#!/usr/bin/perl
##########################################################################
# Daily Run 1.0, managing daily runs of the Upstream Tracker System
# Copyright (C) The Linux Foundation
# Copyright (C) Institute for System Programming, RAS
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License or the GNU Lesser
# General Public License as published by the Free Software Foundation,
# either version 2 of the Licenses, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# and the GNU Lesser General Public License along with this program.
# If not, see <http://www.gnu.org/licenses/>.
##########################################################################
use Getopt::Long;
Getopt::Long::Configure ("posix_default", "no_ignore_case");
use File::Path qw(mkpath rmtree);
use File::Copy qw(copy move);
use Data::Dumper;
use Config;

my ($Help, $TargetLibrary, $MakeSnapshot, $MakePatch, $ApplyPatch,
$UpdateInfo, $SendMailTo, $TransferAddr, $TransferPassword);

my $CmdName = get_FileName($0);
GetOptions("h|help!" => \$Help,
#general options
  "make-snapshot!" => \$MakeSnapshot,
  "make-patch=s" => \$MakePatch,
  "apply-patch=s" => \$ApplyPatch,
  "l|library=s" => \$TargetLibrary,
  "update!" => \$UpdateInfo,
  "mail=s" => \$SendMailTo,
  "transfer|synch=s" => \$TransferAddr,
  "p|password=s" => \$TransferPassword
) or exit(1);

sub HELP_MESSAGE()
{
    print STDERR <<"EOM"

NAME:
  $CmdName - scheduling daily runs and transferring test results
              of the Upstream Tracker System ( http://www.linuxtesting.org/upstream-tracker/ )

USAGE:
  $CmdName [options]

EXAMPLES OF USE:
  $CmdName -update -mail 'a\@mail.ru b\@mail.ru c\@mail.ru'
  $CmdName -transfer root\@site.org:/srv/www/vhosts/site.org/upstream-tracker/ -p PASSWORD

INFORMATION OPTIONS:
  -h|-help
      Print this help.

GENERAL OPTIONS:
  -make-snapshot
      Make snapshot of the current directory and save
      it to the files named <node>-<date>.snapshot and current.snapshot

  -make-patch <snapshot>
      Make patch named <node>-<date>.tar.gz
      with the changes relative to the snapshot.

  -apply-patch <file>
      Apply patch to the current directory.

  -l|-library <name>
      Do all actions for the specified library.

  -update
      Add latest version(s) of a library(s) to the system.
      Will execute "upstream-tracker.pl -update -stat" command.

  -mail <to>
      Send news to the recipients delimited by space or comma.

  -transfer <to>
      Transfer current test results to the central data base site.

  -p|-password <pass>
      Password for the central data base site.

EOM
      ;
}

# Constants
my $REPOSITORY_PATH = "repository";
my %MonNum=(
  "Jan"=>1,
  "Feb"=>2,
  "Mar"=>3,
  "Apr"=>4,
  "May"=>5,
  "Jun"=>6,
  "Jul"=>7,
  "Aug"=>8,
  "Sep"=>9,
  "Oct"=>10,
  "Nov"=>11,
  "Dec"=>12
);

#System
my %SystemPaths;

# Binaries
my %DefaultBinPaths;

my %OperatingSystemAddPaths=(
# this data needed if tool can't determine paths automatically
"default"=>{
    "bin"=>{"/usr/bin"=>1,"/bin"=>1,"/sbin"=>1,"/usr/sbin"=>1}},
"haiku"=>{
    "bin"=>{"/boot/common/bin"=>1,"/boot/system/bin"=>1}}
);

sub get_FileName($)
{
    my $Path = $_[0];
    if($Path=~/\A(.*\/)([^\/]*)\Z/)
    {
        return $2;
    }
    else
    {
        return $Path;
    }
}

sub get_Directory($)
{
    my $Path = $_[0];
    return "" if($Path=~m*\A\./*);
    if($Path=~/\A(.*)[\/]+([^\/]*)\Z/)
    {
        return $1;
    }
    else
    {
        return "";
    }
}

sub transfer_Cmd($)
{
    my $Cmd = $_[0];
    return if(not $Cmd);
    my $TransferCmd = "expect <<TRANSFER 
spawn bash -c \"$Cmd\"
while 1 {
    expect {
    \"(yes/no)?\" {send \"yes\\n\"}
    \"assword\" {send \"$TransferPassword\\n\"}
    eof exit
    }
}
TRANSFER\n";
    writeFile(".tmp_dir/transfer.sh", $TransferCmd);
    system("sh .tmp_dir/transfer.sh>/dev/null");
    unlink(".tmp_dir/transfer.sh");
}

sub getServerAddr($)
{
    my $Addr = $_[0];
    return "" if(not $Addr);
    $Addr=~s/:\/.*\Z//g;
    return $Addr;
}

sub getServerDir($)
{
    my $Addr = $_[0];
    return "" if(not $Addr);
    if($Addr=~/:(\/.*)\Z/)
    {
        return $1;
    }
    else
    {
        return "/";
    }
}

sub transfer_Results($)
{# trasfer patch to the central data base
    my $Addr = $_[0];
    return "" if(not $Addr);
    # copy the tool to the server
    print "transferring tool ...\n";
    transfer_Cmd("scp $0 $Addr");
    # make snapshot of the server
    print "creating snapshot ...\n";
    transfer_Cmd("ssh ".getServerAddr($Addr)." \\'cd ".getServerDir($Addr)." && perl $CmdName -make-snapshot\\'");
    # get snapshot from the server
    transfer_Cmd("scp $Addr/current.snapshot .");
    # create patch using the snapshot
    if(my $Patch_Path = make_Patch("current.snapshot"))
    {
        # copy patch to the server
        print "transferring patch ...\n";
        transfer_Cmd("scp $Patch_Path $Addr");
        # apply patch on the server
        print "applying patch ...\n";
        transfer_Cmd("ssh ".getServerAddr($Addr)." \\'cd ".getServerDir($Addr)." && perl $CmdName -apply-patch ".get_FileName($Patch_Path)."\\'");
    }
}

sub get_State()
{
    my %State = ();
    foreach my $Path (cmd_find("compat_reports","f","abi_compat_report.html"))
    {
        my $ReportDir = get_Directory($Path);
        my $FromTo = get_FileName($ReportDir);
        my $Library = get_FileName(get_Directory($ReportDir));
        if($FromTo=~/\A(.+)_to_(.+)\Z/)
        {
            my ($Version1, $Version2) = ($1, $2);
            if(readFile($Path)=~/generated on (.+) for/i)
            {
                $State{"CompatReports"}{$Library}{$Version1}{$Version2} = $1;
            }
        }
    }
    foreach my $Path (cmd_find("test_results","f","test_results.html"))
    {
        my $ReportDir = get_Directory($Path);
        my $Version = get_FileName($ReportDir);
        my $Library = get_FileName(get_Directory($ReportDir));
        if(readFile($Path)=~/generated on (.+) for/i)
        {
            $State{"TestResults"}{$Library}{$Version} = $1;
        }
    }
    foreach my $Path (cmd_find("changelogs","f","changelog.html"))
    {
        my $ReportDir = get_Directory($Path);
        my $Version = get_FileName($ReportDir);
        my $Library = get_FileName(get_Directory($ReportDir));
        $State{"Changelogs"}{$Library}{$Version} = -s $Path;
    }
    foreach my $Path (cmd_find("tests_t2c","f","*.tar.gz"))
    {
        my $ReportDir = get_Directory($Path);
        my $Version = get_FileName($ReportDir);
        my $Library = get_FileName(get_Directory($ReportDir));
        $State{"TestsT2C"}{$Library}{$Version} = -s $Path;
    }
    if(-d "tests")
    {
        foreach my $Path (split(/\n/,`find tests/ -maxdepth 3 -name "view_tests.html" -type f`))
        {
            my $ReportDir = get_Directory($Path);
            my $Version = get_FileName($ReportDir);
            my $Library = get_FileName(get_Directory($ReportDir));
            if(readFile($Path)=~/generated on (.+) for/i)
            {
                $State{"Tests"}{$Library}{$Version} = $1;
            }
        }
    }
    return \%State;
}

sub make_Snapshot()
{
    my $FName = get_NodeStamp()."-".get_DateStamp(time).".snapshot";
    print "checking current state ...\n";
    my $State = get_State();
    writeFile($FName, Dumper($State));
    writeFile("current.snapshot", Dumper($State));
    print "snapshot has been written to \'$FName\' (copy writtent to \'current.snapshot\')\n";
}

sub apply_Patch($)
{
    my $Patch_Path = $_[0];
    return "" if(not $Patch_Path or not -f $Patch_Path);
    print "applying path\n";
    system("tar -xvzf $Patch_Path");
}

sub make_Patch($)
{
    my $Snapshot_Path = $_[0];
    return "" if(not $Snapshot_Path or not -f $Snapshot_Path);
    my $Content = readFile($Snapshot_Path);
    if($Content!~/};\s*\Z/)
    {
        print STDERR "ERROR: specified snapshot is not valid, try to regenerate it\n";
        exit(1);
    }
    my $Snapshot = eval($Content);
    if(not $Snapshot)
    {
        print STDERR "ERROR: internal error - eval() procedure seem to not working correctly, try to remove 'use strict' and continue\n";
        exit(1);
    }
    print "checking current state ...\n";
    my $State = get_State();
    print "creating patch ...\n";
    rmtree(".tmp/");
    mkpath(".tmp/");
    foreach my $Library (keys(%{$State->{"Changelogs"}}))
    {
        foreach my $Version (keys(%{$State->{"Changelogs"}{$Library}}))
        {
            my $StateSize = $State->{"Changelogs"}{$Library}{$Version};
            my $SnapshotSize = $Snapshot->{"Changelogs"}{$Library}{$Version};
            next if(not $StateSize);
            if(not $SnapshotSize or $StateSize!=$SnapshotSize)
            {
                my $Path = "changelogs/$Library/$Version/changelog.html";
                mkpath(get_Directory(".tmp/$Path"));
                copy($Path, ".tmp/$Path");
            }
        }
    }
    foreach my $Library (keys(%{$State->{"TestResults"}}))
    {
        foreach my $Version (keys(%{$State->{"TestResults"}{$Library}}))
        {
            my $StateDate = $State->{"TestResults"}{$Library}{$Version};
            my $SnapshotDate = $Snapshot->{"TestResults"}{$Library}{$Version};
            next if(not $StateDate);
            if(not $SnapshotDate or cmpDates($StateDate, $SnapshotDate)==1)
            {
                my $Path = "test_results/$Library/$Version/test_results.html";
                mkpath(get_Directory(".tmp/$Path"));
                copy($Path, ".tmp/$Path");
            }
        }
    }
    foreach my $Library (keys(%{$State->{"CompatReports"}}))
    {
        foreach my $Version1 (keys(%{$State->{"CompatReports"}{$Library}}))
        {
            foreach my $Version2 (keys(%{$State->{"CompatReports"}{$Library}{$Version1}}))
            {
                my $StateDate = $State->{"CompatReports"}{$Library}{$Version1}{$Version2};
                my $SnapshotDate = $Snapshot->{"CompatReports"}{$Library}{$Version1}{$Version2};
                next if(not $StateDate);
                if(not $SnapshotDate or cmpDates($StateDate, $SnapshotDate)==1)
                {
                    my $Path = "compat_reports/$Library/$Version1"."_to_".$Version2."/abi_compat_report.html";
                    mkpath(get_Directory(".tmp/$Path"));
                    copy($Path, ".tmp/$Path");
                }
            }
        }
    }
    foreach my $Library (keys(%{$State->{"Tests"}}))
    {
        foreach my $Version (keys(%{$State->{"Tests"}{$Library}}))
        {
            my $StateDate = $State->{"Tests"}{$Library}{$Version};
            my $SnapshotDate = $Snapshot->{"Tests"}{$Library}{$Version};
            next if(not $StateDate);
            if(not $SnapshotDate or cmpDates($StateDate, $SnapshotDate)==1)
            {
                my $Path = "tests/$Library/$Version/view_tests.html";
                mkpath(get_Directory(".tmp/$Path"));
                copy($Path, ".tmp/$Path");
                my $SuitePath = "tests/$Library/$Version/groups";
                mkpath(".tmp/$SuitePath");
                system("cp -fr $SuitePath/* .tmp/$SuitePath/");
            }
        }
    }
    foreach my $Library (keys(%{$State->{"TestsT2C"}}))
    {
        foreach my $Version (keys(%{$State->{"Tests"}{$Library}}))
        {
            my $StateSize = $State->{"TestsT2C"}{$Library}{$Version};
            my $SnapshotSize = $Snapshot->{"TestsT2C"}{$Library}{$Version};
            next if(not $StateSize);
            if(not $SnapshotSize or $SnapshotSize!=$StateSize)
            {
                my $Path = "tests_t2c/$Library/$Version/t2c-$Library-tests-$Version.tar.gz";
                mkpath(get_Directory(".tmp/$Path"));
                copy($Path, ".tmp/$Path");
            }
        }
    }
    system("cp -fr versions .tmp/");
    copy("index.html", ".tmp/");
    copy("statistics.html", ".tmp/statistics.html");
    copy("news.html", ".tmp/news.html");
    my $FName = get_NodeStamp()."-".get_DateStamp(time).".patch.tar";
    unlink($FName);
    system("cd .tmp && tar -cf $FName * && mv $FName ../");
    rmtree(".tmp/");
    unlink($FName.".gz");
    system("gzip $FName --best");
    print "patch has been written to \'$FName.gz\'\n";
    return "$FName.gz";
}

sub cmpDates($$)
{
    my ($Date1, $Date2) = @_;
    return 0 if(not $Date1 or not $Date2);
    return 0 if($Date1 eq $Date2);
    my ($sec1,$min1,$hour1,$day1,$mon1,$year1);
    my ($sec2,$min2,$hour2,$day2,$mon2,$year2);
    return 0 if($Date1!~/[a-z]+\s+([a-z]+)\s+(\d+)\s+(\d+):(\d+):(\d+)\s+(\d+)/i);
    ($sec1,$min1,$hour1,$day1,$mon1,$year1) = ($5, $4, $3, $2, $MonNum{$1}, $6);
    return 0 if($Date2!~/[a-z]+\s+([a-z]+)\s+(\d+)\s+(\d+):(\d+):(\d+)\s+(\d+)/i);
    ($sec2,$min2,$hour2,$day2,$mon2,$year2) = ($5, $4, $3, $2, $MonNum{$1}, $6);
    if($year1>$year2)
    {
        return 1;
    }
    elsif($year2>$year1)
    {
        return -1;
    }
    elsif($mon1>$mon2)
    {
        return 1;
    }
    elsif($mon2>$mon1)
    {
        return -1;
    }
    elsif($day1>$day2)
    {
        return 1;
    }
    elsif($day2>$day1)
    {
        return -1;
    }
    elsif($hour1>$hour2)
    {
        return 1;
    }
    elsif($hour2>$hour1)
    {
        return -1;
    }
    elsif($min1>$min2)
    {
        return 1;
    }
    elsif($min2>$min1)
    {
        return -1;
    }
    elsif($sec1>$sec2)
    {
        return 1;
    }
    elsif($sec2>$sec1)
    {
        return -1;
    }
    else
    {
        return 0;
    }
}

sub writeFile($$)
{
    my ($Path, $Content) = @_;
    return if(not $Path);
    if(my $Dir = get_Directory($Path))
    {
        mkpath($Dir);
    }
    open (FILE, ">".$Path) || die ("can't open file \'$Path\': $!\n");
    print FILE $Content;
    close(FILE);
}

sub readFile($)
{
    my $Path = $_[0];
    return "" if(not $Path or not -f $Path);
    open (FILE, $Path);
    my $Content = join("", <FILE>);
    close(FILE);
    return toUnix($Content);
}

sub toUnix($)
{
    my $Text = $_[0];
    $Text=~s/\r//g;
    return $Text;
}

sub esc($)
{
    my $Str = $_[0];
    $Str =~ s/([()\[\]{}$ &'"`;,<>])/\\$1/g;
    return $Str;
}

sub cmd_find($$$)
{
    my ($Path, $Type, $Name) = @_;
    return () if(not $Path or not -e $Path);
    my $Cmd = "find ".esc($Path);
    if($Type)
    {
        $Cmd .= " -type $Type";
    }
    if($Name)
    {
        $Cmd .= " -name \"$Name\"";
    }
    return split(/\n/, `$Cmd`);
}

sub get_NodeStamp()
{
    my $Node = `uname -n`;
    chomp($Node);
    return $Node;
}

sub get_DateStamp($)
{
    my $Time = $_[0];
    my ($sec,$min,$hour,
    $mday,$mon,$year,$wday,
    $yday,$isdst) = localtime($Time);
    return ($year+1900)."-".normalize_num($mon)."-".normalize_num($mday);
}

sub normalize_num($)
{
    my $Num = $_[0];
    if(int($Num)<10)
    {
        return "0".$Num;
    }
    else
    {
        return $Num;
    }
}

sub updateInfo_Library($)
{
    my $Library = $_[0];
    return if(not $Library);
    if(not -f "upstream-tracker.pl")
    {
        print STDERR "ERROR: upstream-tracker.pl was not found\n";
        exit(1);
    }
    my $Cmd = "perl upstream-tracker.pl -update -stat -l $Library";
    print "running $Cmd\n";
    system($Cmd);
}

sub updateInfo()
{
    if(not -f "upstream-tracker.pl")
    {
        print STDERR "ERROR: upstream-tracker.pl was not found\n";
        exit(1);
    }
    my $Cmd = "perl upstream-tracker.pl -update -stat";
    print "running $Cmd\n";
    system($Cmd);
}

sub makeReport()
{
    my $UTLog = readFile("news.html");
    $UTLog=~s/('|")(compat_reports\/|test_results\/|tests\/|versions\/)/$1\http:\/\/linuxtesting.org\/upstream-tracker\/$2/g;
    return $UTLog;
}

sub get_CmdPath($)
{
    my $Name = $_[0];
    return "" if(not $Name);
    if(my $DefaultPath = get_CmdPath_Default($Name))
    {
        return $DefaultPath;
    }
    foreach my $Path (sort {length($a)<=>length($b)} keys(%{$SystemPaths{"bin"}}))
    {
        if(-f $Path."/".$Name)
        {
            return $Path."/".$Name;
        }
    }
    return "";
}

sub get_CmdPath_Default($)
{
    my $Name = $_[0];
    return "" if(not $Name);
    if(`$Name --version 2>/dev/null`)
    {
        return $Name;
    }
    foreach my $Path (sort {length($a)<=>length($b)} keys(%DefaultBinPaths))
    {
        if(-f $Path."/".$Name)
        {
            return $Path."/".$Name;
        }
    }
    return "";
}

sub detect_bin_default_paths()
{
    my $EnvPaths = $ENV{"PATH"};
    if($Config{"osname"}=~/\A(haiku|beos)\Z/)
    {
        $EnvPaths.=":".$ENV{"BETOOLS"};
    }
    foreach my $Path (sort {length($a)<=>length($b)} split(/:|;/, $EnvPaths))
    {
        if($Path ne "/")
        {
            $Path=~s/[\/]+\Z//g;
            next if(not $Path);
        }
        $DefaultBinPaths{$Path} = 1;
    }
}

sub detect_default_paths()
{
    my $TargetCfg = ($Config{"osname"}=~/\A(haiku|beos)\Z/)?"haiku":"default";
    foreach my $Type (keys(%{$OperatingSystemAddPaths{$TargetCfg}}))
    {# additional search paths
        foreach my $Path (keys(%{$OperatingSystemAddPaths{$TargetCfg}{$Type}}))
        {
            next if(not -d $Path);
            $SystemPaths{$Type}{$Path} = $OperatingSystemAddPaths{$TargetCfg}{$Type}{$Path};
        }
    }
    detect_bin_default_paths();
    foreach my $Path (keys(%DefaultBinPaths))
    {
        $SystemPaths{"bin"}{$Path} = $DefaultBinPaths{$Path};
    }
    foreach my $Path (split(/\n/, `find / -maxdepth 1 -name "*bin*" -type d`))
    {# autodetecting bin directories
        $SystemPaths{"bin"}{$Path} = 1;
    }
}

sub sendReport($$)
{
    my ($To, $Report) = @_;
    return if(not $To or not $Report);
    my $MailCmd = get_CmdPath("mailx");
    return if(not $MailCmd);
    writeFile(".tmp/news.html", $Report);
    writeFile(".tmp/body", "");
    foreach my $Addr (split(/\s+|\,/, $To))
    {
        print "sending mail to <$Addr> ... ";
        system("$MailCmd -s 'Upstream Tracker News' -r upstream-tracker -a .tmp/news.html $Addr < .tmp/body");
        if($?)
        {
            print "fail\n";
            print STDERR "ERROR: can't send mail to $Addr\n";
        }
        else
        {
            print "done\n";
        }
    }
    rmtree(".tmp/");
}

sub scenario()
{
    if(defined $Help)
    {
        HELP_MESSAGE();
        exit(0);
    }
    detect_default_paths();
    if($SendMailTo
    and not get_CmdPath("mailx"))
    {
        print STDERR "ERROR: can't find mailx\n";
        exit(1);
    }
    if($UpdateInfo)
    {
        if($TargetLibrary)
        {
            updateInfo_Library($TargetLibrary);
        }
        else
        {
            updateInfo();
        }
    }
    if($MakeSnapshot)
    {
        make_Snapshot();
    }
    if($MakePatch)
    {
        if(not -e $MakePatch)
        {
            print STDERR "ERROR: can't access file \'$MakePatch\'\n";
            exit(1);
        }
        else
        {
            make_Patch($MakePatch);
        }
    }
    if($ApplyPatch)
    {
        apply_Patch($ApplyPatch);
    }
    if($TransferAddr)
    {
        transfer_Results($TransferAddr);
    }
    if($SendMailTo)
    {
        sendReport($SendMailTo, makeReport());
    }
    exit(0);
}

scenario();
