#!/usr/bin/perl
##########################################################################
# Upstream Tracker 1.3, monitoring and analyzing C/C++ libraries development
# Copyright (C) The Linux Foundation
# Copyright (C) Institute for System Programming, RAS
# Author: Andrey Ponomarenko
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License or the GNU Lesser
# General Public License as published by the Free Software Foundation,
# either version 2 of the Licenses, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# and the GNU Lesser General Public License along with this program.
# If not, see <http://www.gnu.org/licenses/>.
##########################################################################
use Getopt::Long;
Getopt::Long::Configure ("posix_default", "no_ignore_case");
use File::Path qw(mkpath rmtree);
use File::Copy qw(copy move);
use Cwd qw(abs_path);
use Config;
use Net::FTP;

my $UPSTREAM_TRACKER_VERSION = "1.3";
my ($Help, $HelpDB, $ShowVersion, $ListKnownLibs, $GetNew, $TargetVersion, $TargetVersion1,
$TargetVersion2, $IncludeCurrent, $TargetLibrary, $GenT2C, $TestAPI, $CmpABI, $Install,
$ReInstall, $UnInstall, $Alter, $GetChangelogs, $GetCurrent, $ShowStat, $MakeCppCompat,
$UseDebugLib, $UpdateInfo, $RemoveInfo);

my $CmdName = get_FileName($0);
GetOptions("h|help!" => \$Help,
#information options
  "help-db!" => \$HelpDB,
  "about!" => \$ShowVersion,
  "list!" => \$ListKnownLibs,
#general options
  "l|library=s" => \$TargetLibrary,
  "get|download!" => \$GetNew,
  "update!" => \$UpdateInfo,
  "i|install!" => \$Install,
  "r|reinstall!" => \$ReInstall,
  "u|uninstall!" => \$UnInstall,
  "alter!" => \$Alter,
  "test-api!" => \$TestAPI,
  "gen-t2c!" => \$GenT2C,
  "v|version=s" => \$TargetVersion,
  "cmp-abi!" => \$CmpABI,
  "v1=s" => \$TargetVersion1,
  "v2=s" => \$TargetVersion2,
  "changelog" => \$GetChangelogs,
  "debug" => \$UseDebugLib,
  "remove" => \$RemoveInfo,
#extra options
  "get-current|download-current!" => \$GetCurrent,
  "current" => \$IncludeCurrent,
  "mkcpp" => \$MakeCppCompat,
  "stat" => \$ShowStat
) or exit(1);

sub HELP_MESSAGE()
{
    print STDERR <<"EOM"

NAME:
  $CmdName - monitoring and analyzing C/C++ libraries development

DESCRIPTION:
  Upstream Tracker is a tool for monitoring and analyzing C/C++
  libraries development. It was intended for automatically checking new
  versions of libraries for unintentional ABI breaks and critical
  runtime errors.

  The tool is able to find, download and locally install versions of many
  popular libraries for analyzing. It uses ABI Compliance Checker for
  tracking ABI changes and API Sanity Autotest for the runtime testing.

  This tool is free software: you can redistribute it and/or modify
  it under the terms of the GNU GPL or GNU LGPL.

USAGE:
  $CmdName [options]

EXAMPLES OF USE:
  Get new versions:
    $CmdName -l bzip2 -get
  Install all versions:
    $CmdName -l bzip2 -install
  Runtime testing:
    $CmdName -l bzip2 -test-api
  Tracking ABI changes:
    $CmdName -l bzip2 -cmp-abi
  Add latest version(s):
    $CmdName -l bzip2 -update

INFORMATION OPTIONS:
  -h|-help
      Print this help.

  -help-db
      Print help for the library data base entries.

  -about
      Print version information.

  -list
      List all known libraries.

GENERAL OPTIONS:
  -l|-library <name>
      The library name. See lib_info.xml for details.

  -download|-get
      Download all available versions of a library(s).
  
  -update
      Add latest version(s) of a library(s) to the system:
        1. Download new version(s)
        2. Install it locally
        3. Check for ABI compatibility relative to
           the previous version
        4. Test the API for serious failures

  -i|-install
      Install locally library versions.

  -r|-reinstall
      Reinstall versions.

  -u|-uninstall
      Uninstall versions.

  -test-api
      Generate, build and execute runtime tests for the library
      using API-Sanity-Autotest.

  -gen-t2c
      Generate tests in the Template2Code format
      using API-Sanity-Autotest with the -t2c option.
      For more information, please see:
        http://sourceforge.net/projects/template2code/

  -v|-version <ver>
      This option allows to specify version of the library
      for installing or for runtime testing (-install, -test-api).

  -cmp-abi
      Check sequential versions of the library for binary
      compatibility using ABI-compliance-checker.

  -v1 <ver>
  -v2 <ver>
      These options allow to specify versions of the library for
      checking binary compatibility (-cmp-abi).

  -changelog
      Search for changelogs in the source packages/on the web site.
      Use <changelog_links> section in the database for specifying
      web site link templates.

  -alter
      Alter already obtained test results. The system has accumulation
      architecture, so for performing again any actions you can use this option.

  -debug
      Installing or testing of the library debug versions.

  -remove
      Remove particular library version from the system.
      Use it with -l and -v options.

EXTRA OPTIONS

  -download-current|-get-current
      Update/Checkout current (scm) version of the library.

  -current
      Add current (scm) version to the queue of versions for analizing/installing.

  -mkcpp
      Rename "this" and "class" keywords to "c_this" and "c_class"
      in the library header files. This option is needed for correct
      compiling of library header files by the g++ compiler.

  -stat
      Create statistics of ABI changes analysis for all known libraries.

DATABASE ENTRY:
  The tool uses a database (lib_info.xml) with the entries for
  many upstream libraries. Use -help-db option for more information:
    $CmdName -help-db

Report bugs to <upstream-tracker\@linuxtesting.org>
For more information, please see: http://ispras.linux-foundation.org/index.php/Upstream_Tracker
EOM
      ;
}

sub HELP_MESSAGE_DB()
{
    print STDERR <<"EOM"
DATABASE ENTRY:

  The tool uses a database (lib_info.xml) with the entries
  for many upstream libraries. The database entry has the
  following format:

<library>

<!--
     Primary Sections  
                        -->

    <name>
        <!-- Short name of the library (should be as simple as possible)
             It affects on the path to source packages, reports and tests -->
    </name>

    <full_name>
        <!-- Full name of the library.
             It will be displayed in the index.html -->
    </full_name>

    <summary>
        <!-- Summary of the library.
             It will be displayed in the index.html -->
    </summary>

    <description>
        <!-- Library description.
             Wasn't implemented yet -->
    </description>

    <site>
        <!-- Web site/page of the library.
             It will be displayed in the index.html -->
    </site>

    <bugs>
        <!-- Web site/page for reporting bugs.
             Wasn't implemented yet -->
    </bugs>

    <doc>
        <!-- Web site/page with the documentation.
             It will be displayed in the index.html -->
    </doc>

    <download>
        <!-- Web pages with the source packages, one per line.
             Allowed extensions: "tar.gz", "tgz", "tar.bz2", "tar.Z", "zip" and "tar" -->
    </download>

    <package>
        <!-- The source code package template:
             Example: Linux-PAM-{VERSION}.tar.gz -->
    </package>

    <mailing_list>
        <!-- Web site/page with the mailing lists.
             Wasn't implemented yet -->
    </mailing_list>

    <scm>
        <!-- Library SCM name.
             Select it from the following list:
                cvs
                svn
                git
                rsync
                bazaar
                mercurial
                darcs        -->
    </scm>

    <scm_addr>
        <!-- Library SCM address -->
    </scm_addr>

    <scm_module>
        <!-- Library SCM module (cvs only) -->
    </scm_module>

    <scm_branch>
        <!-- Library SCM branch -->
    </scm_branch>

    <scm_password>
        <!-- Library SCM password -->
    </scm_password>

<!--
     Additional Sections
                         -->

    <include_preamble>
        <!-- The list of header files that should be included before other headers, one per line.
             Examples: tree.h in libxml2
                       ft2build.h in freetype2 -->
    </include_preamble>

    <include_paths>
        <!-- The list of paths to be searched for header files
             needed for compiling of library headers, one per line.
             NOTE: If you define this section then the tool
             will not automatically detect include paths -->
    </include_paths>

    <headers>
        <!-- The list of headers that should be processed, one name per line.
             Other headers will not be processed -->
    </headers>

    <libs>
        <!-- The list of shared objects that should be processed, one name per line.
             Other shared objects will not be processed -->
    </libs>

    <skip_headers>
        <!-- The list of header files and/or directories with header files
             that should not be processed, one name per line -->
    </skip_headers>

    <skip_libs>
        <!-- The list of shared objects and/or directories with shared objects
             that should not be processed, one name per line -->
    </skip_libs>

    <skip_interfaces>
        <!-- The list of functions (mangled/symbol names in C++)
             that should be skipped while testing, one per line -->
    </skip_interfaces>

    <changelog_links>
        <!-- The list of templates for changelog web pages, one per line
             Example: http://www.alsa-project.org/main/index.php/Changes_v{V1}_v{V2} -->
    </changelog_links>

    <gcc_options>
        <!-- Additional gcc options, one per line -->
    </gcc_options>

    <out_params>
        <!-- Associating of out(returned)-parameters
             with interfaces, one entry per line:
               func_name:param_name
                    or
               func_name:param_num
             Examples:
               dbus_parse_address:entry
               dbus_parse_address:2    -->
    </out_params>

    <skip_constants>
        <!-- The list of constants that should not be checked, one name per line -->
    </skip_constants>
    
    <skip_warnings>
        <!-- The list of warnings that should not be shown in the report, one pattern per line -->
    </skip_warnings>
    
    <configure_options>
        <!-- The list of additional configure options, one per line -->
    </configure_options>

    <build>
        <!-- The list of build rules that should redefine automatic rules, one per line. -->
    </build>

    <install>
        <!-- The list of install rules that should redefine automatic rules, one per line. -->
    </install>

    <pre_install>
        <!-- Additional pre-installation rules, one per line -->
    </pre_install>

    <post_install>
        <!-- Additional post-installation rules, one per line -->
    </post_install>

    <include_order>
        <!-- Headers include order, one entry per line:
               <header_name_before>:<header_name_after> -->
    </include_order>

    <versioning>
        <!-- Corrections of the standard versioning algorithm,
             one entry per line:
               skip:pattern
                 or
               <version_type>:pattern
             Examples:
               skip:x\-r\d+
               release:\d+\w*
               beta:<pattern>
               alfa:<pattern>    -->
    </versioning>
    
    <skip_on_update>
        <!-- Actions that should not be performed on update, one per line:
               runtime_tests
               abi_tracking
               installing
               get_new_versions
               get_changelog    -->
    </skip_on_update>

    <run_mode>
        <!-- Runtime tests execution mode.
             Select it from the following list:
               xvfb
               default    -->
    </run_mode>

    <install_mode>
        <!-- Installation mode.
             Select it from the following list:
               interactive
               default    -->
    </install_mode>

</library>

EOM
      ;
}

# Constants
my $INSTALL_PATH_DEFAULT = "testing";
my $INSTALL_PATH_DEBUG = $INSTALL_PATH_DEFAULT."_debug";
my $INSTALL_PATH = $UseDebugLib?$INSTALL_PATH_DEBUG:$INSTALL_PATH_DEFAULT;
my $REPOSITORY_PATH = "repository";
my $TIMEOUT = 15;
my $MAX_VIRTUAL_LINKS_ATTEMPTS = 3;
my $MAX_WEB_SEARCH_DEPTH = 5;
my $TOOL_SIGNATURE = "<hr/><div style='width:100%;font-family:Arial;font-size:11px;' align='right'><i>Generated on ".localtime(time)." by <a href='http://ispras.linux-foundation.org/index.php/Upstream_Tracker'>Upstream-Tracker</a> $UPSTREAM_TRACKER_VERSION &nbsp;</i></div>";

my %MonNum=(
  "Jan"=>1,
  "January"=>1,
  "Feb"=>2,
  "February"=>2,
  "Mar"=>3,
  "March"=>3,
  "Apr"=>4,
  "April"=>4,
  "May"=>5,
  "Jun"=>6,
  "June"=>6,
  "Jul"=>7,
  "July"=>7,
  "Aug"=>8,
  "August"=>8,
  "Sep"=>9,
  "September"=>9,
  "Oct"=>10,
  "October"=>10,
  "Nov"=>11,
  "November"=>11,
  "Dec"=>12,
  "December"=>12
);

# Global variables
my %LibraryInfo;
my %RepositoryFiles;
my %RepositoryProvide;
my %RepositoryDate;
my %VersionPath;
my %Cache;
my %TestStat;
my %UpdateLog;

# Web
my %BadLinks;
my %WebDirectory;

#Signals
my %SigNo;
my %SigName;

#System
my %SystemPaths;

# Binaries
my %DefaultBinPaths;

# Recursion locks
my @RecurPage;

my %OperatingSystemAddPaths=(
# this data needed if tool can't determine paths automatically
"default"=>{
    "bin"=>{"/usr/bin"=>1,"/bin"=>1,"/sbin"=>1,"/usr/sbin"=>1}},
"haiku"=>{
    "bin"=>{"/boot/common/bin"=>1,"/boot/system/bin"=>1}}
);

sub get_CmdPath($)
{
    my $Name = $_[0];
    return "" if(not $Name);
    return $Cache{"get_CmdPath"}{$Name} if(defined $Cache{"get_CmdPath"}{$Name});
    if(my $DefaultPath = get_CmdPath_Default($Name))
    {
        $Cache{"get_CmdPath"}{$Name} = $DefaultPath;
        return $DefaultPath;
    }
    foreach my $Path (sort {length($a)<=>length($b)} keys(%{$SystemPaths{"bin"}}))
    {
        if(-f $Path."/".$Name)
        {
            $Cache{"get_CmdPath"}{$Name} = $Path."/".$Name;
            return $Path."/".$Name;
        }
    }
    $Cache{"get_CmdPath"}{$Name} = "";
    return "";
}

sub get_CmdPath_Default($)
{
    my $Name = $_[0];
    return "" if(not $Name);
    return $Cache{"get_CmdPath_Default"}{$Name} if(defined $Cache{"get_CmdPath_Default"}{$Name});
    if(`$Name --version 2>/dev/null`)
    {
        $Cache{"get_CmdPath_Default"}{$Name} = $Name;
        return $Name;
    }
    foreach my $Path (sort {length($a)<=>length($b)} keys(%DefaultBinPaths))
    {
        if(-f $Path."/".$Name)
        {
            $Cache{"get_CmdPath_Default"}{$Name} = $Path."/".$Name;
            return $Path."/".$Name;
        }
    }
    $Cache{"get_CmdPath_Default"}{$Name} = "";
    return "";
}

sub cmpVersions($$)
{
    my ($First, $Second) = @_;
    if($First eq "" and $Second eq "")
    {
        return 0;
    }
    elsif($First eq ""
    or $Second eq "current")
    {
        return -1;
    }
    elsif($Second eq ""
    or $First eq "current")
    {
        return 1;
    }
    $First=~s/(\d)([a-z])/$1.$2/ig;
    $Second=~s/(\d)([a-z])/$1.$2/ig;
    $First=~s/(_|-)/./g;
    $Second=~s/(_|-)/./g;
    $First=~s/\A(_|-|\.)+//g;
    $Second=~s/\A(_|-|\.)+//g;
    $First=~s/\A[0]+([1-9]\d*)\Z/$1/g;
    $Second=~s/\A[0]+([1-9]\d*)\Z/$1/g;
    $First=~s/\A[0]+\Z/0/g;
    $Second=~s/\A[0]+\Z/0/g;
    if($First!~/\./ and $Second!~/\./)
    {
        return mixedCmp($_[0], $_[1]);
    }
    elsif($First!~/\./)
    {
        return cmpVersions($First.".0", $Second);
    }
    elsif($Second!~/\./)
    {
        return cmpVersions($First, $Second.".0");
    }
    else
    {
        my ($Part1, $Part2) = ();
        if($First =~ s/\A([^\.]+)\.//o)
        {
            $Part1 = $1;
        }
        if($Second =~ s/\A([^\.]+)\.//o)
        {
            $Part2 = $1;
        }
        if(my $CmpPartRes = mixedCmp($Part1, $Part2))
        {# compare first parts
            return $CmpPartRes;
        }
        else
        {# compare other parts
            return cmpVersions($First, $Second);
        }
    }
}

sub tokensCmp($$)
{
    my ($First, $Second) = @_;
    if($First eq $Second)
    {
        return 0;
    }
    elsif($First=~/\A[^a-z0-9]+\Z/i
    and $Second=~/\A[^a-z0-9]+\Z/i)
    {
        return 0;
    }
    elsif($First=~/\A[a-z]+\Z/i
    and $Second=~/\A[a-z]+\Z/i)
    {
        return trivialSymbCmp($First, $Second);
    }
    elsif($First=~/\A\d+\Z/
    and $Second=~/\A\d+\Z/)
    {
        return trivialNumerCmp($First, $Second);
    }
    elsif($First=~/\A[a-z]+\Z/i
    and $Second=~/\A\d+\Z/)
    {
        return -1;
    }
    elsif($First=~/\A\d+\Z/
    and $Second=~/\A[a-z]+\Z/i)
    {
        return 1;
    }
    elsif($First and $Second eq "")
    {
        return 1;
    }
    elsif($Second and $First eq "")
    {
        return -1;
    }
    else
    {
        return "undef";
    }
}

sub mixedCmp($$)
{
    my ($First, $Second) = @_;
    if($First eq $Second)
    {
        return 0;
    }
    while($First ne ""
    and $Second ne "")
    {
        my $First_Token = get_Token($First);
        my $Second_Token = get_Token($Second);
        my $CmpRes = tokensCmp($First_Token, $Second_Token);
        if($CmpRes eq "undef")
        {# safety Lock
            return 0;
        }
        elsif($CmpRes != 0)
        {
            return $CmpRes;
        }
        else
        {
            $First =~ s/\A\Q$First_Token\E//g;
            $Second =~ s/\A\Q$Second_Token\E//g;
        }
    }
    if($First ne ""
    or $First eq "0")
    {
        return 1;
    }
    elsif($Second ne ""
    or $Second eq "0")
    {
        return -1;
    }
    else
    {
        return 0;
    }
}

sub get_Token($)
{
    if($_[0]=~/\A(\d+)[a-z]/i)
    {
        return $1;
    }
    elsif($_[0]=~/\A([a-z]+)\d/i)
    {
        return $1;
    }
    elsif($_[0]=~/\A(\d+)[^a-z0-9]/i)
    {
        return $1;
    }
    elsif($_[0]=~/\A([a-z]+)[^a-z0-9]/i)
    {
        return $1;
    }
    elsif($_[0]=~/\A([^a-z0-9]+)/i)
    {
        return $1;
    }
    else
    {
        return $_[0];
    }
}

sub trivialSymbCmp($$)
{
    my ($First, $Second) = @_;
    if($MonNum{$First} and $MonNum{$Second})
    {
        return ($MonNum{$First} <=> $MonNum{$Second});
    }
    elsif($First gt $Second)
    {
        return 1;
    }
    elsif($First eq $Second)
    {
        return 0;
    }
    else
    {
        return -1;
    }
}

sub trivialNumerCmp($$)
{
    if(int($_[0]) > int($_[1]))
    {
        return 1;
    }
    elsif($_[0] eq $_[1])
    {
        return 0;
    }
    else
    {
        return -1;
    }
}

sub check_network()
{
    return 1 if(check_addr("www.google.com"));
    return 1 if(check_addr("www.linuxfoundation.org"));
    return 0;
}

sub check_addr($)
{
    my $Addr = $_[0];
    $Addr=~s/\Aftp\.//g;
    system("ping $Addr -q -c 1 >/dev/null 2>&1");
    if($?)
    {
        return check_page($Addr);
    }
    else
    {
        return 1;
    }
}

sub check_page($)
{
    my $Page = $_[0];
    if(`wget --connect-timeout=$TIMEOUT --spider $Page 2>&1`=~/200\s+OK/i)
    {
        return 1;
    }
    else
    {
        return 0;
    }
}

sub getNewVersions_Library($)
{
    my $Library = $_[0];
    return 0 if(not $LibraryInfo{$Library});
    if($IncludeCurrent)
    {
        print "searching for current (scm) version of the $Library ...\n";
        getCurrentVersion_Library($Library);
    }
    print "searching for release versions of the $Library ...\n";
    my %AllPages = ();
    my $Found = 0;
    foreach my $TopPage (keys(%{$LibraryInfo{$Library}{"Download"}}))
    {# direct search
        $Found+=search_packages($Library, $TopPage);
    }
    foreach my $TopPage (keys(%{$LibraryInfo{$Library}{"Download"}}))
    {# recursive search
        foreach my $Page (sort {cmp_paths($b, $a)} get_pages($Library, $TopPage, $TopPage))
        {
            $Found+=search_packages($Library, $Page);
        }
    }
    if(not $Found and not keys(%{$RepositoryProvide{$Library}}))
    {
        print STDERR "ERROR: can't obtain packages with $Library versions, try to download it manually to \'$REPOSITORY_PATH/$Library/\'\n";
        return 0;
    }
    return $Found;
}

sub cmp_paths($$)
{
    my ($Path1, $Path2) = @_;
    my @Parts1 = split(/\//, $Path1);
    my @Parts2 = split(/\//, $Path2);
    foreach my $Num (0 .. $#Parts1)
    {
        my $Part1 = $Parts1[$Num];
        my $Part2 = $Parts2[$Num];
        if($Part1=~/\d/
        and $Part2=~/\d/)
        {
            return cmpVersions($Part1, $Part2);
        }
        elsif($Part1 cmp $Part2)
        {
            return 1;
        }
    }
    return 0;
}

sub detect_web_directories($)
{
    my $Content = $_[0];
    return if(not $Content);
    while($Content=~s/(\d+)\.\s+(.+)(\s|\Z)//)
    {
        my ($Num, $Link) = ($1, $2);
        if($Content=~/directory\s+\[$Num\]/i)
        {
            $WebDirectory{$Link} = 1;
        }
    }
}

sub list_ftp_dir($)
{
    my $Addr = $_[0];
    my $Content = "";
    if(my $Ftp = Net::FTP->new(getSiteAddr($Addr)))
    {
        if($Ftp->login("anonymous"))
        {
            foreach my $Entry ($Ftp->dir(getSiteDir($Addr)))
            {
                if($Entry=~/\A([^\s]+)\s+.+\s+([^\s]+)\Z/)
                {
                    my ($Attr, $SubDir) = ($1, $2);
                    my $Link = $Addr."/".$SubDir;
                    next if($SubDir eq getSiteDir($Addr));
                    $Content .= $Link."\n";
                    if($Attr=~/\Ad/)
                    {
                        $WebDirectory{$Link} = 1;
                    }
                }
            }
        }
        $Ftp->quit();
    }
    return $Content;
}

sub read_page($$)
{
    my ($Library, $Page) = @_;
    return "" if(not $Library or not $Page);
    mkpath($REPOSITORY_PATH."/".$Library);
    my $DownloadTo = $REPOSITORY_PATH."/".$Library."/page.html";
    system("wget \"$Page\" --quiet --connect-timeout=$TIMEOUT --tries=2 --output-document=\"$DownloadTo\"");
    if(not $?)
    {
        return $DownloadTo;
    }
    if($Page=~/\Aftp:\/\//)
    {
        if(my $DirList = list_ftp_dir($Page))
        {
            writeFile($DownloadTo, $DirList);
            return $DownloadTo;
        }
    }
    if(my $LynxCmd = get_CmdPath("lynx"))
    {
        system(get_safe_cmd("lynx -connect_timeout=10 -dump \"$Page\" > \"$DownloadTo\" 2>&1", 5));
        if(not $? or $SigName{$?} eq "KILL")
        {
            detect_web_directories(readFile($DownloadTo));
            return $DownloadTo;
        }
    }
    print STDERR "ERROR: can't read page \'$Page\'\n";
    @{$Cache{"get_links"}{$Page}} = ();
    return "";
}

sub get_safe_cmd($$)
{
    my ($Cmd, $Timeout) = @_;
    return "" if(not $Cmd or not $Timeout);
    return "{ $Cmd ; kill -s KILL \$\$ & sleep $Timeout; kill -s KILL \$! ;}";
}

sub get_links($$)
{
    my ($Library, $Page) = @_;
    return () if(not $Library or not $Page);
    if($Cache{"get_links"}{$Page})
    {
        return @{$Cache{"get_links"}{$Page}};
    }
    my $Site = getSiteAddr($Page);
    if(not check_addr($Site))
    {
        print STDERR "ERROR: can't access \'$Site\'\n";
        @{$Cache{"get_links"}{$Page}} = ();
        return ();
    }
    my $DownloadTo = read_page($Library, $Page);
    if(not $DownloadTo)
    {
        return ();
    }
    my $Content = readFile($DownloadTo);
    unlink($DownloadTo);
    my (%Links1, %Links2, %Links3) = ();
    while($Content =~ s/(src|href)\s*\=\s*["']\s*((ftp|http|https):\/\/[^"'<>\s]+?)\s*["']//i)
    {
        $Links1{$2}=1;
    }
    while($Content =~ s/(src|href)\s*\=\s*["']\s*([^"'<>\s]+?)\s*["']//i)
    {
        $Links2{linkSum($Page, $2)}=1;
    }
    while($Content=~s/((ftp|http|https):\/\/[^"'<>\s]+?(\.tar\.bz2|\.tar\.gz|\.tar\.Z|\.tgz|\.tar|\.zip|))([\s"']|\Z)//i)
    {
        $Links3{$1}=1;
    }
    my @Links = ();
    foreach my $Link (sort {$b cmp $a} keys(%Links1), sort {$b cmp $a} keys(%Links2), sort {$b cmp $a} keys(%Links3))
    {
        next if($Link=~/\/\.\.\// and $Link!~/\.(tar\.bz2|tar\.gz|\.tar\.Z|tgz|tar|zip)\Z/);
        next if($Link=~/\/(apps\/wordpress|builds|build)/i);
        $Link=~s/\?.+\Z//g;
        if(not getSiteProtocol($Link))
        {
            $Link = getSiteProtocol($Page).$Link;
        }
        push(@Links, $Link);
    }
    @{$Cache{"get_links"}{$Page}} = @Links;
    return @{$Cache{"get_links"}{$Page}};
}

sub canonify_version($)
{
    my $Version = $_[0];
    return "" if(not $Version);
    $Version=~s/\Acurrent[0-9._:-]+\Z/current/i; # current (scm)
    while($Version=~s/(\d)(_|-)(\d)/$1.$3/g){ };
    return $Version;
}

sub filter_version($)
{
    my $Version = $_[0];
    return "" if(not $Version);
    $Version=~s/\Acurrent[0-9._:-]+\Z/current/i; # current (scm)
    $Version=~s/\A(v|ver|version)(\d)/$2/i;
    $Version=~s/\A(source|src|sources|open|stable)(-|_|\.|)(\d)/$3/i;
    $Version=~s/(-|_|\.)(source|src|sources|open|stable|orig|m2|m3|m4|x86|linux|unix)\Z//i;
    return $Version;
}

sub get_package_link_version($$)
{
    my ($Library, $Link) = @_;
    return "" if(not $Library or not $Link);
    return get_default_package_link_version($Library, $Link) if(not keys(%{$LibraryInfo{$Library}{"Package"}}));
    foreach my $PackageTmpl (keys(%{$LibraryInfo{$Library}{"Package"}}))
    {
        if(my ($LinkPrefix, $LinkSuffix) = split(/\{VERSION\}/, $PackageTmpl))
        {
            if($LinkSuffix)
            {
                if($LinkPrefix)
                {
                    if($Link=~/([^"'<>]+?\/|\A)\Q$LinkPrefix\E([^\/"'<>+]+?)\Q$LinkSuffix\E(\.tar\.bz2|\.tar\.gz|\.tar\.Z|\.tgz|\.tar|\.zip|)\Z/i)
                    {
                        my $Version = $2;
                        return filter_version($Version);
                    }
                }
                else
                {
                    $LinkPrefix = $Library;
                    $LinkPrefix=~s/\Alib//i;
                    if($Link=~/([^"'<>]+?\/|\A)(lib|)\Q$LinkPrefix\E[_\-.]*(\d+[^\/"'<>+]+?)\Q$LinkSuffix\E(\.tar\.bz2|\.tar\.gz|\.tar\.Z|\.tgz|\.tar|\.zip|)\Z/i)
                    {
                        my $Version = $3;
                        return filter_version($Version);
                    }
                }
            }
            else
            {
                if($Link=~/([^"'<>]+?\/|\A)\Q$LinkPrefix\E([^\/"'<>+]+?)\.(tar\.bz2|tar\.gz|tar\.Z|tgz|tar|zip)\Z/i)
                {
                    my $Version = $2;
                    return filter_version($Version);
                }
            }
        }
    }
}

sub get_default_package_link_version($$)
{
    my ($Library, $Link) = @_;
    return "" if(not $Library or not $Link);
    my $LinkPrefix = $Library;
    $LinkPrefix=~s/\Alib//i;
    if($Link=~/([^"'<>]+?\/|\A)(lib|)\Q$LinkPrefix\E[_\-.]*([^\/"'<>+]+?)\.(tar\.bz2|tar\.gz|tar\.Z|tgz|tar|zip)\Z/i)
    {
        my $Version = $3;
        return filter_version($Version);
    }
    return "";
}

sub search_packages($$)
{
    my ($Library, $Page) = @_;
    return 0 if(not $Library or not $Page);
    my %MatchedLinks = ();
    my $Found = 0;
    foreach my $Link (sort {$b=~/\.bz2\Z/ <=> $a=~/\.bz2\Z/} get_links($Library, $Page))
    {
        if(my $Version = get_package_link_version($Library, $Link))
        {
            push(@{$MatchedLinks{$Version}}, $Link);
        }
    }
    foreach my $Version (sort {cmpVersions($b, $a)} keys(%MatchedLinks)) 
    {
        foreach my $Link (sort {length($b) <=> length($a)} @{$MatchedLinks{$Version}})
        {
            $Found+=download_LibraryVersion($Library, $Version, $Link, 0);
            if($Link=~s/:\d+//)
            {# remove port (:21)
                $Found+=download_LibraryVersion($Library, $Version, $Link, 0);
            }
        }
    }
    if(not $UpdateInfo and my @Ver = sort {cmpVersions($b, $a)} keys(%MatchedLinks))
    {# try to download old packages
        my $HighVer = $Ver[0];
        if($HighVer=~/[a-z]/i)
        {
            my $Index = 1;
            while($Index<=$#Ver and $HighVer=~/[a-z]/i)
            {
                $HighVer = $Ver[$Index];
                $Index+=1;
            }
        }
        if(my @Links = sort @{$MatchedLinks{$HighVer}} and $HighVer!~/[a-z-_]/i)
        {
            my $LinkTemplate = $Links[0];
            my $ReducedVer = reduce_version($HighVer);
            my $Miss = 0;
            while($Miss < $MAX_VIRTUAL_LINKS_ATTEMPTS and $ReducedVer)
            {
                my $VirtualLink = $LinkTemplate;
                last if(not $VirtualLink=~s/\Q$HighVer\E/$ReducedVer/g);
                my $ShortVer = $ReducedVer;
                $ShortVer=~s/\.0\Z//g;
                if($RepositoryProvide{$Library}{$ReducedVer}
                or $RepositoryProvide{$Library}{$ShortVer})
                {
                    $ReducedVer = reduce_version($ReducedVer);
                    $Miss=0;
                    next;
                }
                if(download_LibraryVersion($Library, $ReducedVer, $VirtualLink, 1))
                {
                    $Miss=0;
                    $Found+=1;
                }
                elsif(($VirtualLink=~s/\.tar\.bz2\Z/.tar.gz/g
                or $VirtualLink=~s/\.tar\.gz\Z/.tar.bz2/g or $VirtualLink=~s/\.zip\Z/.tar.gz/g)
                and download_LibraryVersion($Library, $ReducedVer, $VirtualLink, 1))
                {
                    $Miss=0;
                    $Found+=1;
                }
                elsif(($ShortVer ne $ReducedVer) and $VirtualLink=~s/\Q$ReducedVer\E/$ShortVer/g
                and download_LibraryVersion($Library, $ShortVer, $VirtualLink, 1))
                {
                    $Miss=0;
                    $Found+=1;
                }
                $ReducedVer = reduce_version($ReducedVer);
                $Miss+=1;
            }
        }
    }
    return $Found;
}

sub unmask_link($)
{
    my $Link = $_[0];
    return "" if(not $Link);
    $Link=~s/\.(tar\.bz2|tar\.gz|tar\.Z|tgz|tar|zip)\/download\Z/\.$1/g; # sourceforge.net
    return $Link;
}

sub download_LibraryVersion($$$$)
{
    my ($Library, $Version, $Link, $Virtual) = @_;
    return 0 if(not $Library or not $Version or not $Link);
    return 0 if($Link=~/\.(php|perl|cgi)\?/i);
    return 0 if($RepositoryProvide{$Library}{canonify_version($Version)});
    return 0 if($BadLinks{$Link});
    if(checkVersion_Pattern($Library, $Version) ne "release")
    {
        return 0 if($Version=~/\A[a-z]{2,}[\-_.]/i);
        return 0 if($Version=~/(\A|[-_\.])(32|64)bit([-_\.]|\Z)/i);
        return 0 if($Version=~/\A(-|_)/i);
        return 0 if($Version=~/\A\d\d\d\d\d\d\d\d\Z/i);# skipping daily snapshot versions
        return 0 if($Version!~/\d/i);
        return 0 if(checkVersion_Pattern($Library, $Version) eq "skip");
        return 0 if(getVersion_Type($Library, $Version) eq "unknown");
    }
    my $LastRelease = getLastVersion_Library($Library, "release");
    return 0 if(getVersion_Type($Library, $Version) ne "release" and $LastRelease and cmpVersions($Version, $LastRelease)==-1);
    my $LastVersion_SameType = getLastVersion_Library($Library, getVersion_Type($Library, $Version));
    return 0 if(getVersion_Type($Library, $Version) ne "release" and $LastVersion_SameType and cmpVersions($Version, $LastVersion_SameType)==-1);
    return 0 if($UpdateInfo and $LastRelease and cmpVersions($Version, $LastRelease)==-1);
    mkpath($REPOSITORY_PATH."/".$Library);
    my $FileName = get_FileName($Link);
    my $DownloadTo = $REPOSITORY_PATH."/".$Library."/".$FileName;
    my $GetCmd = "wget \"$Link\" --connect-timeout=5 --tries=1 --output-document=\"$DownloadTo\" 2>&1";
    my $Log = `$GetCmd`;
    if($Log=~/\[text\/html\]/)
    {# miss
        unlink($DownloadTo);
        $BadLinks{$Link} = 1;
        return 0;
    }
    elsif($?)
    {
        unlink($DownloadTo);
        print STDERR "ERROR: can't access \'$Link\'\n" if(not $Virtual);
        $BadLinks{$Link} = 1;
        return 0;
    }
    else
    {
        print "downloading $Library-".canonify_version($Version)."\n  to \'$DownloadTo\'\n";
        $RepositoryProvide{$Library}{canonify_version($Version)} = getVersion_Type($Library, $Version);
        $VersionPath{$Library}{canonify_version($Version)} = $DownloadTo;
        add_attr($Library, canonify_version($Version), "dates", getDate(time));
        if($LastRelease and cmpVersions($Version, $LastRelease)==1)
        {
            $UpdateLog{$Library}{"New"}{canonify_version($Version)} = $Link;
        }
        else
        {
            $UpdateLog{$Library}{"Added"}{canonify_version($Version)} = $Link;
        }
        return 1;
    }
}

sub reduce_version($)
{
    my $Version = $_[0];
    return "" if(not $Version);
    my ($MainVer, $MinorVer, $MicroVer) = split(/\./, $Version);
    if($MicroVer)
    {
        return $MainVer.".".$MinorVer.".".($MicroVer-1);
    }
    elsif($MinorVer)
    {
        if($MicroVer eq "")
        {# no micro version
            return $MainVer.".".($MinorVer-1);
        }
        else
        {# micro == 0
            return $MainVer.".".($MinorVer-1).".9";
        }
    }
    elsif($MainVer)
    {
        if($MinorVer eq "")
        {
            return ($MainVer-1);
        }
        else
        {
            if($MicroVer eq "")
            {# minor == 0 && no micro version
                return ($MainVer-1).".9";
            }
            else
            {# minor == 0 && micro == 0
                return ($MainVer-1).".9.0";
            }
        }
    }
    else
    {
        return "";
    }
}

sub getSiteAddr($)
{
    if($_[0] =~ /\A([a-z]+\:\/\/|)([^\/]+\.[a-z]+)(:[0-9]+|)(\/|\Z)/i)
    {
        return $2;
    }
    else
    {
        return $_[0];
    }
}

sub getSiteDir($)
{
    if($_[0] =~ /\A([a-z]+\:\/\/|)([^\/]+\.[a-z]+)((:[0-9]+|).+\Z)/i)
    {
        return $3;
    }
    return "";
}

sub getSiteProtocol($)
{
    if($_[0] =~ /\A([a-z]+\:\/\/)/i)
    {
        return $1;
    }
    return "";
}

sub linkSum($$)
{
    my ($PageUrl, $Path) = @_;
    return "" if(not $PageUrl);
    return $PageUrl if(not $Path);
    $PageUrl=~s/\?.+?\Z//g;
    if($Path=~/\A(ftp|http|https)\:/)
    {
        return $Path;
    }
    elsif($Path=~/\A\//)
    {
        return getSiteAddr($PageUrl).$Path;
    }
    elsif($PageUrl=~/\.(htm|html|cgi|php|perl)/i)
    {
        return get_Directory($PageUrl)."/".$Path;
    }
    else
    {
        return $PageUrl."/".$Path;
    }
    
}

sub get_FileName($)
{
    my $Path = $_[0];
    if($Path=~/\A(.*\/)([^\/]*)\Z/)
    {
        return $2;
    }
    else
    {
        return $Path;
    }
}

sub get_Directory($)
{
    my $Path = $_[0];
    return "" if($Path=~m*\A\./*);
    if($Path=~/\A(.*)[\/]+([^\/]*)\Z/)
    {
        return $1;
    }
    else
    {
        return "";
    }
}

sub writeFile($$)
{
    my ($Path, $Content) = @_;
    return if(not $Path);
    if(my $Dir = get_Directory($Path))
    {
        mkpath($Dir);
    }
    open (FILE, ">".$Path) || die ("can't open file \'$Path\': $!\n");
    print FILE $Content;
    close(FILE);
}

sub readFile($)
{
    my $Path = $_[0];
    return "" if(not $Path or not -f $Path);
    open (FILE, $Path);
    my $Content = join("", <FILE>);
    close(FILE);
    return toUnix($Content);
}

sub readFirstLine($)
{
    my $Path = $_[0];
    return "" if(not $Path or not -f $Path);
    open (FILE, $Path);
    my $FirstLine = <FILE>;
    close(FILE);
    return toUnix($FirstLine);
}

sub toUnix($)
{
    my $Text = $_[0];
    $Text=~s/\r//g;
    return $Text;
}

sub isCyclical($$)
{
    return (grep {$_ eq $_[1]} @{$_[0]});
}

sub get_depth($)
{
    return $Cache{"get_depth"}{$_[0]} if(defined $Cache{"get_depth"}{$_[0]});
    $Cache{"get_depth"}{$_[0]} = scalar ( ( ) = $_[0]=~/(\/|::)/g );
    return $Cache{"get_depth"}{$_[0]};
}

sub get_pages($$$)
{
    my ($Library, $TopPage, $CurrentPage) = @_;
    return if(not $Library or not $TopPage or not $CurrentPage);
    return if(get_depth($CurrentPage) - get_depth($TopPage) > $MAX_WEB_SEARCH_DEPTH);
    return @{$Cache{"get_pages"}{$CurrentPage}} if($Cache{"get_pages"}{$CurrentPage});
    return ($CurrentPage) if(isCyclical(\@RecurPage, $CurrentPage));
    push (@RecurPage, $CurrentPage);
    my (%Pages, %ResPages) = ();
    foreach my $Link (get_links($Library, $CurrentPage))
    {# filter directories
        if(($Link=~/\/\Z/ or $WebDirectory{$Link})
        and get_Directory($Link) ne getSiteAddr($CurrentPage)
        and getSiteAddr($Link) eq getSiteAddr($CurrentPage)
        and get_depth($Link)>get_depth($CurrentPage))
        {
            $Pages{$Link}=1;
        }
    }
    foreach my $Page (sort keys(%Pages))
    {
        next if($Page eq $CurrentPage);
        next if($Page=~/\/(python|list[_\-]*archive(s|)|logo(s|)|contrib(s|)|plugin(s|)|test(s|)|win\d*|windows\d*|delphi|os2|misc)\//i);
        next if($Page=~/\.OBJ/);
        $ResPages{$Page} = 1;
        foreach my $SubPage (get_pages($Library, $TopPage, $Page))
        {
            next if($SubPage eq $CurrentPage);
            $ResPages{$SubPage} = 1;
        }
    }
    pop(@RecurPage);
    @{$Cache{"get_pages"}{$CurrentPage}} = sort keys(%ResPages);
    return (sort keys(%ResPages));
}

sub parseTag($$)
{
    my ($CodeRef, $Tag) = @_;
    return "" if(not $CodeRef or not ${$CodeRef} or not $Tag);
    if(${$CodeRef} =~ s/\<$Tag\>((.|\n)+?)\<\/$Tag\>//)
    {
        my $Content = $1;
        $Content=~s/(\A\s+|\s+\Z)//g;
        return $Content;
    }
    else
    {
        return "";
    }
}

sub checkRepositoryFiles()
{
    foreach my $Path (split(/\n/, `find \"$REPOSITORY_PATH\" -maxdepth 1 -type d`))
    {
        my $Library = get_FileName($Path);
        next if($Path=~/\A$REPOSITORY_PATH([\/]+|)\Z/);
        foreach my $Dir (split(/\n/, `find \"$Path\" -maxdepth 1 -type d`))
        {
            if($Dir!~/-source\Z/
            and $Dir ne $Path)
            {
                rmtree($Dir);
            }
        }
        foreach my $FilePath (split(/\n/, `find \"$Path\" -maxdepth 1 -type f`))
        {
            if(-s $FilePath)
            {
                $RepositoryFiles{$Library}{get_FileName($FilePath)} = $FilePath;
            }
            else
            {
                unlink($FilePath);
            }
        }
    }
}

sub getDate($)
{
    my $Time = $_[0];
    my ($sec,$min,$hour,
    $mday,$mon,$year,$wday,
    $yday,$isdst) = localtime($Time);
    return ($year+1900)."-".normalize_num($mon+1)."-".normalize_num($mday)." ".normalize_num($hour).":".normalize_num($min);
}

sub getDateStamp($)
{
    my $Date = $_[0];
    return "" if(not $Date);
    $Date=~s/[ :]/_/g;
    return $Date;
}

sub normalize_num($)
{
    my $Num = $_[0];
    if(int($Num)<10)
    {
        return "0".$Num;
    }
    else
    {
        return $Num;
    }
}

sub checkRepository($)
{
    my $Library = $_[0];
    return if(not $Library);
    mkpath($REPOSITORY_PATH);
    checkRepositoryFiles() if(not keys(%RepositoryFiles));
    foreach my $FileName (keys(%{$RepositoryFiles{$Library}}))
    {
        if(my $Version = get_package_link_version($Library, $FileName))
        {
            $RepositoryProvide{$Library}{canonify_version($Version)} = getVersion_Type($Library, $Version);
            $VersionPath{$Library}{canonify_version($Version)} = $RepositoryFiles{$Library}{$FileName};
        }
    }
}

sub checkVersion_Pattern($$)
{
    my ($Library, $Version) = @_;
    return "" if(not $Library or $Version eq "");
    foreach my $Type (keys(%{$LibraryInfo{$Library}{"Versioning"}}))
    {
        foreach my $Pattern (keys(%{$LibraryInfo{$Library}{"Versioning"}{$Type}}))
        {
            if($Version=~/$Pattern/)
            {
                return $Type;
            }
        }
    }
    return "";
}

sub getVersion_Type($$)
{
    my ($Library, $Version) = @_;
    return "" if(not $Library or $Version eq "");
    if(my $Type = checkVersion_Pattern($Library, $Version))
    {
        return $Type;
    }
    if($Version=~/(\A|[^a-z])(devel|dev|exp)([^a-z]|\Z)/i)
    {
        return "devel";
    }
    elsif($Version=~/(\A|[^a-z])(snapshot|snap|master|current)([^a-z]|\Z)/i)
    {
        return "snapshot";
    }
    elsif($Version=~/(\A|[^a-z])(alfa|alpha|a)([^a-z]|\Z)/i)
    {
        return "alfa";
    }
    elsif($Version=~/(\A|[^a-z])tp([^a-z]|\Z)/i)
    {
        return "technology-preview";
    }
    elsif($Version=~/(\A|[^a-z])(beta|b)([^a-z]|\Z)/i)
    {
        return "beta";
    }
    elsif($Version=~/(\A|[^a-z])pre([^a-z]|\Z)/i)
    {
        return "pre-release";
    }
    elsif($Version=~/(\A|[^a-z])(rc|cr)([^a-z]|\Z)/i)
    {
        return "release-candidate";
    }
    elsif($Version=~/\Acurrent[0-9._:-]\Z/)
    {
        return "current";
    }
    elsif($Version=~/[a-z]/i)
    {
        return "unknown";
    }
    else
    {
        return "release";
    }
}

sub check_state()
{
    if(-f "lib_info.xml")
    {
        my $Content = readFile("lib_info.xml");
        if($Content)
        {
            $Content =~ s/\/\*(.|\n)+?\*\///g;
            $Content =~ s/<\!--(.|\n)+?-->//g;
            while(my $Library = parseTag(\$Content, "library"))
            {
                my %Attr = ();
                my $Name = parseTag(\$Library, "name");
                $Attr{"Name"} = $Name;
                $Attr{"FullName"} = parseTag(\$Library, "full_name");
                $Attr{"Summary"} = parseTag(\$Library, "summary");
                $Attr{"Description"} = parseTag(\$Library, "description");
                $Attr{"Site"} = parseTag(\$Library, "site");
                $Attr{"Bugs"} = parseTag(\$Library, "bugs");
                $Attr{"Doc"} = parseTag(\$Library, "doc");
                foreach my $Page (split(/\n/, parseTag(\$Library, "download")))
                {
                    $Page=~s/\A\s+|\s+\Z//g;
                    $Page=~s/[\/]+\Z//g;
                    $Attr{"Download"}{$Page} = 1;
                }
                $Attr{"MailingList"} = parseTag(\$Library, "mailing_list");
                $Attr{"LinkTemplate"} = parseTag(\$Library, "link_template");
                $Attr{"Scm"} = parseTag(\$Library, "scm");
                $Attr{"ScmAddr"} = parseTag(\$Library, "scm_addr");
                $Attr{"ScmModule"} = parseTag(\$Library, "scm_module");
                $Attr{"ScmBranch"} = parseTag(\$Library, "scm_branch");
                $Attr{"ScmPassword"} = parseTag(\$Library, "scm_password");
                foreach my $PackageTmpl (split(/\n/, parseTag(\$Library, "package")))
                {
                    $PackageTmpl=~s/\A\s+|\s+\Z//g;
                    $Attr{"Package"}{$PackageTmpl} = 1;
                }
                my $Pos = 0;
                foreach my $Header (split(/\n/, parseTag(\$Library, "include_preamble")))
                {
                    $Header=~s/\A\s+|\s+\Z//g;
                    $Attr{"IncludePreamble"}{$Header} = $Pos;
                    $Pos+=1;
                }
                foreach my $Header (split(/\n/, parseTag(\$Library, "headers")))
                {
                    $Header=~s/\A\s+|\s+\Z//g;
                    $Attr{"Headers"}{$Header} = $Pos;
                    $Pos+=1;
                }
                foreach my $Header (split(/\n/, parseTag(\$Library, "libs")))
                {
                    $Header=~s/\A\s+|\s+\Z//g;
                    $Attr{"Libs"}{$Header} = 1;
                }
                foreach my $IncPath (split(/\n/, parseTag(\$Library, "include_paths")))
                {
                    $IncPath=~s/\A\s+|\s+\Z//g;
                    $Attr{"IncludePaths"}{$IncPath} = 1;
                }
                foreach my $Name (split(/\n/, parseTag(\$Library, "skip_headers")))
                {
                    $Name=~s/\A\s+|\s+\Z//g;
                    $Attr{"SkipHeaders"}{$Name} = 1;
                }
                foreach my $Name (split(/\n/, parseTag(\$Library, "skip_libs")))
                {
                    $Name=~s/\A\s+|\s+\Z//g;
                    $Attr{"SkipLibs"}{$Name} = 1;
                }
                foreach my $AddAction (split(/\n/, parseTag(\$Library, "actions")))
                {
                    $AddAction=~s/\A\s+|\s+\Z//g;
                    $Attr{"Actions"}{$AddAction} = 1;
                }
                foreach my $SkipAction (split(/\n/, parseTag(\$Library, "skip_on_update")))
                {
                    $SkipAction=~s/\A\s+|\s+\Z//g;
                    $Attr{"SkipOnUpdate"}{$SkipAction} = 1;
                }
                foreach my $Lib (split(/\n/, parseTag(\$Library, "libs_depend")))
                {
                    $Lib=~s/\A\s+|\s+\Z//g;
                    $Attr{"LibsDepend"}{$Lib} = 1;
                }
                foreach my $Interface (split(/\n/, parseTag(\$Library, "skip_interfaces")))
                {
                    $Interface=~s/\A\s+|\s+\Z//g;
                    $Attr{"SkipInterfaces"}{$Interface} = 1;
                }
                foreach my $LinkTmpl (split(/\n/, parseTag(\$Library, "changelog_links")))
                {
                    $LinkTmpl=~s/\A\s+|\s+\Z//g;
                    $Attr{"ChangeLogLinks"}{$LinkTmpl} = 1;
                }
                foreach my $Interface (split(/\n/, parseTag(\$Library, "gcc_options")))
                {
                    $Interface=~s/\A\s+|\s+\Z//g;
                    $Attr{"GccOptions"}{$Interface} = 1;
                }
                foreach my $IntParam (split(/\n/, parseTag(\$Library, "out_params")))
                {
                    $IntParam=~s/\A\s+|\s+\Z//g;
                    $Attr{"OutParams"}{$IntParam} = 1;
                }
                foreach my $Warning (split(/\n/, parseTag(\$Library, "skip_warnings")))
                {
                    $Warning=~s/\A\s+|\s+\Z//g;
                    $Attr{"SkipWarnings"}{$Warning} = 1;
                }
                foreach my $Constant (split(/\n/, parseTag(\$Library, "skip_constants")))
                {
                    $Constant=~s/\A\s+|\s+\Z//g;
                    $Attr{"SkipConstants"}{$Constant} = 1;
                }
                $Attr{"RunMode"} = parseTag(\$Library, "run_mode");
                $Attr{"InstallMode"} = parseTag(\$Library, "install_mode");
                foreach my $ConfOpt (split(/\n/, parseTag(\$Library, "configure_options")))
                {
                    $ConfOpt=~s/\A\s+|\s+\Z//g;
                    $Attr{"ConfigureOptions"}{$ConfOpt} = 1;
                }
                foreach my $Rule (split(/\n/, parseTag(\$Library, "versioning")))
                {
                    $Rule=~s/\A\s+|\s+\Z//g;
                    if($Rule=~/(\w+?):(.+)\Z/)
                    {
                        my ($Type, $Pattern) = ($1, $2);
                        $Attr{"Versioning"}{$Type}{$Pattern} = 1;
                    }
                }
                my $RuleNum = 0;
                foreach my $Rule (split(/\n/, parseTag(\$Library, "pre_install")))
                {
                    $Rule=~s/\A\s+|\s+\Z//g;
                    if($Rule=~/\A([^\s=&]+):(.+)\Z/)
                    {
                        $Attr{"PreInstall"}{$1}{$2} = $RuleNum;
                    }
                    else
                    {
                        $Attr{"PreInstallGeneric"}{$Rule} = $RuleNum;
                    }
                    $RuleNum+=1;
                }
                foreach my $Rule (split(/\n/, parseTag(\$Library, "post_install")))
                {
                    $Rule=~s/\A\s+|\s+\Z//g;
                    if($Rule=~/\A([^\s=&]+):(.+)\Z/)
                    {
                        $Attr{"PostInstall"}{$1}{$2} = $RuleNum;
                    }
                    else
                    {
                        $Attr{"PostInstallGeneric"}{$Rule} = $RuleNum;
                    }
                    $RuleNum+=1;
                }
                foreach my $Rule (split(/\n/, parseTag(\$Library, "build")))
                {
                    $Rule=~s/\A\s+|\s+\Z//g;
                    if($Rule=~/\A([^\s=&]+):(.+)\Z/)
                    {
                        $Attr{"Build"}{$1}{$2} = $RuleNum;
                    }
                    else
                    {
                        $Attr{"BuildGeneric"}{$Rule} = $RuleNum;
                    }
                    $RuleNum+=1;
                }
                foreach my $Rule (split(/\n/, parseTag(\$Library, "install")))
                {
                    $Rule=~s/\A\s+|\s+\Z//g;
                    if($Rule=~/\A([^\s=&]+):(.+)\Z/)
                    {
                        $Attr{"Install"}{$1}{$2} = $RuleNum;
                    }
                    else
                    {
                        $Attr{"InstallGeneric"}{$Rule} = $RuleNum;
                    }
                    $RuleNum+=1;
                }
                foreach my $Order (split(/\n/, parseTag(\$Library, "include_order")))
                {
                    $Order=~s/\A\s+|\s+\Z//g;
                    $Attr{"IncludeOrder"}{$Order} = 1;
                }
                %{$LibraryInfo{$Name}} = %Attr;
            }
        }
        else
        {
            print STDERR "ERROR: lib_info.xml is empty\n";
        }
    }
    else
    {
        print STDERR "ERROR: can't find lib_info.xml\n";
    }
    foreach my $Library (sort keys(%LibraryInfo))
    {
        checkRepository($Library);
    }
}

sub skip_lib($$)
{
    my ($Library, $SoPath) = @_;
    return 1 if(not $Library or not $SoPath);
    foreach my $Pattern (keys(%{$LibraryInfo{$Library}{"SkipLibs"}}))
    {
        if($Pattern=~/\*|\//)
        {
            $Pattern=~s/\*/.*/g;
            return 1 if($SoPath=~/$Pattern/);
        }
        else
        {
            return 1 if(get_FileName($SoPath)=~/\A$Pattern[0-9\.]*\Z/);
        }
    }
    return 0;
}

sub getLastVersion_Library($$)
{
    my ($Library, $Type) = @_;
    my $LastVersion = "";
    foreach my $Version (sort {cmpVersions($a, $b)} keys(%{$RepositoryProvide{$Library}}))
    {
        if($RepositoryProvide{$Library}{$Version} eq $Type)
        {
            $LastVersion = $Version;
        }
    }
    return $LastVersion;
}

sub isInstalled($$)
{
    my ($Library, $Version) = @_;
    return 0 if(not $Library or not $Version);
    my $InstallPath = "$INSTALL_PATH/$Library/$Version";
    if(-d $InstallPath)
    {
        foreach my $Path (cmd_find($InstallPath,"f","*.so*"))
        {
            if(get_FileName($Path)=~/\.so[0-9.]*\Z/)
            {
                return 1;
            }
        }
    }
    return 0;
}

sub clearUnpacked($$)
{
    my ($Library, $Version) = @_;
    rmtree("$REPOSITORY_PATH/$Library/install/");
}

sub get_unpack_command($$)
{
    my ($Package, $OutDir) = @_;
    if($Package=~/\.(tar\.gz|tgz|tar\.Z)\Z/)
    {
        return "tar -xzf $Package --directory=$OutDir";
    }
    elsif($Package=~/\.tar\.bz2\Z/)
    {
        return "tar -xjf $Package --directory=$OutDir";
    }
    elsif($Package=~/\.tar\Z/)
    {
        return "tar -xf $Package --directory=$OutDir";
    }
    elsif($Package=~/\.zip\Z/)
    {
        return "unzip $Package -d $OutDir";
    }
    else
    {
        return "";
    }
}

sub unPack($$)
{
    my ($Library, $Version) = @_;
    clearUnpacked($Library, $Version);
    return "" if(not $VersionPath{$Library}{$Version});
    my $Package = get_FileName($VersionPath{$Library}{$Version});
    if(my $UnpackCmd = get_unpack_command($Package, "install"))
    {
        my $BuildDir = "$REPOSITORY_PATH/$Library/install";
        mkpath($BuildDir);
        system("cd \"$REPOSITORY_PATH/$Library/\" && $UnpackCmd >/dev/null 2>&1");
        return $BuildDir;
    }
    else
    {
        print STDERR "ERROR: unknown format $Package\n";
        return "";
    }
}

sub get_RootDir($)
{
    my $Dir = $_[0];
    rmtree($Dir."/CVS") if(-d $Dir."/CVS");
    my @Root_Files = split(/\n/, `find \"$Dir\" -maxdepth 1`);
    if($#Root_Files==1)
    {
        return get_RootDir($Root_Files[1]);
    }
    else
    {
        return $Dir;
    }
}


sub esc($)
{
    my $Str = $_[0];
    $Str =~ s/([()\[\]{}$ &'"`;,<>])/\\$1/g;
    return $Str;
}

sub cmd_find($$$)
{
    my ($Path, $Type, $Name) = @_;
    return () if(not $Path or not -e $Path);
    my $Cmd = "find ".esc($Path);
    if($Type)
    {
        $Cmd .= " -type $Type";
    }
    if($Name)
    {
        $Cmd .= " -name \"$Name\"";
    }
    return split(/\n/, `$Cmd`);
}

sub get_prefix_defines($$)
{
    my ($Path, $InstallPath) = @_;
    my $Content = readFile($Path);
    my %Defines = ();
    while($Content=~s/\$\((\w+)\)(\/|)(\$\((\w+)\)\/|)(include|lib|bin|man)//i)
    {
        my $Define = $1;
        $Defines{$Define} = 1 if($Define!~/src|man|build|top|home|package|test/i);
    }
    if($Content=~/\$\(DESTDIR\)(\s|\/)/)
    {# Xen
        $Defines{"DESTDIR"} = 1;
    }
    if($Defines{"INSTALL_PREFIX"})
    {
        return "INSTALL_PREFIX=$InstallPath ";
    }
    elsif($Defines{"DESTDIR"})
    {
        return "DESTDIR=$InstallPath ";
    }
    else
    {
        my $DefineString = "";
        foreach my $Define (keys(%Defines))
        {
            $DefineString .= $Define."=".$InstallPath." ";
        }
        return $DefineString;
    }
}

sub check_for_splint_specs($)
{
    my $Path = $_[0];
    return 0 if(not $Path or not -e $Path);
    my $Content = readFile($Path);
    return ($Content=~/\/\*\s*\@\s*(returned|out|notnull|null)\s*\@\s*\*\//
    or $Content=~/\/\*\s*\@\s*(modifies|requires|ensures)\s*(.+?)\s*\@\s*\*\//);
}

sub sort_unique_paths(@)
{
    my @Paths = @_;
    my (@Filtered, %Complex) = ();
    foreach my $Path1 (sort {length($b)<=>length($a)} @_)
    {
        $Path1 =~s/[\/]+\Z//g;
        foreach my $Path2 (sort {length($a)<=>length($b)} @_)
        {
            $Path2 =~s/[\/]+\Z//g;
            if($Path1=~/\A\Q$Path2\E\//)
            {
                $Complex{$Path1}=1;
                last;
            }
        }
    }
    foreach my $Path (sort {length($a)<=>length($b)} @_)
    {
        if(not $Complex{$Path})
        {
            push(@Filtered, $Path);
        }
    }
    return @Filtered;
}

sub getPkgConfigLibs($$)
{
    my ($Library, $Version) = @_;
    return "" if(not $Version or not $Library);
    return @{$Cache{"getPkgConfigLibs"}{$Library}{$Version}} if(defined $Cache{"getPkgConfigLibs"}{$Library}{$Version});
    my $InstallPath = "$INSTALL_PATH/$Library/$Version";
    my @Names = ();
    foreach my $PkgCfgDir (cmd_find($InstallPath,"d","pkgconfig"))
    {
        foreach my $Path (cmd_find($PkgCfgDir,"f",""))
        {
            my $Content = readFile($Path);
            foreach my $Line (split(/\n/, $Content))
            {
                if($Line=~/\A(\w+)\s*=\s*(.*)\Z/)
                {
                    my ($Define, $Value) = ($1, $2);
                    $Content=~s/\${$Define}/$Value/g;
                }
            }
            while($Content=~s/\s+-l([\w-.]+)//)
            {
                push(@Names, "lib".$1.".so");
            }
        }
    }
    $Cache{"getPkgConfigLibs"}{$Library}{$Version} = \@Names;
    return @Names;
}

sub createDescriptor($$)
{
    my ($Library, $Version) = @_;
    return "" if(not $Version or not $Library);
    my $InstallPath = "$INSTALL_PATH/$Library/$Version";
    my (%LibDirs, %Libs, %Includes, @Headers, %AddIncludePaths) = ();
    if(not keys(%{$LibraryInfo{$Library}{"Libs"}}))
    {
        foreach my $Name (getPkgConfigLibs($Library, $Version))
        {
            $LibraryInfo{$Library}{"Libs"}{$Name}=1;
        }
    }
    if(my @Files = keys(%{$LibraryInfo{$Library}{"Libs"}}))
    {
        foreach my $Name (@Files)
        {
            my @Paths = cmd_find($InstallPath,"",$Name);
            my @Paths = cmd_find($InstallPath,"f",$Name."*") if(not @Paths);
            my @Paths = cmd_find($InstallPath,"",$Name."*") if(not @Paths);
            foreach my $Path (@Paths)
            {
                $Libs{cut_path_prefix($Path, $InstallPath)}=1;
            }
        }
    }
    if(not keys(%Libs))
    {
        if(-d $InstallPath."/lib")
        {
            $LibDirs{"lib"}=1;
        }
        else
        {
            foreach my $Path (cmd_find($InstallPath,"","*\.so*"))
            {
                next if($Path!~/\.so([0-9\.]*)\Z/);
                my $RelPath = cut_path_prefix($Path, $InstallPath);
                next if($RelPath=~/(\A|\/)(examples|plugins)(\/|\Z)/i);
                $LibDirs{get_Directory($RelPath)}=1;
            }
        }
    }
    my ($HasAnnotations, $OneHeaderChecked) = (0, 0);
    if(my @Files = sort {$LibraryInfo{$Library}{"Headers"}{$a}<=>$LibraryInfo{$Library}{"Headers"}{$b}}
    keys(%{$LibraryInfo{$Library}{"Headers"}}))
    {
        foreach my $Name (@Files)
        {
            if($Name=~s/\A\{RELPATH\}//g)
            {
                push(@Headers, $Name);
                next;
            }
            foreach my $Path (cmd_find($InstallPath,"f",$Name))
            {
                my $RelPath = cut_path_prefix($Path, $InstallPath);
                next if($RelPath=~/(\A|\/)(pkgconfig|examples)(\/|\Z)/i);
                push(@Headers, $RelPath);
                if(not $OneHeaderChecked)
                {
                    if(check_for_splint_specs($Path))
                    {
                        $LibraryInfo{$Library}{"Annotations"}=1;
                    }
                    $OneHeaderChecked=1;
                }
            }
        }
        foreach my $Path (cmd_find($InstallPath,"f","*\.h"))
        {
            my $RelPath = cut_path_prefix($Path, $InstallPath);
            if($RelPath=~/\Alib(|s)\//i)
            {
                $AddIncludePaths{get_Directory($RelPath)}=1;
                if(my $Dir = get_Directory(get_Directory($RelPath)))
                {
                    $AddIncludePaths{$Dir}=1;
                }
            }
        }
    }
    else
    {
        foreach my $Path (cmd_find($InstallPath,"f","*\.h"))
        {
            $Includes{get_Directory(cut_path_prefix($Path, $InstallPath))}=1;
            if(not $OneHeaderChecked)
            {
                if(check_for_splint_specs($Path))
                {
                    $LibraryInfo{$Library}{"Annotations"}=1;
                }
                $OneHeaderChecked=1;
            }
        }
        foreach my $Path (cmd_find($InstallPath,"d","include"))
        {
            $Includes{cut_path_prefix($Path, $InstallPath)}=1;
        }
        if(-d $InstallPath."/include")
        {
            foreach my $Path (keys(%Includes))
            {
                if($Path!~/\A(lib|include)(|\/)/)
                {
                    delete($Includes{$Path});
                }
            }
        }
    }
    return "" if((not keys(%Includes) and $#Headers<0) or (not keys(%LibDirs) and not keys(%Libs)));
    my $DPath = "descriptors/$Library/$Version.xml";
    my $Descriptor = "<version>
    $Version
  </version>

  <headers>
    {RELPATH}/".(keys(%{$LibraryInfo{$Library}{"Headers"}})?join("\n    {RELPATH}/",@Headers):join("\n    {RELPATH}/",sort_unique_paths(keys(%Includes))))."
  </headers>

  <libs>
    {RELPATH}/".(keys(%{$LibraryInfo{$Library}{"Libs"}})?join("\n    {RELPATH}/",keys(%Libs)):join("\n    {RELPATH}/",sort_unique_paths(keys(%LibDirs))))."
  </libs>";
    if(my @IncludePreamble = sort {$LibraryInfo{$Library}{"IncludePreamble"}{$a}<=>$LibraryInfo{$Library}{"IncludePreamble"}{$b}} keys(%{$LibraryInfo{$Library}{"IncludePreamble"}}))
    {
        $Descriptor .= "\n\n  <include_preamble>\n    ".join("\n    ",@IncludePreamble)."\n  </include_preamble>";
    }
    if(my @IncludePaths = keys(%{$LibraryInfo{$Library}{"IncludePaths"}}))
    {
        $Descriptor .= "\n\n  <include_paths>\n    ".join("\n    ",@IncludePaths)."\n  </include_paths>";
    }
    if(my @AddIncPaths = keys(%AddIncludePaths))
    {
        $Descriptor .= "\n\n  <add_include_paths>\n    {RELPATH}/".join("\n    {RELPATH}/",@AddIncPaths)."\n  </add_include_paths>";
    }
    if(my @SkipHeaders = keys(%{$LibraryInfo{$Library}{"SkipHeaders"}}))
    {
        $Descriptor .= "\n\n  <skip_headers>\n    ".join("\n    ",@SkipHeaders)."\n  </skip_headers>";
    }
    if(my @SkipLibs = keys(%{$LibraryInfo{$Library}{"SkipLibs"}}))
    {
        $Descriptor .= "\n\n  <skip_libs>\n    ".join("\n    ",@SkipLibs)."\n  </skip_libs>";
    }
    if(my @IncludeOrder = keys(%{$LibraryInfo{$Library}{"IncludeOrder"}}))
    {
        $Descriptor .= "\n\n  <include_order>\n    ".join("\n    ",@IncludeOrder)."\n  </include_order>";
    }
    if(my @LibsDepend = keys(%{$LibraryInfo{$Library}{"LibsDepend"}}))
    {
        $Descriptor .= "\n\n  <libs_depend>\n    ".join("\n    ",@LibsDepend)."\n  </libs_depend>";
    }
    if(my @SkipInterfaces = keys(%{$LibraryInfo{$Library}{"SkipInterfaces"}}))
    {
        $Descriptor .= "\n\n  <skip_interfaces>\n    ".join("\n    ",@SkipInterfaces)."\n  </skip_interfaces>";
    }
    if(my @GccOptions = keys(%{$LibraryInfo{$Library}{"GccOptions"}}))
    {
        $Descriptor .= "\n\n  <gcc_options>\n    ".join("\n    ",@GccOptions)."\n  </gcc_options>";
    }
    if(my @OutParams = keys(%{$LibraryInfo{$Library}{"OutParams"}}))
    {
        $Descriptor .= "\n\n  <out_params>\n    ".join("\n    ",@OutParams)."\n  </out_params>";
    }
    if(my @SkipWarnings = keys(%{$LibraryInfo{$Library}{"SkipWarnings"}}))
    {
        $Descriptor .= "\n\n  <skip_warnings>\n    ".join("\n    ",@SkipWarnings)."\n  </skip_warnings>";
    }
    if(my @SkipConstants = keys(%{$LibraryInfo{$Library}{"SkipConstants"}}))
    {
        $Descriptor .= "\n\n  <skip_constants>\n    ".join("\n    ",@SkipConstants)."\n  </skip_constants>";
    }
    writeFile($DPath, "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n<descriptor>\n  $Descriptor\n</descriptor>");
    return $DPath;
}

sub test_LibraryVersion($$$)
{
    my ($Library, $Version, $TestFormat) = @_;
    return if(not $Library or not $Version or not $TestFormat);
    if($TestFormat eq "standalone")
    {
        if(-e "test_results/$Library/$Version/test_results.html"
        and not $Alter)
        {
            print "this version of library was already tested (use -alter option for testing again)\n";
            return;
        }
    }
    elsif($TestFormat eq "t2c")
    {
        if(-e "tests_t2c/$Library/$Version/t2c-$Library-tests-$Version.tar.gz"
        and not $Alter)
        {
            print "tests in T2C format were already generated (use -alter option for generating again)\n";
            return;
        }
    }
    if(not isInstalled($Library, $Version))
    {
        print "$Library $Version".($UseDebugLib?" (debug)":"")." was not installed yet\n";
        return;
    }
    if(my $DPath = createDescriptor($Library, $Version))
    {
        my $Cmd = "api-sanity-autotest.pl -l $Library -d $DPath";
        my $InstallPath = "$INSTALL_PATH/$Library/$Version";
        $Cmd .= " -relpath ".$InstallPath;
        if(my $FullName = $LibraryInfo{$Library}{"FullName"})
        {
            $Cmd .= " -library-full-name ".$FullName
        }
        if($LibraryInfo{$Library}{"RunMode"} eq "xvfb")
        {
            $Cmd .= " -xvfb";
        }
        if($TestFormat eq "standalone")
        {
            $Cmd .= " -show-retval -gen -build -run";
        }
        elsif($TestFormat eq "t2c")
        {
            $Cmd .= " -t2c -gen";
        }
        if($LibraryInfo{$Library}{"Annotations"})
        {
            $Cmd .= " -splint-specs";
        }
        if(-f "spectypes/$Library-$Version.st")
        {
            $Cmd .= " -st spectypes/$Library-$Version.st";
        }
        else
        {
            my ($HVer, $LVer) = ("", "");
            my (@Higher, @Lower) = ();
            foreach my $STpath (cmd_find("spectypes","f","*\.st"))
            {
                my $Name = get_FileName($STpath);
                if($Name=~/$Library>=(.+)\.st/)
                {
                    push(@Higher, $1);
                }
                elsif($Name=~/$Library<=(.+)\.st/)
                {
                    push(@Lower, $1);
                }
            }
            foreach my $STVer (sort {cmpVersions($a, $b)} @Lower)
            {
                if(cmpVersions($Version, $STVer)!=1)
                {
                    $LVer = $STVer;
                }
            }
            foreach my $STVer (sort {cmpVersions($a, $b)} @Higher)
            {
                if(cmpVersions($Version, $STVer)!=-1)
                {
                    $HVer = $STVer;
                }
            }
            if($HVer)
            {
                $Cmd .= " -st \"spectypes/$Library>=$HVer.st\"";
            }
            elsif($LVer)
            {
                $Cmd .= " -st \"spectypes/$Library<=$LVer.st\"";
            }
            elsif(-f "spectypes/$Library.st")
            {
                $Cmd .= " -st spectypes/$Library.st";
            }
        }
        if(-f "parameters/$Library.prms")
        {
            $Cmd .= " -params parameters/$Library.prms";
        }
        if(-d "testdata/$Library-$Version")
        {
            $Cmd .= " -td testdata/$Library-$Version";
        }
        elsif(-d "testdata/$Library")
        {
            $Cmd .= " -td testdata/$Library";
        }
        print "running perl $Cmd\n";
        system("perl $Cmd");
        if($TestFormat eq "standalone")
        {# cleaning
            my $CleanCmd = "api-sanity-autotest.pl -l $Library -d $DPath -view-only";
            print "running perl $CleanCmd\n";
            system("perl $CleanCmd");
        }
        elsif($TestFormat eq "t2c")
        {# packaging and cleaning
            my $TestsVersion = `perl api-sanity-autotest.pl --dumpversion`;
            chomp($TestsVersion);
            my $TestsPath = "tests_t2c/$Library/$Version";
            if(-d $TestsPath."/$Library-t2c")
            {
                writeFile("$TestsPath/$Library-t2c/VERSION", $TestsVersion);
                print "creating package with t2c tests: 'tests/$Library/$Version/t2c-$Library-tests-$Version.tar.gz'\n";
                create_Package("$TestsPath/$Library-t2c", "t2c-$Library-tests-$Version", "tar.gz");
                rmtree("$TestsPath/$Library-t2c");
            }
            else
            {
                print STDERR "ERROR: can't create package with t2c tests\n";
                rmtree($TestsPath);
            }
        }
        if($TestFormat eq "standalone")
        {
            my %ResAttr = readAttributes("test_results/$Library/$Version/test_results.html");
            foreach my $Name (keys(%ResAttr))
            {
                $UpdateLog{$Library}{"TestRes"}{$Version}{$Name}=$ResAttr{$Name};
            }
            my %ViewAttr = readAttributes("tests/$Library/$Version/view_tests.html");
            foreach my $Name (keys(%ViewAttr))
            {
                $UpdateLog{$Library}{"Tests"}{$Version}{$Name}=$ViewAttr{$Name};
            }
        }
        elsif($TestFormat eq "t2c")
        {
            # FIXME: add log for T2C tests
        }
        if($Version eq "current")
        {
            add_attr($Library, "current", "dates", readFile("$INSTALL_PATH/$Library/current/date.txt"));
        }
    }
    else
    {
        print STDERR "ERROR: can't create descriptor for $Library $Version\n";
    }
}

sub mkcpp_LibraryVersion($$)
{# renaming "this" and "class" keywords to "c_this" and "c_class"
    my ($Library, $Version) = @_;
    return if(not $Library or not $Version);
    print "making $Library $Version headers C++ compatible\n";
    foreach my $Path (cmd_find("$INSTALL_PATH/$Library/$Version","f","*\.h"))
    {
        my $Content = readFile($Path);
        $Content=~s/(\*\s*|\s+)(this|class|export|private|public|protected)(\s*(,|\)|;))/$1c_$2$3/g;
        writeFile($Path, $Content);
    }
}

sub mkcpp_Library($)
{
    my $Library = $_[0];
    return if(not $Library);
    my @Versions = sort_VersionSequence(keys(%{$RepositoryProvide{$Library}}), $Library);
    sort_Installed($Library, \@Versions);
    if($#Versions<0)
    {
        print STDERR "ERROR: $Library versions".($UseDebugLib?" (debug)":"")." were not installed yet\n";
        return;
    }
    foreach my $Version (@Versions)
    {
        mkcpp_LibraryVersion($Library, $Version);
    }
}

sub cmp_LibraryVersions($$$)
{
    my ($Library, $Version1, $Version2) = @_;
    return if(not $Library or not $Version1 or not $Version2);
    if(-e "compat_reports/$Library/$Version1"."_to_$Version2/abi_compat_report.html"
    and not $Alter)
    {
        print "versions $Version1 and $Version2 of the ABI were already compared (use -alter option for comparing again)\n";
        return;
    }
    if(not isInstalled($Library, $Version1))
    {
        print STDERR "ERROR: $Library $Version1".($UseDebugLib?" (debug)":"")." was not installed yet\n";
        return;
    }
    if(not isInstalled($Library, $Version2))
    {
        print STDERR "ERROR: $Library $Version2".($UseDebugLib?" (debug)":"")." was not installed yet\n";
        return;
    }
    my $DPath1 = createDescriptor($Library, $Version1);
    if(not $DPath1)
    {
        print STDERR "ERROR: can't create descriptor for $Library $Version1\n";
        return;
    }
    my $DPath2 = createDescriptor($Library, $Version2);
    if(not $DPath2)
    {
        print STDERR "ERROR: can't create descriptor for $Library $Version2\n";
        return;
    }
    my $Cmd = "abi-compliance-checker.pl -l $Library";
    if(my $FullName = $LibraryInfo{$Library}{"FullName"})
    {
        $Cmd .= " -library_full_name ".$FullName
    }
    $Cmd .= " -d1 $DPath1 -d2 $DPath2";
    my $InstallPath1 = "$INSTALL_PATH/$Library/$Version1";
    my $InstallPath2 = "$INSTALL_PATH/$Library/$Version2";
    $Cmd .= " -relpath1 $InstallPath1 -relpath2 $InstallPath2";
    if(-f "parameters/$Library.prms")
    {
        $Cmd .= " -params parameters/$Library.prms";
    }
    print "running perl $Cmd\n";
    system("perl $Cmd");
    my %Attr = readAttributes("compat_reports/$Library/$Version1"."_to_$Version2/abi_compat_report.html");
    foreach my $Name (keys(%Attr))
    {
        $UpdateLog{$Library}{"Compat"}{$Version1}{$Version2}{$Name}=$Attr{$Name};
    }
    if($Version2 eq "current")
    {
        add_attr($Library, "current", "dates", readFile("$INSTALL_PATH/$Library/current/date.txt"));
    }
}

sub get_PreviousVersion($$)
{
    my ($Library, $Version) = @_;
    return "" if(not $Library or not $Version);
    my @RepVersions = sort {cmpVersions($a, $b)} keys(%{$RepositoryProvide{$Library}});
    foreach my $Num (0 .. $#RepVersions)
    {
        if($RepVersions[$Num] eq $Version)
        {
            if($Num>=1)
            {
                return $RepVersions[$Num-1];
            }
            else
            {
                return "";
            }
        }
    }
    return "";
}

sub remove_local_links($)
{
    my $Content = $_[0];
    return if(not $Content);
    while($Content=~s/<img\s+[^>]*src\s*=\s*("|')[^:'"]*("|')[^>]*>//g){};#images
    while($Content=~s/<a\s+[^>]*href\s*=\s*("|')[^:'"]*("|')[^>]*>([^>]*)<\/a>/$3/g){};#refs
    return $Content;
}

sub changelog_LibraryVersion($$)
{
    my ($Library, $Version) = @_;
    return if(not $Version or not $Library);
    if(not $RepositoryProvide{$Library}{$Version})
    {
        print STDERR "ERROR: $Library-$Version is not presented in the repository\n";
        return;
    }
    if($Version eq "current"
    and my $ScmLog = getChangeLog_SCM($Library))
    {
        if(length($ScmLog)>100000) {
            $ScmLog=substr($ScmLog,0,100000)." ...";
        }
        writeFile("changelogs/$Library/current/changelog.html", convert_to_html($Library, "current", $ScmLog));
        return;
    }
    my $ChangelogDir = "changelogs/$Library/$Version";
    mkpath($ChangelogDir);
    if(-f "changelogs/$Library/$Version/changelog.html" or readFile("changelogs/$Library/$Version/link"))
    {
        if($Alter or $Version eq "current")
        {
            unlink("changelogs/$Library/$Version/changelog.html");
            unlink("changelogs/$Library/$Version/link");
        }
        else
        {
            print "changelog for $Library-$Version was already found (use -alter option for searching again)\n";
            return;
        }
    }
    print "searching for changelog in $Library $Version ... ";
    # searching on the web site
    if(my @Links = sort {$b=~/{VERSION}|{V}/} keys(%{$LibraryInfo{$Library}{"ChangeLogLinks"}}))
    {
        foreach my $LinkTmpl (@Links)
        {
            my $PreviousVersion = get_PreviousVersion($Library, $Version);
            $LinkTmpl=~s/[\/]+\Z//g;
            $LinkTmpl=~s/{V1}/$PreviousVersion/g;
            $LinkTmpl=~s/{V2}|{VERSION}|{V}/$Version/g;
            next if($BadLinks{$LinkTmpl});
            if($Cache{"changelog_LibraryVersion.check_page"}{$LinkTmpl}
            or check_page($LinkTmpl))
            {
                $BadLinks{$LinkTmpl}=0;
                $Cache{"changelog_LibraryVersion.check_page"}{$LinkTmpl}=1;
                writeFile($ChangelogDir."/link", $LinkTmpl);
                print $LinkTmpl."\n";
                return;
            }
            else
            {
                $BadLinks{$LinkTmpl}=1;
            }
        }
    }
    # searching in the package
    my $UnpackDir = unPack($Library, $Version);
    return if(not $UnpackDir or not -d $UnpackDir);
    if(my $Path = find_changelog($Library, $Version, get_RootDir($UnpackDir)))
    {
        my $Content = readFile($Path);
        if(cmd_file($Path)!~/HTML/i and $Path!~/\.(html|htm)\Z/)
        {
            $Content = convert_to_html($Library, $Version, $Content);
        }
        else
        {
            $Content = remove_local_links($Content);
        }
        writeFile("$ChangelogDir/changelog.html", $Content);
        my $RelPath = cut_path_prefix($Path, $UnpackDir);
        $RelPath=~s/\A[^\/]+?\///;
        print $RelPath."\n";
    }
    else
    {
        print "not found\n";
    }
    clearUnpacked($Library, $Version);
}

sub cmd_file($)
{
    my $Path = $_[0];
    return "" if(not $Path or not -e $Path);
    if(my $CmdPath = get_CmdPath("file"))
    {
        my $Cmd = $CmdPath." ".esc($Path);
        my $Cmd_Out = `$Cmd`;
        return $Cmd_Out;
    }
    else
    {
        return "";
    }
}

sub convert_to_html($$$)
{
    my ($Library, $Version, $Content) = @_;
    return "" if(not $Library or not $Version or not $Content);
    $Content = htmlSpecChars($Content);
    $Content=~s/\n/<br\/>/;
    my $Title = "Changelog for the library $Library-$Version";
    my $Keywords = "$Library, changelog, shared, library, $Version";
    $Content = composeHTML_Head($Title, $Keywords, "   body{font-family:Arial;font-size:14px;background:#ffffff;color:Black;}\ntable.changelog{cursor:text;margin-top:7px;width:50%;margin-left:20px;font-family:Monaco, \"Courier New\", Courier;font-size:14px;padding:10px;border:1px solid #e0e8e5;color:#444444;background-color:#eff3f2;overflow:auto;}")."\n<body>\n".back_link($Library)."\n<table class='changelog'>\n<tr><td>\n$Content</td></tr>\n</table>\n</body>\n</html>\n";
    return $Content;
}

sub htmlSpecChars($)
{
    my $Str = $_[0];
    $Str=~s/\&([^#])/&amp;$1/g;
    $Str=~s/</&lt;/g;
    $Str=~s/>/&gt;/g;
    $Str=~s/([^ ])( )([^ ])/$1\@ALONE_SP\@$3/g;
    $Str=~s/ /&nbsp;/g;
    $Str=~s/\@ALONE_SP\@/ /g;
    $Str=~s/\n/<br\/>/g;
    return $Str;
}

sub back_link($)
{
    my $Library = $_[0];
    return "<a style='font-size:11px;color:#336699;' href=\'../../../versions/$Library.html\'>list of versions</a>";
}

sub getInitialVersion($)
{
    my $Library = $_[0];
    return "" if(not $Library);
    my @Versions = sort {cmpVersions($a, $b)} keys(%{$RepositoryProvide{$Library}});
    if($#Versions>=0)
    {
        return $Versions[0];
    }
    else
    {
        return "";
    }
}

sub get_short_version($)
{
    my $Version = $_[0];
    $Version=~s/[._-][^._-]+\Z//g;
    return $Version;
}

sub find_changelog($$$)
{
    my ($Library, $Version, $Directory) = @_;
    return "" if(not $Version or not $Library or not $Directory);
    my $ShortVersion = get_short_version($Version);
    my (@Revision, @VRelNotes, @WhatsNew, @Changes, @ChangeLog, @News, @VersionInfo, @Readme, @Log, @RelNotes, @History) = ();
    my @LogDirs = split(/\n/,`find \"$Directory\" -maxdepth 1 -type d`);
    if(my $DocDir = (-d "$Directory/docs"?"$Directory/docs":-d "$Directory/doc"?"$Directory/doc":""))
    {
        foreach my $SubDir (split(/\n/,`find \"$DocDir\" -maxdepth 2 -type d`))
        {
            if(get_FileName($SubDir)=~/ref|log|change|install|api/i)
            {
                push(@LogDirs, $SubDir);
            }
        }
    }
    foreach my $SubDir (sort {get_depth($a) <=> get_depth($b)} sort {length(get_FileName($a))<=>length(get_FileName($b))}
    sort {cmpVersions(get_FileName($b), get_FileName($a))} @LogDirs)
    {# directories with logs
        foreach my $Path (sort {get_depth($a) <=> get_depth($b)} sort {length(get_FileName($a))<=>length(get_FileName($b))}
        sort {cmpVersions(get_FileName($b), get_FileName($a))} split(/\n/,`find \"$SubDir\" -maxdepth 1 -type f`))
        {
            next if(-s $Path<20); # almost empty
            next if(-s $Path<100 and $Version ne getInitialVersion($Library));
            next if($Path=~/\.\w+\Z/i and $Path!~/\.(log|html|htm|txt)\Z/i);
            next if(get_FileName($Path)=~/(config)/i);
            next if(length(get_FileName($Path))>25);
            my $RelPath = cut_path_prefix($Path, $Directory);
            next if($RelPath=~/\// and $RelPath!~/\A(doc|log|change)/i);
            my $Name = get_FileName($Path);
            if($Name=~/relnotes/i
            and $Name=~/(\A|\D)\Q$Version\E(\D|\Z)/)
            {
                push(@VRelNotes, $Path);
            }
            elsif($Name=~/Whats[._-]*New/i) {
                push(@WhatsNew, $Path);
            }
            elsif($Name=~/changes/i) {
                push(@Changes, $Path);
            }
            elsif($Name=~/revision/i) {
                push(@Revision, $Path);
            }
            elsif($Name=~/change/i) {
                push(@ChangeLog, $Path);
            }
            elsif($Name=~/history/i) {
                push(@History, $Path);
            }
            elsif($Name=~/news([^a-z]|\Z)/i) {
                push(@News, $Path);
            }
            elsif($Name=~/$Version/i) {
                push(@VersionInfo, $Path);
            }
            elsif($Name=~/readme/i) {
                push(@Readme, $Path);
            }
            elsif($Name=~/log(\W|\Z)/i) {
                push(@Log, $Path);
            }
            elsif($Name=~/relnotes/i)
            {
                push(@RelNotes, $Path);
            }
            elsif($Name=~/\A(\Q$Version\E|\Q$ShortVersion\E)[\d._-]*(\.htm|\Z)/i)
            {
                push(@ChangeLog, $Path);
            }
        }
    }
    
    if($#VRelNotes>=0) {
        return $VRelNotes[0];
    }
    elsif($#WhatsNew>=0) {
        return $WhatsNew[0];
    }
    elsif($#Changes>=0) {
        return $Changes[0];
    }
    elsif($#RelNotes>=0) {
        return $RelNotes[0];
    }
    elsif($#News>=0 and $#ChangeLog>=0
    and ((-s $News[0]) < (-s $ChangeLog[0]))
    and (-s $News[0]<500)) {
        return $ChangeLog[0];
    }
    elsif($#News>=0) {
        return $News[0];
    }
    elsif($#ChangeLog>=0) {
        return $ChangeLog[0];
    }
    elsif($#VersionInfo>=0) {
        return $VersionInfo[0];
    }
    elsif($#History>=0) {
        return $History[0];
    }
    elsif($#Log>=0) {
        return $Log[0];
    }
    elsif($#Revision>=0) {
        return $Revision[0];
    }
    elsif($#Readme>=0
    and readFile($Readme[0])=~/version\s+(\Q$Version\E)/i) {
        return $Readme[0];
    }
    return "";
}

sub get_conf_cmd($$)
{
    my ($Library, $BuildDir) = @_;
    return "" if(not $BuildDir or not -e $BuildDir or not $Library);
    if(-f "$BuildDir/configure.pl")
    {# perl
        return "perl ./configure.pl";
    }
    elsif(-f "$BuildDir/configure.py")
    {# python
        return "python ./configure.py";
    }
    elsif(-f "$BuildDir/configure")
    {# shell
        return "sh ./configure";
    }
    elsif(-f "$BuildDir/config")
    {# openssl
        return "sh ./config";
    }
    elsif(-f "$BuildDir/bootstrap-configure")
    {# blueZ
        return "sh ./bootstrap-configure";
    }
    elsif(-f "$BuildDir/bootstrap.sh")
    {# Boost
        return "sh ./bootstrap.sh";
    }
    elsif(-f "$BuildDir/bootstrap")
    {
        return "sh ./bootstrap";
    }
    elsif(-f "$BuildDir/Configure")
    {# other
        return "sh ./Configure";
    }
    elsif(-f "$BuildDir/../dist/configure"
    and -d "$BuildDir/../build_unix")
    {# libdb
        return "sh ../dist/configure";
    }
    elsif($Library eq "glibc")
    {
        return "sh ../configure";
    }
    return "";
}

sub get_conf_params($$$$)
{
    my ($Library, $ConfCmd, $ConfHelp, $InstallPath) = @_;
    return () if(not $Library or not $ConfCmd or not $ConfHelp or not $InstallPath);
    my ($Cmd, $DestPrms) = ($ConfCmd, "");
    if($ConfHelp=~/(\W|\A)([-]+prefix(\s*=\s*|\s+))/i)
    {
        $DestPrms .= " $2$InstallPath";
    }
    if($ConfHelp=~/(\W|\A)([-]+exec-prefix(\s*=\s*|\s+))/i)
    {
        $DestPrms .= " $2$InstallPath";
    }
    if($ConfHelp=~/(\W|\A)([-]+with-mysql)(\W|\Z)/i and $Library!~/mysql/i)
    {
        $Cmd .= " $2";
    }
    if($ConfHelp=~/(\W|\A)([-]+non-interactive)(\W|\Z)/i)
    {
        $Cmd .= " $2";
    }
    if($ConfHelp=~/(\W|\A)([-]+opensource)(\W|\Z)/i)
    {
        $Cmd .= " $2";
    }
    if($ConfHelp=~/(\W|\A)([-]+enable-shared|-shared)(\W|\Z)/)
    {
        $Cmd .= " $2";
    }
    foreach my $ConfOpt (keys(%{$LibraryInfo{$Library}{"ConfigureOptions"}}))
    {
        if($ConfOpt=~/\A([-]+[^\s=]+)(\s|=|\Z)/)
        {
            my $Option = $1;
            if($ConfHelp=~/(\W|\A)$Option(\W|\Z)/)
            {
                $Cmd .= " ".$ConfOpt;
            }
        }
        else
        {
            $Cmd .= " ".$ConfOpt;
        }
    }
    if($ConfHelp=~/--without-/)
    {# disabling other languages bindings
        if($ConfHelp=~/--with-python(\s+([^\=]|\Z)|\Z)/)
        {
            $Cmd .= " --without-python";
        }
        if($ConfHelp=~/--with-java(\s+([^\=]|\Z)|\Z)/)
        {
            $Cmd .= " --without-java";
        }
    }
    if($ConfHelp=~/--disable-feature(\W|\Z)/i)
    {# disabling demo programs
        if($ConfHelp=~/--enable-demos(\W|\Z)/i)
        {
            $Cmd .= " --disable-demos";
        }
        if($ConfHelp=~/--enable-mozilla-plugin(\W|\Z)/i)
        {
            $Cmd .= " --disable-mozilla-plugin";
        }
    }
    if($ConfHelp=~/(--disable-tcl)(\W|\Z)/i)
    {
        $Cmd .= " $1";
    }
    return ($Cmd.$DestPrms, $DestPrms);
}

sub modify_makefile($$)
{
    my ($Library, $Path) = @_;
    return if(not $Library or not $Path or not -e $Path);
    my $Makefile = readFile($Path);
    $Makefile=~s/(\s|\A)(-Werror|-mpentiumpro|-mpentium|-m486)(\s|\Z)/$1$3/g;
    writeFile($Path, $Makefile);
}

sub modify_makefile_tmpl($$)
{
    my ($Library, $Path) = @_;
    return if(not $Library or not $Path or not -e $Path);
    my $Makefile = readFile($Path);
    $Makefile=~s/(\n|\A)[\t ]*\$\(INSTALL_DATA\)[^\n]+\/usr\/share\/aclocal(\n|\Z)/$1$2/g;
    writeFile($Path, $Makefile);
}

sub install_LibraryVersion($$)
{
    my ($Library, $Version) = @_;
    return 0 if(not $Version or not $Library);
    if(not $VersionPath{$Library}{$Version})
    {
        print STDERR "ERROR: package with $Library $Version is not presented in the repository\n";
        return 0;
    }
    my $InstallPath = "$INSTALL_PATH/$Library/$Version";
    print "installing $Library-$Version".($UseDebugLib?" (debug)":"")." to \'$InstallPath\' ... ";
    if(isInstalled($Library, $Version))
    {
        if($ReInstall or $Alter)
        {
            rmtree($InstallPath);
        }
        else
        {
            print "\n$Library $Version was already installed (you can -reinstall version)\n";
            return 1;
        }
    }
    my $UnpackDir = unPack($Library, $Version);
    return 0 if(not $UnpackDir or not -d $UnpackDir);
    mkpath($InstallPath);
    $InstallPath = abs_path($InstallPath);
    my $ErrorDir = "install_logs/$Library/$Version";
    rmtree($ErrorDir);
    mkpath($ErrorDir);
    my $BuildDir = get_RootDir($UnpackDir);
    if(-d "$BuildDir/build_unix")
    {# libdb
        $BuildDir = "$BuildDir/build_unix";
    }
    elsif(-d "$BuildDir/make/linux")
    {# libdb
        $BuildDir = "$BuildDir/make/linux";
    }
    elsif(-d "$BuildDir/source"
    and -f "$BuildDir/source/configure")
    {# icu4c
        $BuildDir = "$BuildDir/source";
    }
    elsif(-d "$BuildDir/src"
    and -f "$BuildDir/src/configure")
    {# icu4c
        $BuildDir = "$BuildDir/src";
    }
    elsif(-f "$BuildDir/unix/configure")
    {# Tcl, Tk
        $BuildDir = "$BuildDir/unix";
    }
    elsif($Library eq "glibc")
    {#in a separate build directory
        mkpath("$BuildDir/build_unix");
        $BuildDir = "$BuildDir/build_unix";
    }
    my @IContext = ($Library, $Version, $BuildDir);
    my $ConfCmd = get_conf_cmd($Library, $BuildDir);
    my $ConfHelp = $ConfCmd?`cd $BuildDir && $ConfCmd --help 2>&1`:"";
    $ConfHelp="" if($ConfHelp!~/[a-z]/i);
    if(defined $LibraryInfo{$Library}{"Install"}{$Version})
    {# manual build and manual install
        foreach my $Rule (sort {$LibraryInfo{$Library}{"Install"}{$Version}{$a}<=>$LibraryInfo{$Library}{"Install"}{$Version}{$b}}
        keys(%{$LibraryInfo{$Library}{"Install"}{$Version}}))
        {
            $Rule=~s/{INSTALL}/$InstallPath/g;
            return 0 if(not exec_install_cmd(@IContext, "install", $Rule));
        }
    }
    elsif(defined $LibraryInfo{$Library}{"InstallGeneric"})
    {# manual build and manual install (generic)
        foreach my $Rule (sort {$LibraryInfo{$Library}{"InstallGeneric"}{$a}<=>$LibraryInfo{$Library}{"InstallGeneric"}{$b}}
        keys(%{$LibraryInfo{$Library}{"InstallGeneric"}}))
        {
            $Rule=~s/{INSTALL}/$InstallPath/g;
            return 0 if(not exec_install_cmd(@IContext, "install", $Rule));
        }
    }
    elsif(defined $LibraryInfo{$Library}{"Build"}{$Version})
    {# manual build and auto install
        foreach my $Rule (sort {$LibraryInfo{$Library}{"Build"}{$Version}{$a}<=>$LibraryInfo{$Library}{"Build"}{$Version}{$b}}
        keys(%{$LibraryInfo{$Library}{"Build"}{$Version}}))
        {
            $Rule=~s/{INSTALL}/$InstallPath/g;
            return 0 if(not exec_install_cmd(@IContext, "build", $Rule));
        }
        manual_install($Library, $Version, $BuildDir);
    }
    elsif(defined $LibraryInfo{$Library}{"BuildGeneric"})
    {# manual build and auto install (generic)
        foreach my $Rule (sort {$LibraryInfo{$Library}{"BuildGeneric"}{$a}<=>$LibraryInfo{$Library}{"BuildGeneric"}{$b}}
        keys(%{$LibraryInfo{$Library}{"BuildGeneric"}}))
        {
            $Rule=~s/{INSTALL}/$InstallPath/g;
            return 0 if(not exec_install_cmd(@IContext, "build", $Rule));
        }
        manual_install($Library, $Version, $BuildDir);
    }
    elsif(-f $BuildDir."/CMakeLists.txt"
    and (not $ConfCmd or not $ConfHelp or $ConfHelp=~/instead of configure|uses the cmake|use cmake/i))
    {# cmake
        return 0 if(not exec_install_cmd(@IContext, "preamble", "mkdir -p ut_build"));
        return 0 if(not exec_install_cmd(@IContext, "configure", "cd ut_build && cmake ../ -DCMAKE_INSTALL_PREFIX=$InstallPath".($UseDebugLib?" -DCMAKE_BUILD_TYPE=Debug":"")));
        return 0 if(not exec_install_cmd(@IContext, "build", "cd ut_build && make"));
        return 0 if(not exec_install_cmd(@IContext, "install", "cd ut_build && make install"));
    }
    elsif(-f $BuildDir."/SConstruct" and (not $ConfCmd or not $ConfHelp))
    {# scons
        return 0 if(not exec_install_cmd(@IContext, "build", "scons all"));
        return 0 if(not exec_install_cmd(@IContext, "install", "scons --prefix=$InstallPath install"));
    }
    else
    {# auto build and auto install
        if(my $Dos2Unix = get_CmdPath("dos2unix"))
        {
            foreach my $File ("config.guess", "config.sub", "mkinstalldirs", "install-sh", "configure")
            {
                if(-f $BuildDir."/".$File)
                {# removing windows symbols from the shell script
                    exec_install_cmd(@IContext, "preamble", "$Dos2Unix $File");
                }
            }
        }
        
        my $AutoGenCmd = "autogen.sh";
        $AutoGenCmd = "autogen" if(not -f $BuildDir."/".$AutoGenCmd);
        $AutoGenCmd = "" if(not -f $BuildDir."/".$AutoGenCmd);
        if(not $ConfCmd)
        {
            if($AutoGenCmd)
            {
                return 0 if(not exec_install_cmd(@IContext, "preamble", "sh $AutoGenCmd"));
            }
            elsif(-f $BuildDir."/buildconf")
            {
                return 0 if(not exec_install_cmd(@IContext, "preamble", "./buildconf"));
            }
            if($ConfCmd = get_conf_cmd($Library, $BuildDir))
            {
                $ConfHelp = `cd $BuildDir && $ConfCmd --help`;
            }
        }
        my ($CfgCmd, $DestPrms) = get_conf_params($Library, $ConfCmd, $ConfHelp, $InstallPath);
        if($UseDebugLib)
        {
            my $DebugOpts = "";
            if($ConfHelp=~/--with-debug(\W|\Z)/)
            {
                $DebugOpts .= " --with-debug";
            }
            if($ConfHelp=~/--enable-debug=\[no\/minimum\/yes\](\W|\Z)/)
            {
                $DebugOpts .= " --enable-debug=yes";
            }
            elsif($ConfHelp=~/--enable-debug(\W|\Z)/)
            {
                $DebugOpts .= " --enable-debug";
            }
            if($DebugOpts)
            {
                $CfgCmd .= $DebugOpts;
            }
            else
            {
                print "failed\n";
                print STDERR "ERROR: can't find debugging options\n";
                clearUnpacked($Library, $Version);
                rmtree($InstallPath);
                return 0;
            }
        }
        
        modify_makefile_tmpl($Library, $BuildDir."/Makefile.in") if(-f $BuildDir."/Makefile.in");
        modify_makefile_tmpl($Library, $BuildDir."/Makefile.am") if(-f $BuildDir."/Makefile.am");
        return 0 if($CfgCmd and not exec_install_cmd(@IContext, "configure", $CfgCmd));
        modify_makefile($Library, $BuildDir."/Makefile") if(-f $BuildDir."/Makefile");
        modify_makefile($Library, $BuildDir."/src/Makefile") if(-f $BuildDir."/Makefile");
        if(defined $LibraryInfo{$Library}{"PreInstall"}{$Version})
        {
            foreach my $Rule (sort {$LibraryInfo{$Library}{"PreInstall"}{$Version}{$a}<=>$LibraryInfo{$Library}{"PreInstall"}{$Version}{$b}}
            keys(%{$LibraryInfo{$Library}{"PreInstall"}{$Version}}))
            {
                $Rule=~s/{INSTALL}/$InstallPath/g;
                return 0 if(not exec_install_cmd(@IContext, "pre_install", $Rule));
            }
        }
        elsif(defined $LibraryInfo{$Library}{"PreInstallGeneric"})
        {
            foreach my $Rule (sort {$LibraryInfo{$Library}{"PreInstallGeneric"}{$a}<=>$LibraryInfo{$Library}{"PreInstallGeneric"}{$b}}
            keys(%{$LibraryInfo{$Library}{"PreInstallGeneric"}}))
            {
                $Rule=~s/{INSTALL}/$InstallPath/g;
                return 0 if(not exec_install_cmd(@IContext, "pre_install", $Rule));
            }
        }
        # building and installing
        if(-f $BuildDir."/Makefile" and $CfgCmd and $DestPrms)
        {
            return 0 if(not exec_install_cmd(@IContext, "build", "make"));
            my $MakeContent = readFile($BuildDir."/Makefile");
            if($MakeContent=~/(\A|\s)(install:|include\s)/)
            {
                return 0 if(not exec_install_cmd(@IContext, "install", "make install"));
            }
            if($MakeContent=~/(\A|\s)install-headers:/)
            {
                return 0 if(not exec_install_cmd(@IContext, "install", "make install-headers"));
            }
            if($MakeContent=~/(\A|\s)install-lib:/)
            {
                return 0 if(not exec_install_cmd(@IContext, "install", "make install-lib"));
            }
        }
        elsif(-f $BuildDir."/Makefile"
        and my $PrefixDefines = get_prefix_defines($BuildDir."/Makefile", $InstallPath))
        {
            return 0 if(not exec_install_cmd(@IContext, "build", "make"));
            my $MakeContent = readFile($BuildDir."/Makefile");
            if($MakeContent=~/(\A|\s)install:/)
            {
                return 0 if(not exec_install_cmd(@IContext, "install", "make install $PrefixDefines"));
            }
            if($MakeContent=~/(\A|\s)install-headers:/)
            {
                return 0 if(not exec_install_cmd(@IContext, "install", "make install-headers $PrefixDefines"));
            }
            if($MakeContent=~/(\A|\s)install-lib:/)
            {
                return 0 if(not exec_install_cmd(@IContext, "install", "make install-lib $PrefixDefines"));
            }
        }
        elsif(-f $BuildDir."/Makefile")
        {
            return 0 if(not exec_install_cmd(@IContext, "build", "make"));
            manual_install($Library, $Version, $BuildDir);
        }
        elsif(-f $BuildDir."/Jamroot" and $CfgCmd and $DestPrms)
        {# bjam
            return 0 if(not exec_install_cmd(@IContext, "install", "bjam install"));
        }
        elsif(-f $BuildDir."/scripts/makefile.lnx"
        or -f $BuildDir."/scripts/makefile.linux"
        or -f $BuildDir."/scripts/makefile.elf")
        {# libpng
            copy($BuildDir."/scripts/makefile.linux", "$BuildDir/Makefile");
            copy($BuildDir."/scripts/makefile.lnx", "$BuildDir/Makefile") if(not -f "$BuildDir/Makefile");
            copy($BuildDir."/scripts/makefile.elf", "$BuildDir/Makefile") if(not -f "$BuildDir/Makefile");
            return 0 if(not exec_install_cmd(@IContext, "build", "make"));
            if(my $PrefixDefines = get_prefix_defines($BuildDir."/Makefile", $InstallPath))
            {
                return 0 if(not exec_install_cmd(@IContext, "install", "make install $PrefixDefines"));
            }
        }
        elsif(-f $BuildDir."/makefile"
        and my $PrefixDefines = get_prefix_defines($BuildDir."/makefile", $InstallPath))
        {
            return 0 if(not exec_install_cmd(@IContext, "build", "make"));
            return 0 if(not exec_install_cmd(@IContext, "install", "make install $PrefixDefines"));
        }
        elsif(-f $BuildDir."/GNUmakefile"
        and my $PrefixDefines = get_prefix_defines($BuildDir."/GNUmakefile", $InstallPath))
        {# stdcxx
            return 0 if(not exec_install_cmd(@IContext, "build", "make -f GNUmakefile"));
            return 0 if(not exec_install_cmd(@IContext, "install", "make -f GNUmakefile install $PrefixDefines"));
        }
        elsif(-f $BuildDir."/src/Makefile")
        {
            return 0 if(not exec_install_cmd(@IContext, "build", "cd src && make"));
            manual_install($Library, $Version, $BuildDir);
        }
        elsif(-d $BuildDir."/include"
        and -d $BuildDir."/lib")
        {# NVIDIA Cg
            move($BuildDir."/include", $InstallPath."/include");
            move($BuildDir."/lib", $InstallPath."/lib");
        }
        else
        {
            print "failed\n";
            print STDERR "ERROR: can't find/create a makefile\n";
            clearUnpacked($Library, $Version);
            rmtree($InstallPath);
            return 0;
        }
    }
    
    if(defined $LibraryInfo{$Library}{"PostInstall"}{$Version})
    {
        foreach my $Rule (sort {$LibraryInfo{$Library}{"PostInstall"}{$Version}{$a}<=>$LibraryInfo{$Library}{"PostInstall"}{$Version}{$b}}
        keys(%{$LibraryInfo{$Library}{"PostInstall"}{$Version}}))
        {
            $Rule=~s/{INSTALL}/$InstallPath/g;
            exec_install_cmd(@IContext, "post_install", $Rule);
        }
    }
    elsif(defined $LibraryInfo{$Library}{"PostInstallGeneric"})
    {
        foreach my $Rule (sort {$LibraryInfo{$Library}{"PostInstallGeneric"}{$a}<=>$LibraryInfo{$Library}{"PostInstallGeneric"}{$b}}
        keys(%{$LibraryInfo{$Library}{"PostInstallGeneric"}}))
        {
            $Rule=~s/{INSTALL}/$InstallPath/g;
            exec_install_cmd(@IContext, "post_install", $Rule);
        }
    }
    
    if(not defined $LibraryInfo{$Library}{"Install"}{$Version}
    and not defined $LibraryInfo{$Library}{"InstallGeneric"})
    {
        manual_install_additions($Library, $BuildDir, $InstallPath);
        
        if(not cmd_find($InstallPath,"","*\.so*"))
        {
            if(not static_to_shared($InstallPath))
            {
                manual_install($Library, $Version, $BuildDir);
            }
        }
    }
    clearUnpacked($Library, $Version);
    if(isInstalled($Library, $Version))
    {
        post_install($Library, $Version);
        print "success\n";
        if($LibraryInfo{$Library}{"Actions"}{"mkcpp"})
        {
            mkcpp_LibraryVersion($Library, $Version);
        }
        if($Version eq "current")
        {
            if(get_FileName($VersionPath{$Library}{$Version})=~/-current-(.+)\.tar\.gz\Z/)
            {
                my $DateStamp = $1;
                $DateStamp=~s/(\d+)_(\d+)\Z/$1:$2/g;
                $DateStamp=~s/_(\d+:\d+)\Z/ $1/g;
                writeFile($InstallPath."/date.txt", $DateStamp);
            }
        }
        return 1;
    }
    else
    {
        print "failed\n";
        print STDERR "ERROR: incorrect installation (shared objects were not installed)\n";
        rmtree($InstallPath);
        return 0;
    }
}

sub post_install($$)
{
    my ($Library, $Version) = @_;
    return 0 if(not $Version or not $Library);
    my $InstallPath = "$INSTALL_PATH/$Library/$Version";
    foreach my $Path (split(/\n/, `find $InstallPath -maxdepth 1 -type d`))
    {
        if(get_FileName($Path)=~/\A(share|bin|examples|demos|doc|info|man|etc)\Z/)
        {
            rmtree($Path);
        }
    }
}

sub separatePath($)
{
    my $Path = $_[0];
    return (get_Directory($Path), get_FileName($Path));
}

sub manual_install_additions($$$)
{
    my ($Library, $BuildDir, $InstallPath) = @_;
    return if(not $Library or not -d $BuildDir or not -d $InstallPath);
    if($Library eq "beecrypt")
    {
        if(-f "$BuildDir/config.h")
        {
            copy("$BuildDir/config.h", "$InstallPath/");
        }
        if(-f "$BuildDir/config.gnu.h")
        {
            copy("$BuildDir/config.gnu.h", "$InstallPath/");
        }
        if(-f "$BuildDir/gnu/config.gnu.h")
        {
            mkpath("$InstallPath/gnu/");
            copy("$BuildDir/gnu/config.gnu.h", "$InstallPath/gnu/");
        }
    }
}

sub static_to_shared($)
{
    my $Dir = $_[0];
    return () if(not $Dir or not -d $Dir);
    my @StaticObjects = cmd_find($Dir,"f","*\.a");
    return () if($#StaticObjects<0);
    my @SharedObjects = ();
    foreach my $Path (@StaticObjects)
    {
        my $Name = get_FileName($Path);
        my $SoName = $Name;
        $SoName=~s/\.a\Z/.so/;
        mkpath(".tmpdir");
        copy($Path, ".tmpdir/");
        system("cd .tmpdir/ && ar -x $Name && gcc *.o -shared -o $SoName");
        my $SoPath = get_Directory($Path)."/".$SoName;
        copy(".tmpdir/".$SoName, $SoPath);
        push(@SharedObjects, $SoPath);
        rmtree(".tmpdir");
    }
    return @SharedObjects;
}

sub manual_install($$$)
{
    my ($Library, $Version, $BuildDir) = @_;
    return 0 if(not $Library or not $Version or not $BuildDir);
    my $InstallPath = "$INSTALL_PATH/$Library/$Version";
    my (@Headers, @SharedObjects) = ();
    foreach my $Path (cmd_find($BuildDir,"f","*\.h"))
    {
        my $RelPath = cut_path_prefix(get_Directory($Path), $BuildDir);
        if(($RelPath=~/\A(src|include|lib|$Library|core|crypto)(.*)(\/|\Z)/ or not $RelPath)
        and get_FileName($Path)!~/private|(_p\Z)/i)
        {
            push(@Headers, $Path);
        }
    }
    return 0 if($#Headers<0);
    @SharedObjects = find_shared_objects($BuildDir);
    if($#SharedObjects<0)
    {
        @SharedObjects = static_to_shared($BuildDir);
    }
    return 0 if($#SharedObjects<0);
    # installing
    foreach my $Path (@Headers)
    {
        my $Suffix = cut_path_prefix($Path, $BuildDir);
        while($Suffix=~s/(\A|\/)(src|headers|include)\//\//g){};
        my $Dest = $InstallPath."/include/".$Suffix;
        mkpath(get_Directory($Dest));
        copy($Path, $Dest);
    }
    foreach my $Path (@SharedObjects)
    {
        my $Suffix = cut_path_prefix($Path, $BuildDir);
        while($Suffix=~s/(\A|\/)(\.|)(lib|libs|obj|objs)\//\//g){};
        my $Dest = $InstallPath."/lib/".$Suffix;
        mkpath(get_Directory($Dest));
        move($Path, $Dest);
    }
    return 1;
}

sub find_shared_objects($)
{# all shared objects and symbolic links
    my $Dir = $_[0];
    return () if(not $Dir or not -e $Dir);
    my @SharedObjects = ();
    foreach my $Path (cmd_find($Dir,"","*\.so*"))
    {
        if(get_FileName($Path)=~/\.so[.0-9]*\Z/)
        {
            push(@SharedObjects, $Path);
        }
    }
    return @SharedObjects;
}

sub cut_path_prefix($$)
{
    my ($Path, $Prefix) = @_;
    $Prefix=~s/[\/]+\Z//;
    $Path=~s/\A\Q$Prefix\E([\/]+|\Z)//;
    return $Path;
}

sub separate_streams($$$)
{
    my ($Cmd, $Out_Path, $StdErr_Path) = @_;
    return $Cmd if(not $Out_Path or not $StdErr_Path);
    return "(( $Cmd 3>&1 1>&2 2>&3 | tee -a $StdErr_Path ) 3>&1 1>&2 2>&3) >> $Out_Path 2>&1";
}

sub exec_install_cmd($$$$$)
{
    my ($Library, $Version, $BuildDir, $CommandType, $Command) = @_;
    return 0 if(not $Library or not $Version or not $BuildDir or not $Command or not $CommandType);
    my $ErrorDir = abs_path("install_logs/$Library/$Version");
    my $LogPath = $ErrorDir."/".$CommandType;
    writeFile($LogPath, $Command."\n\n");
    if($LibraryInfo{$Library}{"InstallMode"} eq "interactive")
    {# FIXME: check for errors
        interactive_Cmd("cd $BuildDir && $Command|tee -a $LogPath", "");
        return 1;
    }
    else
    {
        system("cd $BuildDir && $Command >> $LogPath 2>&1");
        if($?)
        {
            print "failed\n";
            print STDERR "ERROR: failed to \'$Command\'\n";
            print STDERR "see error log in the \'$ErrorDir\'\n";
            clearUnpacked($Library, $Version);
            rmtree("$INSTALL_PATH/$Library/$Version");
            return 0;
        }
        else
        {
            return 1;
        }
    }
    
}

sub install_Library($)
{
    my $Library = $_[0];
    return if(not $Library);
    if($IncludeCurrent)
    {# uninstalling current (scm) version
        rmtree("$INSTALL_PATH_DEFAULT/$Library/current");
        rmtree("$INSTALL_PATH_DEBUG/$Library/current");
    }
    my $AlreadyInstalled=1;
    my @Versions = sort_VersionSequence(keys(%{$RepositoryProvide{$Library}}), $Library);
    if($#Versions<0)
    {
        print STDERR "ERROR: there are no packages of $Library in the repository (use -get option for dowloading packages or manually download it to \'$REPOSITORY_PATH/$Library\')\n";
        return;
    }
    foreach my $Version (@Versions)
    {
        next if(not $IncludeCurrent and $Version eq "current");
        if($UpdateInfo and not $UpdateLog{$Library}{"New"}{$Version} and ($Version ne "current" or not $IncludeCurrent))
        {
            $AlreadyInstalled=0;
            next;
        }
        if(not isInstalled($Library, $Version) or defined $ReInstall or defined $Alter)
        {
            install_LibraryVersion($Library, $Version);
            $AlreadyInstalled=0;
        }
    }
    if($AlreadyInstalled)
    {
        print "all versions of $Library were already installed (you can -reinstall versions)\n";
    }
}

sub changelog_Library($)
{
    my $Library = $_[0];
    return if(not $Library);
    my $AlreadyObtained=1;
    my @Versions = sort {cmpVersions($b, $a)} keys(%{$RepositoryProvide{$Library}});
    if($#Versions<0)
    {
        print STDERR "ERROR: there are no packages of $Library in the repository (use -get option for dowloading packages or manually download it to \'$REPOSITORY_PATH/$Library\')\n";
        return;
    }
    foreach my $Version (@Versions)
    {
        next if($Version eq "current" and not $IncludeCurrent);
        if($UpdateInfo and not $UpdateLog{$Library}{"New"}{$Version} and ($Version ne "current" or not $IncludeCurrent))
        {
            $AlreadyObtained=0;
            next;
        }
        if(not -f "changelogs/$Library/$Version/changelog.html" or $Alter or $Version eq "current")
        {
            changelog_LibraryVersion($Library, $Version);
            $AlreadyObtained=0;
        }
    }
    if($AlreadyObtained and keys(%{$RepositoryProvide{$Library}}))
    {
        print "all $Library changelogs were already found (use -alter option for searching again)\n";
    }
}

sub get_InstalledVersions($)
{
    my $Library = $_[0];
    return () if(not $Library);
    my @Versions = ();
    foreach my $Path (split(/\n/, `find $INSTALL_PATH/$Library -maxdepth 1 -type d`))
    {
        push(@Versions, get_FileName($Path));
    }
    @Versions = sort_VersionSequence(@Versions, $Library);
    sort_Installed($Library, \@Versions);
    return @Versions;
}

sub test_Library($)
{
    my $Library = $_[0];
    return if(not $Library);
    if($IncludeCurrent and isInstalled($Library, "current"))
    {
        remove_LibraryVersion($Library, "current", "tests");
    }
    my $AlreadyTested=1;
    my @Versions = sort_VersionSequence(keys(%{$RepositoryProvide{$Library}}), $Library);
    sort_Installed($Library, \@Versions);
    if($#Versions<0)
    {
        print STDERR "ERROR: $Library versions".($UseDebugLib?" (debug)":"")." were not installed yet\n";
        return;
    }
    foreach my $Version (@Versions)
    {
        next if(not $IncludeCurrent and $Version eq "current");
        if($UpdateInfo and not $UpdateLog{$Library}{"New"}{$Version} and ($Version ne "current" or not $IncludeCurrent))
        {
            $AlreadyTested=0;
            next;
        }
        if(not -f "test_results/$Library/$Version/test_results.html"
        or not -f "tests/$Library/$Version/view_tests.html"
        or defined $Alter)
        {
            test_LibraryVersion($Library, $Version, "standalone");
            $AlreadyTested=0;
        }
    }
    if($AlreadyTested and $#Versions>=0)
    {
        print "all installed versions of $Library were already tested (use -alter option for testing again)\n";
    }
}

sub genT2C_Library($)
{
    my $Library = $_[0];
    return if(not $Library);
    if($IncludeCurrent and isInstalled($Library, "current"))
    {
        remove_LibraryVersion($Library, "current", "tests_t2c");
    }
    my $AlreadyTested=1;
    my @Versions = sort_VersionSequence(keys(%{$RepositoryProvide{$Library}}), $Library);
    sort_Installed($Library, \@Versions);
    if($#Versions<0)
    {
        print STDERR "ERROR: $Library versions".($UseDebugLib?" (debug)":"")." were not installed yet\n";
        return;
    }
    foreach my $Version (@Versions)
    {
        next if(not $IncludeCurrent and $Version eq "current");
        if($UpdateInfo and not $UpdateLog{$Library}{"New"}{$Version} and ($Version ne "current" or not $IncludeCurrent))
        {
            $AlreadyTested=0;
            next;
        }
        if(not -f "tests_t2c/$Library/$Version/t2c-$Library-tests-$Version.tar.gz" or defined $Alter)
        {
            test_LibraryVersion($Library, $Version, "t2c");
            $AlreadyTested=0;
        }
    }
    if($AlreadyTested and $#Versions>=0)
    {
        print "tests in t2c format were already generated for all installed versions of $Library (use -alter option for generating tests again)\n";
    }
}

sub cmp_Library($)
{
    my $Library = $_[0];
    return if(not $Library);
    if($IncludeCurrent and isInstalled($Library, "current"))
    {
        remove_LibraryVersion($Library, "current", "compat_reports");
    }
    my $AlreadyTested=1;
    my @Versions = sort_VersionSequence(keys(%{$RepositoryProvide{$Library}}), $Library);
    sort_Installed($Library, \@Versions);
    sort_CreateDescriptor($Library, \@Versions);
    if($#Versions<0)
    {
        print STDERR "ERROR: $Library versions".($UseDebugLib?" (debug)":"")." were not installed yet\n";
        return;
    }
    return if($#Versions<=0);# needed more than 2 versions
    foreach my $Num (0 .. $#Versions-1)
    {
        my $Version1 = $Versions[$Num+1];
        my $Version2 = $Versions[$Num];
        next if(not $IncludeCurrent and $Version2 eq "current");
        if($UpdateInfo and not $UpdateLog{$Library}{"New"}{$Version1} and not $UpdateLog{$Library}{"New"}{$Version2}
        and ($Version2 ne "current" or not $IncludeCurrent))
        {
            $AlreadyTested=0;
            next;
        }
        if($Version1 and $Version2 and (not -e "compat_reports/$Library/$Version1"."_to_$Version2/abi_compat_report.html" or defined $Alter))
        {
            cmp_LibraryVersions($Library, $Version1, $Version2);
            $AlreadyTested=0;
        }
    }
    if($AlreadyTested and $#Versions>=0)
    {
        print "all installed versions of $Library were already checked for ABI changes (use -alter option for checking again)\n";
    }
}

sub cmpLibVersions($$$)
{
    my ($V1, $V2, $Library) = @_;
    return cmpVersions($V1, $V2) if(not $Library);
    return 0 if($V1 eq $V2);
    if(getVersion_Type($Library, $V1) eq getVersion_Type($Library, $V2))
    {
        if($V1=~/\Q$V2\E/)
        {
            return 1;
        }
        elsif($V2=~/\Q$V1\E/)
        {
            return -1;
        }
    }
    return cmpVersions($V1, $V2);
}

sub sort_VersionSequence(@)
{
    my $Library = pop(@_);
    return () if(not $Library);
    my @Versions = @_;
    my ($LastFinal, $Current, $LastUnstable) = ("", "", "");
    my @Releases = ();
    foreach my $Version (sort {cmpLibVersions($a, $b, $Library)} @Versions)
    {
        my $VType = getVersion_Type($Library, $Version);
        if($VType eq "release")
        {
            $LastFinal = $Version;
            push(@Releases, $Version);
        }
        elsif($VType eq "current"
        or $Version eq "current")
        {
            $Current = $Version;
        }
        elsif($VType eq "unknown")
        {
            next;
        }
        else
        {
            $LastUnstable = $Version;
        }
    }
    my @NaturalSequence = @Releases;
    if($LastUnstable)
    {
        if($LastFinal)
        {
            if(cmpVersions($LastUnstable, $LastFinal)==1)
            {
                push(@NaturalSequence, $LastUnstable);
            }
        }
        else
        {
            push(@NaturalSequence, $LastUnstable);
        }
    }
    if($Current)
    {
        push(@NaturalSequence, $Current);
    }
    return reverse(@NaturalSequence);
}

sub sort_Installed($$)
{
    my ($Library, $VersionsRef) = @_;
    return () if(not $Library or not $VersionsRef);
    my @SortedVersions = ();
    foreach my $Version (@{$VersionsRef})
    {
        if(isInstalled($Library, $Version))
        {
            push(@SortedVersions, $Version);
        }
    }
    @{$VersionsRef} = @SortedVersions;
    return @SortedVersions;
}

sub sort_CreateDescriptor($$)
{
    my ($Library, $VersionsRef) = @_;
    return () if(not $Library or not $VersionsRef);
    my @SortedVersions = ();
    foreach my $Version (@{$VersionsRef})
    {
        if(createDescriptor($Library, $Version))
        {
            push(@SortedVersions, $Version);
        }
    }
    @{$VersionsRef} = @SortedVersions;
    return @SortedVersions;
}

sub detect_bin_default_paths()
{
    my $EnvPaths = $ENV{"PATH"};
    if($Config{"osname"}=~/\A(haiku|beos)\Z/)
    {
        $EnvPaths.=":".$ENV{"BETOOLS"};
    }
    foreach my $Path (sort {length($a)<=>length($b)} split(/:|;/, $EnvPaths))
    {
        if($Path ne "/")
        {
            $Path=~s/[\/]+\Z//g;
            next if(not $Path);
        }
        $DefaultBinPaths{$Path} = 1;
    }
}

sub detect_default_paths()
{
    my $TargetCfg = ($Config{"osname"}=~/\A(haiku|beos)\Z/)?"haiku":"default";
    foreach my $Type (keys(%{$OperatingSystemAddPaths{$TargetCfg}}))
    {# additional search paths
        foreach my $Path (keys(%{$OperatingSystemAddPaths{$TargetCfg}{$Type}}))
        {
            next if(not -d $Path);
            $SystemPaths{$Type}{$Path} = $OperatingSystemAddPaths{$TargetCfg}{$Type}{$Path};
        }
    }
    detect_bin_default_paths();
    foreach my $Path (keys(%DefaultBinPaths))
    {
        $SystemPaths{"bin"}{$Path} = $DefaultBinPaths{$Path};
    }
    foreach my $Path (split(/\n/, `find / -maxdepth 1 -name "*bin*" -type d`))
    {# autodetecting bin directories
        $SystemPaths{"bin"}{$Path} = 1;
    }
}

sub get_library_full_name($)
{
    my $Library = $_[0];
    return "" if(not $Library);
    if(my $FullName = $LibraryInfo{$Library}{"FullName"})
    {
        return $FullName;
    }
    return $Library;
}

sub cut_off_number($$)
{
    my ($num, $digs_to_cut) = @_;
    if($num!~/\./)
    {
        $num .= ".";
        foreach (1 .. $digs_to_cut-1)
        {
            $num .= "0";
        }
    }
    elsif($num=~/\.(.+)\Z/ and length($1)<$digs_to_cut-1)
    {
        foreach (1 .. $digs_to_cut - 1 - length($1))
        {
            $num .= "0";
        }
    }
    elsif($num=~/\d+\.(\d){$digs_to_cut,}/)
    {
      $num=sprintf("%.".($digs_to_cut-1)."f", $num);
    }
    return $num;
}

sub detect_language($$)
{
    my ($Library, $Version) = @_;
    return "" if(not $Library or not $Version);
    my $ReadelfCmd = get_CmdPath("readelf");
    if(not $ReadelfCmd)
    {
        print STDERR "ERROR: can't find readelf\n";
        exit(1);
    }
    foreach my $SoPath (sort {length($a)<=>length($b)} sort {lc($a) cmp lc($b)}
    cmd_find("$INSTALL_PATH/$Library/$Version","f","*.so*"))
    {
        next if($SoPath!~/\.so([0-9\.]*)\Z/);
        if(`objdump -T $SoPath|grep __gxx_personality`)
        {
            return "C++";
        }
    }
    return "C";
}

sub is_public_soname($$$)
{
    my ($Library, $Version, $Soname) = @_;
    return "" if(not $Library or not $Version or not $Soname);
    my @PublicSonames = getPkgConfigLibs($Library, $Version);
    return 1 if($#PublicSonames==-1);
    foreach my $PkgCfgName (@PublicSonames)
    {
        return 1 if($Soname=~/\Q$PkgCfgName\E/);
    }
    return 0;
}

sub get_soname_version($$)
{
    my ($Library, $Version) = @_;
    return "" if(not $Library or not $Version);
    if((my $SoVer = get_attr($Library, $Version, "sonames")) ne "")
    {
        return $SoVer;
    }
    my %Versions = ();
    my $InstallPath = "$INSTALL_PATH/$Library/$Version";
    $InstallPath .= "/lib" if(-d $InstallPath."/lib");
    foreach my $SoPath (cmd_find("$INSTALL_PATH/$Library/$Version","","*.so*"))
    {
        next if($SoPath!~/\.so([0-9\.]*)\Z/);
        next if(skip_lib($Library, $SoPath));
        next if(not is_public_soname($Library, $Version, get_FileName($SoPath)));
        my $NameInfo = `objdump -x $SoPath 2>&1 | grep SONAME`;
        my @VersionParts = ();
        $NameInfo=~s/(-|_)([a-z]+\d+)(-|_|\.)/$1$2/ig;
        $NameInfo=~s/\Q$Library\E//ig;
        next if(get_FileName($SoPath)=~/\A\Q$Library\E(-|_)[a-z]+/i);
        if($NameInfo=~/SONAME\s+.+?(\d[0-9\.\-_]*)\.so/)
        {
            my $V1 = $1;
            push(@VersionParts, $1) if($V1!~/_/);
        }
        if($NameInfo=~/SONAME\s+.+\.so\.([0-9\.]+)/)
        {
            push(@VersionParts, $1);
        }
        if((my $FVersion = join(".", @VersionParts)) ne "")
        {
            $Versions{$FVersion} = 1;
        }
    }
    my @V = sort {cmpVersions($b, $a)} keys(%Versions);
    if($#V>=0)
    {
        add_attr($Library, $Version, "sonames", $V[0]);
        return $V[0];
    }
    else
    {
        add_attr($Library, $Version, "sonames", "none");
        return "none";
    }
}

sub createLog()
{
    return if(not $UpdateInfo);
    unlink("news.html");
    return if(not keys(%UpdateLog));
    
    my $LogStyles = "    body{font-family:Arial;font-size:15px;background:#ffffff;color:Black;}
    hr{color:Black;background-color:Black;height:1px;border:0;}
    h1.title1{font-size:26px;}
    h2.title2{margin:0px;padding:0px;font-size:20px;}
    table.wikitable{border-collapse:collapse;margin:1em 1em 1em 0;background:#f9f9f9;border:1px #aaaaaa solid;}
    table.wikitable td{border:1px #aaaaaa solid;text-align:center;}
    table.wikitable th{background:#eeeeee;border:1px #aaaaaa solid;padding:0.2em;padding-left:0.4em;padding-right:0.4em;}
    table.wikitable td.left{text-align:left;background-color:#f2f2f2;}
    table.wikitable td.value{padding-left:5px;padding-right:5px;width:35px;font-size:15px;}
    table.wikitable td.destr{padding:5px;width:35px;font-size:16px;background-color:#f2f2f2;}";
    
    my ($LogNew, $LogAdded, $LogCompat, $LogTest) = "";
    foreach my $Library (sort {lc($a) cmp lc($b)} keys(%UpdateLog))
    {
        foreach my $Version (sort {cmpVersions($b, $a)} keys(%{$UpdateLog{$Library}{"New"}}))
        {
            next if($Version eq "current");
            $LogNew .= "<a href='versions/$Library.html'>".get_library_full_name($Library)."</a>-$Version (<a href='".$UpdateLog{$Library}{"New"}{$Version}."'>source</a>)<br/>\n";
        }
        foreach my $Version (sort {cmpVersions($b, $a)} keys(%{$UpdateLog{$Library}{"Added"}}))
        {
            $LogAdded .= "<a href='versions/$Library.html'>".get_library_full_name($Library)."</a>-$Version (<a href='".$UpdateLog{$Library}{"Added"}{$Version}."'>source</a>)<br/>\n";
        }
        foreach my $Version1 (sort {cmpVersions($b, $a)} keys(%{$UpdateLog{$Library}{"Compat"}}))
        {
            foreach my $Version2 (sort {cmpVersions($b, $a)} keys(%{$UpdateLog{$Library}{"Compat"}{$Version1}}))
            {
                next if($Version2 eq "current");
                my %Attributes = %{$UpdateLog{$Library}{"Compat"}{$Version1}{$Version2}};
                my ($Added, $Withdrawn) = ($Attributes{"added"}, $Attributes{"withdrawn"});
                my $TotalProblemsLow = $Attributes{"type_problems_low"}+$Attributes{"interface_problems_low"}+$Attributes{"changed_constants"};
                my $TotalProblemsMedium = $Attributes{"type_problems_medium"}+$Attributes{"interface_problems_medium"};
                my $TotalProblemsHigh = $Attributes{"type_problems_high"}+$Attributes{"interface_problems_high"};
                
                $Added="<a href='compat_reports/$Library/$Version1\_to_$Version2/abi_compat_report.html#Added\'>$Added</a>" if($Added);
                $Withdrawn="<a href='compat_reports/$Library/$Version1\_to_$Version2/abi_compat_report.html#Withdrawn\'>$Withdrawn</a>" if($Withdrawn);
                $TotalProblemsLow="<a href='compat_reports/$Library/$Version1\_to_$Version2/abi_compat_report.html\'>$TotalProblemsLow</a>" if($TotalProblemsLow);
                $TotalProblemsMedium="<a href='compat_reports/$Library/$Version1\_to_$Version2/abi_compat_report.html\'>$TotalProblemsMedium</a>" if($TotalProblemsMedium);
                $TotalProblemsHigh="<a href='compat_reports/$Library/$Version1\_to_$Version2/abi_compat_report.html\'>$TotalProblemsHigh</a>" if($TotalProblemsHigh);
                
                $LogCompat .= "<a href='versions/$Library.html'>".get_library_full_name($Library)."</a>-$Version2 (relative to $Version1): $Added new interface".(int($Attributes{"added"})==1?"":"s").", $Withdrawn withdrawn, $TotalProblemsHigh high severity problems, $TotalProblemsMedium medium and $TotalProblemsLow low<br/>\n";
            }
        }
        foreach my $Version (sort {cmpVersions($b, $a)} keys(%{$UpdateLog{$Library}{"TestRes"}}))
        {
            next if($Version eq "current");
            my %Attributes = %{$UpdateLog{$Library}{"TestRes"}{$Version}};
            my ($Total, $Failed) = ($Attributes{"total"}, $Attributes{"failed"});
            $Total="<a href='tests/$Library/$Version/view_tests.html\'>$Total</a>" if($Total);
            $Failed="<a href='test_results/$Library/$Version/test_results.html\'>$Failed</a>" if($Failed);
            $LogTest .= "<a href='versions/$Library.html'>".get_library_full_name($Library)."</a>-$Version: $Total total, $Failed failed<br/>\n";
        }
    }
    
    return if(not $LogNew and not $LogAdded and not $LogCompat and not $LogTest);
    
    $LogNew = "<h2 class='title2'>New Versions</h2><hr/>\n".$LogNew."<br/>" if($LogNew);
    $LogAdded = "<h2 class='title2'>Added Old Versions</h2><hr/>\n".$LogAdded."<br/>" if($LogAdded);
    $LogCompat = "<h2 class='title2'>ABI Changes</h2><hr/>\n".$LogCompat."<br/>" if($LogCompat);
    $LogTest = "<h2 class='title2'>API Shallow Tests</h2><hr/>\n".$LogTest."<br/>" if($LogTest);
    
    my $Log = "<h1 class='title1'><span style='color:Brown;'>Upstream Tracker</span> News</h1><br/>\n";
    $Log .= "<table cellpadding='0' cellspacing='0' style='padding-left:13px;width:100%;'><tr><td valign='top'>\n$LogNew$LogAdded$LogCompat$LogTest\n</td></tr></table>";
    my $Title = "upstream C/C++ libraries development news";
    my $Keywords = "upstream, news, shared, library, C, C++";
    writeFile("news.html", composeHTML_Head($Title, $Keywords, $LogStyles)."\n<body>\n<div>\n$Log</div>\n"."<br/><br/>$TOOL_SIGNATURE\n</body></html>");
    print "see results in the news.html\n";
}

sub composeHTML_Head($$$)
{
    my ($Title, $Keywords, $Styles) = @_;
    return composeHTML_Head_Descr($Title, $Keywords, "", $Styles, "");
}

sub composeHTML_Head_Descr($$$$$)
{
    my ($Title, $Keywords, $Description, $Styles, $JScript) = @_;
    return "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\" lang=\"en\">\n<head>
    <meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\" />
    <meta name=\"keywords\" content=\"$Keywords\" />".($Description?"\n    <meta name=\"description\" content=\"$Description\" />":"")."
    <title>\n        $Title\n    </title>\n    <style type=\"text/css\">\n$Styles\n    </style>\n    <script type=\"text/javascript\">\n    <!--\n$JScript\n    //-->\n    </script>\n</head>";
}

sub readAttributes($)
{
    my $Path = $_[0];
    return () if(not $Path or not -f $Path);
    my %Attributes = ();
    my $FirstLine = readFirstLine($Path);
    if($FirstLine=~/<!--\s+(.+)\s+-->/)
    {
        foreach my $AttrVal (split(/;/, $1))
        {
            if($AttrVal=~/(.+):(.+)/)
            {
                $Attributes{$1} = $2;
            }
        }
    }
    return %Attributes;
}

sub read_attrs($)
{
    my $Name = $_[0];
    my %LibVerAttr = ();
    foreach my $Line (split(/\n/, readFile("versions/$Name.txt")))
    {
        if($Line=~/\A([^;]+);([^;]+);([^;]+)\Z/)
        {
            $LibVerAttr{$1}{$2}=$3;
        }
    }
    return \%LibVerAttr;
}

sub add_attr($$$$)
{
    my ($Library, $Version, $Name, $Value) = @_;
    return if(not $Library or not $Version or not $Name or $Value eq "");
    writeFile("versions/$Name.txt", "") if(not -e "versions/$Name.txt");
    return if(get_attr($Library, $Version, $Name) eq $Value);
    $Cache{"get_attr"}{$Name}{$Library}{$Version} = $Value;
    my $AddCmd = "cat versions/$Name.txt|grep -v \'$Library;$Version;\' > versions/.$Name";
    $AddCmd .= " && echo \'$Library;$Version;$Value\' >> versions/.$Name";
    $AddCmd .= " && sort versions/.$Name > versions/$Name.txt";
    system($AddCmd);
}

sub get_attr($$$)
{
    my ($Library, $Version, $Name) = @_;
    return if(not $Library or not $Version or not $Name);
    return $Cache{"get_attr"}{$Name}{$Library}{$Version} if(defined $Cache{"get_attr"}{$Name}{$Library}{$Version});
    my $LibVerAttr = read_attrs($Name);
    $Cache{"get_attr"}{$Name} = $LibVerAttr;
    return $LibVerAttr->{$Library}{$Version};
}

sub createIndex()
{
    my (%TestResults, %Tests, %CompatReports, %LibVer) = ();
    foreach my $Path (cmd_find("compat_reports","f","abi_compat_report.html"))
    {
        my $ReportDir = get_Directory($Path);
        my $FromTo = get_FileName($ReportDir);
        my $Library = get_FileName(get_Directory($ReportDir));
        if($FromTo=~/\A(.+)_to_(.+)\Z/)
        {
            my ($Version1, $Version2) = ($1, $2);
            $CompatReports{$Library}{$Version1}{$Version2}{"path"} = $Path;
            $LibVer{$Library}{$Version1}=1;
            $LibVer{$Library}{$Version2}=1;
            my %Attr = readAttributes($Path);
            foreach my $Name (keys(%Attr))
            {
                $CompatReports{$Library}{$Version1}{$Version2}{$Name}=$Attr{$Name};
                if($ShowStat)
                {
                    $TestStat{$Name}+=$Attr{$Name};
                }
            }
        }
        if($ShowStat)
        {
            my $Content = readFile($Path);
            $Content=~s/\n//g;
            foreach my $Problem (split(/<span class='problem_body'>/, $Content))
            {
                $Problem=~s/<[^<>]*>//g;
                next if($Problem=~/ABI compliance report for the library/i);
                if($Problem=~/therefore the layout of virtual table has been changed/i)
                {
                    $TestStat{"vtable_changes"}+=1;
                }
                elsif($Problem=~/has been added to the interface signature/i)
                {
                    $TestStat{"added_parameter"}+=1;
                }
                elsif($Problem=~/has been withdrawn from the interface signature/i)
                {
                    $TestStat{"withdrawn_parameter"}+=1;
                }
                elsif(($Problem=~/type of (1st|2nd|3rd|\d+th) parameter/i and $Problem=~/ bytes\)/)
                or $Problem=~/type pointer level of (1st|2nd|3rd|\d+th) parameter/i
                or $Problem=~/member \w+ has been added to this type/i
                or $Problem=~/member \w+ has been added at the middle position of this structural type/i
                or $Problem=~/member \w+ has been withdrawn from this type/i
                or $Problem=~/member \w+ has been withdrawn from the middle position of this structural type/i
                or ($Problem=~/type of member/i and $Problem=~/ bytes\)/)
                or $Problem=~/type pointer level of member/i
                or $Problem=~/size of this type has been changed/i
                # compatibility with abi-compliance-checker<=1.8
                or $Problem=~/added between the first member and the last member/i
                or $Problem=~/from this structural type between the first member and the last member/i)
                {
                    $TestStat{"destructive_parameter_changes"}+=1;
                }
                elsif($Problem=~/type of (1st|2nd|3rd|\d+th) parameter/i and $Problem!~/ bytes\)/i
                or ($Problem=~/type of member/i and $Problem!~/ bytes\)/)
                or $Problem=~/has been renamed/)
                {
                    $TestStat{"non_destructive_parameter_changes"}+=1;
                }
                elsif(($Problem=~/type of return value has been changed/i and $Problem=~/ bytes\)/i)
                or $Problem=~/type pointer level of return value has been changed/i)
                {
                    $TestStat{"destructive_return_changes"}+=1;
                }
                elsif($Problem=~/type of return value has been changed/i and $Problem!~/ bytes\)/i
                or $Problem=~/base of this type has been changed/i)
                {
                    $TestStat{"non_destructive_return_changes"}+=1;
                }
                elsif($Problem=~/function become (non-|)static/i)
                {
                    $TestStat{"static_changes"}+=1;
                }
                elsif($Problem=~/has been redefined by/i)
                {
                    $TestStat{"virtual_overriding"}+=1;
                }
                elsif($Problem=~/value of member|name of member/i)
                {
                    $TestStat{"enum_changes"}+=1;
                }
                elsif($Problem=~/value of constant/i)
                {
                    $TestStat{"macros_changes"}+=1;
                }
            }
        }
    }
    
    foreach my $Path (cmd_find("test_results","f","test_results.html"))
    {
        my $ReportDir = get_Directory($Path);
        my $Version = get_FileName($ReportDir);
        my $Library = get_FileName(get_Directory($ReportDir));
        $TestResults{$Library}{$Version}{"path"} = $Path;
        $LibVer{$Library}{$Version}=1;
        my %Attr = readAttributes($Path);
        foreach my $Name (keys(%Attr))
        {
            $TestResults{$Library}{$Version}{$Name}=$Attr{$Name};
        }
    }
    if(-d "tests")
    {
        foreach my $Path (split(/\n/,`find tests/ -mindepth 3 -maxdepth 3 -name "view_tests.html" -type f`))
        {
            my $ReportDir = get_Directory($Path);
            my $Version = get_FileName($ReportDir);
            my $Library = get_FileName(get_Directory($ReportDir));
            $Tests{$Library}{$Version}{"path"} = $Path;
            $LibVer{$Library}{$Version}=1;
            my %Attr = readAttributes($Path);
            foreach my $Name (keys(%Attr))
            {
                $Tests{$Library}{$Version}{$Name}=$Attr{$Name};
            }
        }
    }
    
    my $StatStyles = "    body{font-family:Arial;font-size:15px;background:#ffffff;color:Black;}
    hr{color:Black;background-color:Black;height:1px;border:0;}
    h1.title1{margin-bottom:0px;padding-bottom:0px;font-size:26px;margin-top:0px;padding-top:2px;}
    h2.title2{margin-bottom:0px;padding-bottom:0px;font-size:20px;margin-top:0px;padding-top:0px;}
    table.wikitable{border-collapse:collapse;margin:1em 1em 1em 0;background:#f9f9f9;border:1px #aaaaaa solid;}
    table.wikitable td{border:1px #aaaaaa solid;text-align:center;}
    table.wikitable th{background:#eeeeee;border:1px #aaaaaa solid;padding:0.2em;padding-left:0.4em;padding-right:0.4em;}
    table.wikitable td.left{text-align:left;background-color:#f2f2f2;}
    table.wikitable td.value{padding-left:5px;padding-right:5px;width:35px;font-size:15px;}
    table.wikitable td.destr{padding:5px;width:35px;font-size:16px;background-color:#f2f2f2;}";
    
    my $VListStyles = "    body{font-family:Arial;font-size:12px;background:#ffffff;color:Black;}
    hr{color:Black;background-color:Black;height:1px;border:0;}
    h1.title1{margin-bottom:0px;padding-bottom:0px;font-size:26px;margin-top:0px;padding-top:2px;}
    h2.title2{margin-bottom:0px;padding-bottom:0px;font-size:20px;margin-top:0px;padding-top:0px;}
    table.wikitable{border-collapse:collapse;margin:1em 1em 1em 0;background:#f9f9f9;border:1px #aaaaaa solid;}
    table.wikitable td{border:1px #aaaaaa solid;text-align:center;}
    table.wikitable th{background:#f2f2f2;border:1px #aaaaaa solid;padding:0.2em;padding-left:0.4em;padding-right:0.4em;}
    table.wikitable th.left{text-align:left;}
    td.passed{background-color:#CCFFCC;}
    td.warning{background-color:#F4F4AF;}
    td.failed{background-color:#FFC3CE;}
    a.default{color:#336699;}
    a.passed{color:Green;}
    a.failed{color:Red;}
    a.changelog{color:#336699;}
    sup{font-size:12px;}
    span.date{font-weight:100;font-size:10px;}";

    my $IndexStyles = "    body{font-family:Arial;font-size:14px;background:#ffffff;color:Black;}
    hr{color:Black;background-color:Black;height:1px;border:0;}
    h1.title1{margin-bottom:0px;padding-bottom:0px;font-size:26px;margin-top:0px;padding-top:0px;}
    h2.title2{margin-bottom:0px;padding-bottom:0px;font-size:20px;margin-top:0px;padding-top:0px;}
    table.wikitable{border-collapse:collapse;margin:1em 1em 1em 0;background:#f9f9f9;border:1px #aaaaaa solid;}
    table.wikitable td{border:1px #aaaaaa solid;text-align:center;padding:0.2em;padding-left:5px;padding-right:5px;}
    table.wikitable th{background:#f2f2f2;border:1px #aaaaaa solid;padding:0.2em;padding-left:0.4em;padding-right:0.4em;}
    table.wikitable th.left{text-align:left;background-color:#ffffff;}
    a.link{color:#336699;}
    td.summary{background-color:#F4F4AF;}
    td.results{background-color:#ffffff;}
    sup{font-size:10px;}";
    
    my $IndexJScript = "    function show_item(item){
        item.style.display = '';
        item.style.visibility = 'visible';
    }
    function hide_item(item){
        item.style.display = 'none';
        item.style.visibility = 'hidden';
    }
    function applyFilter(filter, table_id){
        var table = document.getElementById(table_id);
        var items = table.getElementsByTagName('tr');
        var pattern = filter.value.toLowerCase();
        pattern = pattern.replace(/\\+/gi, '\\\\+');
        var count = 0;
        for (var i in items)
        {
            if(isNaN(parseInt(i)))
                continue;
            if(items[i].id=='listHeader1' || items[i].id=='listHeader2')
                continue;
            var text = items[i].title.toLowerCase();
            if(pattern=='' || text.match(pattern)){
                show_item(items[i]);
                count += 1;
            }
            else{
                hide_item(items[i]);
            }
        }
        var listHeader1 = document.getElementById('listHeader1');
        var listHeader2 = document.getElementById('listHeader2');
        
        var counter1 = listHeader1.getElementsByTagName('sup')[0];
        var counter2 = listHeader2.getElementsByTagName('sup')[0];
        
        counter1.innerHTML = count;
        counter2.innerHTML = count;
        
        show_item(listHeader1);
        if(count>=10){
            show_item(listHeader2);
            show_item(counter1);
        }
        else if(count==1){
            hide_item(listHeader2);
            hide_item(counter1);
        }
        else if(count==0){
            hide_item(listHeader1);
            hide_item(listHeader2);
            hide_item(counter1);
        }
        else{
            hide_item(listHeader2);
            show_item(counter1);
        }
    }";
    
    if($ShowStat)
    {# visualizing
        foreach my $Library (keys(%LibVer))
        {
            my @Versions = sort_VersionSequence(keys(%{$LibVer{$Library}}), $Library);
            if($#Versions>=0)
            {# detect language (C or C++) of the latest library version
                $TestStat{detect_language($Library, $Versions[0])}+=1;
            }
        }
        
        $TestStat{"total"} = $TestStat{"macros_changes"}+$TestStat{"enum_changes"}+$TestStat{"virtual_overriding"}+
        $TestStat{"static_changes"}+$TestStat{"non_destructive_return_changes"}+$TestStat{"destructive_return_changes"}+
        $TestStat{"non_destructive_parameter_changes"}+$TestStat{"destructive_parameter_changes"}+$TestStat{"withdrawn_parameter"}+
        $TestStat{"added_parameter"}+$TestStat{"vtable_changes"}+$TestStat{"withdrawn"};
        my $TotalVersions = 0;
        foreach my $LName (keys(%LibVer))
        {
            $TotalVersions+=keys(%{$LibVer{$LName}});
        }
        
        my $VisualStat = "<a style='font-size:11px;color:#336699;' href=\'index.html\'>to main page</a><h1 class='title1'>&nbsp;Statistics of binary changes analysis<br/>&nbsp;&nbsp;for the shared <span style='color:Blue;'>C/C++</span> libraries on <span style='color:Blue;'>x86</span></h1>";
        my $VisualStatSummary = "<table cellpadding='2' class='wikitable'>";
        $VisualStatSummary .= "<tr><td class='left'>Total libraries</td><td class='value'>".keys(%LibVer)."</td></tr>";
        $VisualStatSummary .= "<tr><td class='left'>C++/C language</td><td class='value'>".cut_off_number($TestStat{"C++"}*100/($TestStat{"C"}+$TestStat{"C++"}), 2)."%&nbsp;/&nbsp;".cut_off_number($TestStat{"C"}*100/($TestStat{"C"}+$TestStat{"C++"}), 2)."%</td></tr>";
        $VisualStatSummary .= "<tr><td class='left'>Total versions</td><td class='value'>".$TotalVersions."</td></tr>";
        $VisualStatSummary .= "<tr><td class='left'>Total problems</td><td class='value'>".$TestStat{"total"}."</td></tr></table>";
        my $VisualStatList = "<table cellpadding='2' class='wikitable' style='width:470px;'><tr><th colspan='2'>Kind of Changes</th><th>Severity</th><th>Count</th><th>Percentage</th></tr>";
        if(my $Withdrawn = $TestStat{"withdrawn"})
        {
            $VisualStatList .= "<tr><td colspan='2' class='left'>Removing functions from the library</td><td class='value'>High</td><td class='value'>$Withdrawn</td><td class='value'>".(cut_off_number($Withdrawn*100/$TestStat{"total"}, 3))."%</td></tr>";
        }
        if(my $VTableChg = $TestStat{"vtable_changes"})
        {
            $VisualStatList .= "<tr><td colspan='2' class='left'>Changing virtual table structure</td><td class='value'>High</td><td class='value'>$VTableChg</td><td class='value'>".(cut_off_number($VTableChg*100/$TestStat{"total"}, 3))."%</td></tr>";
        }
        if(my $ParamsChg = $TestStat{"added_parameter"}+$TestStat{"withdrawn_parameter"})
        {
            $VisualStatList .= "<tr><td colspan='2' class='left'>Changing the number or order of parameters</td><td class='value'>Medium</td><td class='value'>$ParamsChg</td><td class='value'>".(cut_off_number($ParamsChg*100/$TestStat{"total"}, 3))."%</td></tr>";
        }
        if(my $ParamsChg_Destr = $TestStat{"destructive_parameter_changes"}
        and my $ParamsChg_NonDestr = $TestStat{"non_destructive_parameter_changes"})
        {
            $VisualStatList .= "<tr><td rowspan='2' class='left'>Changing type of a parameter</td><td class='destr'>destructive<sup>*</sup></td><td class='value'>Medium</td><td class='value'>$ParamsChg_Destr</td><td class='value'>".(cut_off_number($ParamsChg_Destr*100/$TestStat{"total"}, 3))."%</td></tr>";
            $VisualStatList .= "<tr><td class='destr'>non-destructive</td><td>Low</td><td>$ParamsChg_NonDestr</td><td>".(cut_off_number($ParamsChg_NonDestr*100/$TestStat{"total"}, 3))."%</td></tr>";
        }
        if(my $ReturnChg_Destr = $TestStat{"destructive_return_changes"}
        and my $ReturnChg_NonDestr = $TestStat{"non_destructive_return_changes"})
        {
            $VisualStatList .= "<tr><td rowspan='2' class='left'>Changing type of return value</td><td class='destr'>destructive</td><td class='value'>Medium</td><td class='value'>$ReturnChg_Destr</td><td class='value'>".(cut_off_number($ReturnChg_Destr*100/$TestStat{"total"}, 3))."%</td></tr>";
            $VisualStatList .= "<tr><td class='destr'>non-destructive</td><td>Low</td><td>$ReturnChg_NonDestr</td><td>".(cut_off_number($ReturnChg_NonDestr*100/$TestStat{"total"}, 3))."%</td></tr>";
        }
        if(my $StaticChg = $TestStat{"static_changes"})
        {
            $VisualStatList .= "<tr><td colspan='2' class='left'>Adding/removing “static” specifier of a method</td><td class='value'>Medium</td><td class='value'>$StaticChg</td><td class='value'>".(cut_off_number($StaticChg*100/$TestStat{"total"}, 3))."%</td></tr>";
        }
        if(my $MacrosChg = $TestStat{"macros_changes"}+$TestStat{"enum_changes"})
        {
            $VisualStatList .= "<tr><td colspan='2' class='left'>Changing Values in Enumeration Types or Macros</td><td class='value'>Low</td><td class='value'>$MacrosChg</td><td class='value'>".(cut_off_number($MacrosChg*100/$TestStat{"total"}, 3))."%</td></tr>";
        }
        if(my $VirtualOverriding = $TestStat{"virtual_overriding"})
        {
            $VisualStatList .= "<tr><td colspan='2' class='left'>Overriding Virtual Functions</td><td class='value'>Low</td><td class='value'>$VirtualOverriding</td><td class='value'>".(cut_off_number($VirtualOverriding*100/$TestStat{"total"}, 3))."%</td></tr>";
        }
        $VisualStatList .= "</table><sup>*</sup> changes that affect the size or internal structure of a type.";
        $VisualStat .= "<table cellpadding='0' cellspacing='0' style='padding-left:13px;'><tr><td valign='top'>$VisualStatSummary$VisualStatList</td></tr></table>";
        my $Title = "Statistics of binary changes analysis for the shared C/C++ libraries on x86";
        my $Keywords = "C, C++, ABI, change, library, shared, binary, statistics";
        writeFile("statistics.html", composeHTML_Head($Title, $Keywords, $StatStyles)."\n<body>\n<div>\n$VisualStat</div>\n"."<br/><br/>$TOOL_SIGNATURE\n</body></html>");
        print "see statistics of binary changes analysis in the statistics.html\n";
    }
    
    my $LibraryList = "";
    foreach my $Library (sort {lc(get_library_full_name($a)) cmp lc(get_library_full_name($b))} keys(%LibVer))
    {
        my $VersionList = "<table cellpadding='0' cellspacing='0' style='padding-left:11px;'><tr><td valign='top'><table class='wikitable'><tr><th rowspan='2'>Version</th><th rowspan='2'>Soname<br/>Version</th><th colspan='2'>Runtime Unit Tests</th><th rowspan='2'>T2C<br/>Tests</th><th rowspan='2'>Change<br/>Log</th><th rowspan='2'>Added<br/>Interfaces</th><th rowspan='2'>Withdrawn<br/>Interfaces</th><th colspan='3' style='white-space:nowrap;'>Backward Binary Compatibility Problems</th></tr><tr><th>Total</th><th>Failed</th><th style='width:80px;color:Olive;'>High</th><th style='width:80px;color:Olive;'>Medium</th><th style='width:90px;color:Olive;'>Low</th></tr>";
        my @Versions = sort_VersionSequence(keys(%{$LibVer{$Library}}), $Library);
        foreach my $Num (0 .. $#Versions)
        {# version entry
            my $Version = $Versions[$Num];
            my $PreVersion = ($Num<$#Versions)?$Versions[$Num+1]:"";
            $VersionList.="<tr><th class='left'>$Version".(($Version eq "current" and $LibraryInfo{$Library}{"Scm"})?" (".$LibraryInfo{$Library}{"Scm"}.")<br/><span class='date'>".get_attr($Library, "current", "dates")."</span>":"")."</th>";
            $VersionList.="<td>".get_soname_version($Library, $Version)."</td>";
            my $RuntimeTests="";
            if($Tests{$Library}{$Version}{"total"})
            {
                $RuntimeTests="<a href='../".$Tests{$Library}{$Version}{"path"}."' class='default'>".$Tests{$Library}{$Version}{"total"}."</a>";
            }
            if($RuntimeTests)
            {
                $VersionList.="<td>$RuntimeTests</td>";
            }
            else
            {
                $VersionList.="<td></td>";
            }
            my $FailedTests="";
            if($TestResults{$Library}{$Version}{"failed"})
            {
                $FailedTests="<a href='../".$TestResults{$Library}{$Version}{"path"}."' class='default'>".$TestResults{$Library}{$Version}{"failed"}."</a>";
            }
#             my $Warnings="";
#             if(my $WCount = $TestResults{$Library}{$Version}{"warnings"})
#             {
#                 $Warnings="<a href='../".$TestResults{$Library}{$Version}{"path"}."' class='default'>$WCount warning".($WCount==1?"":"s")."</a>";
#             }
            if($FailedTests)
            {
                $VersionList.="<td class='failed'>$FailedTests</td>";
            }
#             elsif($Warnings)
#             {
#                 $VersionList.="<td class='warning'>&nbsp;0&nbsp;/&nbsp;$Warnings&nbsp;</td>";
#             }
            elsif(not defined $TestResults{$Library}{$Version}{"failed"})
            {
                $VersionList.="<td></td>";
            }
            else
            {
                $VersionList.="<td class='passed'>0</td>";
            }
            my $T2C_Ref="";
            if(-f "tests_t2c/$Library/$Version/t2c-$Library-tests-$Version.tar.gz")
            {
                $T2C_Ref = "&nbsp;&nbsp;&nbsp;<a class='changelog' href=\'../tests_t2c/$Library/$Version/t2c-$Library-tests-$Version.tar.gz\'>tar.gz</a>&nbsp;&nbsp;&nbsp;";
            }
            if($T2C_Ref)
            {
                $VersionList.="<td>$T2C_Ref</td>";
            }
            else
            {
                $VersionList.="<td></td>";
            }
            my $ChangeLog="";
            if(my $Link = readFile("changelogs/$Library/$Version/link"))
            {
                $ChangeLog="&nbsp;&nbsp;<a href=\'$Link\' class='changelog'>changelog</a>&nbsp;&nbsp;";
            }
            elsif(-f "changelogs/$Library/$Version/changelog.html")
            {
                $ChangeLog="&nbsp;&nbsp;<a href='../changelogs/$Library/$Version/changelog.html' class='changelog'>changelog</a>&nbsp;&nbsp;";
            }
            if($ChangeLog)
            {
                $VersionList.="<td>$ChangeLog</td>";
            }
            else
            {
                $VersionList.="<td></td>";
            }
            my %CompatAttr = (defined $CompatReports{$Library}{$PreVersion}{$Version})?%{$CompatReports{$Library}{$PreVersion}{$Version}}:();
            my $Added="";
            if(my $Count = $CompatAttr{"added"})
            {
                $Added="<a href='../".$CompatAttr{"path"}."#Added' class='default'>$Count new</a>";
            }
            if($Added)
            {
                $VersionList.="<td class='warning'>$Added</td>";
            }
            elsif(not defined $CompatAttr{"added"})
            {
                $VersionList.="<td></td>" if($Num!=$#Versions);
            }
            else
            {
                $VersionList.="<td class='passed'>0</td>";
            }
            my $Withdrawn="";
            if(my $Count = $CompatAttr{"withdrawn"})
            {
                $Withdrawn="<a href='../".$CompatAttr{"path"}."#Withdrawn' class='default'>$Count withdrawn</a>";
            }
            if($Withdrawn)
            {
                $VersionList.="<td class='failed'>$Withdrawn</td>";
            }
            elsif(not defined $CompatAttr{"withdrawn"})
            {
                $VersionList.="<td></td>" if($Num!=$#Versions);
            }
            else
            {
                $VersionList.="<td class='passed'>0</td>";
            }
            my $High="";
            if(my $Count = $CompatAttr{"type_problems_high"}+$CompatAttr{"interface_problems_high"})
            {
                $High="<a href='../".$CompatAttr{"path"}."' class='default'>".problem_title($Count)."</a>";
            }
            if($High)
            {
                $VersionList.="<td class='failed'>$High</td>";
            }
            elsif(not defined $CompatAttr{"type_problems_high"}
            and not defined $CompatAttr{"interface_problems_high"})
            {
                $VersionList.="<td></td>" if($Num!=$#Versions);
            }
            else
            {
                $VersionList.="<td class='passed'>0</td>";
            }
            my $Medium="";
            if(my $Count = $CompatAttr{"type_problems_medium"}+$CompatAttr{"interface_problems_medium"})
            {
                $Medium="<a href='../".$CompatAttr{"path"}."' class='default'>".problem_title($Count)."</a>";
            }
            if($Medium)
            {
                $VersionList.="<td class='failed'>$Medium</td>" ;
            }
            elsif(not defined $CompatAttr{"type_problems_medium"}
            and not defined $CompatAttr{"interface_problems_medium"})
            {
                $VersionList.="<td></td>" if($Num!=$#Versions);
            }
            else
            {
                $VersionList.="<td class='passed'>0</td>";
            }
            my $Low="";
            if(my $Count = $CompatAttr{"type_problems_low"}+$CompatAttr{"interface_problems_low"}+$CompatAttr{"changed_constants"})
            {
                $Low="<a href='../".$CompatAttr{"path"}."' class='default'>".warning_title($Count)."</a>";
            }
            if($Low)
            {
                $VersionList.="<td class='warning'>$Low</td>";
            }
            elsif(not defined $CompatAttr{"type_problems_low"}
            and not defined $CompatAttr{"interface_problems_low"}
            and not defined $CompatAttr{"changed_constants"})
            {
                $VersionList.="<td></td>" if($Num!=$#Versions);
            }
            else
            {
                $VersionList.="<td class='passed'>0</td>";
            }
            if($Num==$#Versions)
            {
                $VersionList.="<td colspan='5' align='center' style='color:#bbbbbb'><b>initial version</b></td>";
            }
            $VersionList.="</tr>\n";
        }
        if($#Versions>=10)
        {# bottom header
            $VersionList .= "<tr><th rowspan='2'>Version</th><th rowspan='2'>Soname<br/>Version</th><th>Total</th><th>Failed</th><th rowspan='2'>T2C<br/>Tests</th><th rowspan='2'>Change<br/>Log</th><th rowspan='2'>Added<br/>Interfaces</th><th rowspan='2'>Withdrawn<br/>Interfaces</th><th style='width:80px;color:Olive;'>High</th><th style='width:80px;color:Olive;'>Medium</th><th style='width:90px;color:Olive;'>Low</th></tr><tr><th colspan='2'>Runtime Unit Tests</th><th colspan='3' style='white-space:nowrap;'>Backward Binary Compatibility Problems</th></tr>";
        }
        $VersionList .= "</table></td>".(($#Versions>=1)?"<td valign='top' style='padding-top:50px;padding-left:2px;padding-right:1px;font-size:20px'>&#8595;</td><td valign='top' style='width:130px;padding-top:55px;'><span style='font-family:Verdana;'>Analysis of each version has been done <b>relative to the previous version</b> in the list</span></td>\n":"")."</tr></table>\n";
        $VersionList = "<a style='font-size:11px;color:#336699;' href=\'../index.html\'>list of libraries</a><h1 class='title1'>&nbsp;Test results and ABI changes analysis<br/>&nbsp;&nbsp;for the library <span style='color:Blue;'>".get_library_full_name($Library)."</span> on <span style='color:Blue;'>x86</span></h1>\n".$VersionList;
        my $Title = "Test results and ABI changes analysis for the library ".get_library_full_name($Library)." on x86";
        my $Keywords = get_library_full_name($Library).", library, ABI, change, runtime, test";
        writeFile("versions/$Library.html", composeHTML_Head($Title, $Keywords, $VListStyles)."\n<body>\n<div><a name='Top'></a>\n$VersionList</div>\n"."<br/><br/>$TOOL_SIGNATURE\n</body></html>");
        # libraries list
        my $Library_Id = get_library_full_name($Library);
        $Library_Id=~s/\+\+/pp/g;
        $Library_Id=~s/\+//g;
        $Library_Id=~s/[\s()]+\Z//g;
        $Library_Id=~s/[\s()]+/_/g;
        $LibraryList .= "<tr id=\'$Library_Id\' title='".get_library_full_name($Library)."'><th class='left'>".get_library_full_name($Library);
        if(my $Site = $LibraryInfo{$Library}{"Site"})
        {
            $LibraryList .= "<sup style='padding-left:3px;'><a class='link' href=\'$Site\'>site</a></sup>";
        }
        if(my $Doc = $LibraryInfo{$Library}{"Doc"})
        {
            $LibraryList .= "<sup style='padding-left:3px;'><a class='link' style='color:Green;' href=\'$Doc\'>doc</a></sup>";
        }
        $LibraryList .= "</th>";
        $LibraryList .= "<td width='215px' class='summary' style='text-align:left;'>".$LibraryInfo{$Library}{"Summary"}."</td>";
        if($#Versions>-1)
        {
            $LibraryList .= "<td class='results'><a class='link' href='versions/$Library.html'>".version_title($#Versions+1)."</a></td>";
        }
        else
        {
            $LibraryList .= "<td class='results'></td>";
        }
        $LibraryList .= "</tr>\n";
    }
    my $HideCounter = (keys(%LibVer)>1?"":" style='display:none;visibility:hidden;'");
    my $HideHeader1 = (keys(%LibVer)>0?"":" style='display:none;visibility:hidden;'");
    my $ListHeader1 = "<tr id='listHeader1'$HideHeader1><th>Library&nbsp;<sup$HideCounter>".keys(%LibVer)."</sup></th><th>Summary</th><th>Test Results</th></tr>\n";
    my $HideHeader2 = (keys(%LibVer)>=10?"":" style='display:none;visibility:hidden;'");
    my $ListHeader2 = "<tr id='listHeader2'$HideHeader2><th>Library&nbsp;<sup$HideCounter>".keys(%LibVer)."</sup></th><th>Summary</th><th>Test Results</th></tr>\n";
    my $StatPart = (-f "statistics.html")?"\n<br/>&nbsp;&nbsp;&nbsp;&nbsp;See statistics of binary changes analysis <a href='statistics.html'>here</a>.":"";
    my $NewsPart = (-f "news.html")?"\n<br/>&nbsp;&nbsp;&nbsp;&nbsp;See latest DB update <a href='news.html'>here</a>.":"";
    $LibraryList = "<h1 class='title1'>&nbsp;<span style='color:Brown;'>Upstream Tracker</span><br/>&nbsp;Test results and ABI changes analysis<br/>&nbsp;&nbsp;for the shared <span style='color:Blue;'>C/C++</span> libraries on <span style='color:Blue;'>x86</span></h1>$StatPart$NewsPart
<br/><br/>&nbsp;&nbsp;&nbsp;&nbsp;Filter by name: <input type='text' name='filter' onkeyup=\"javascript:applyFilter(this, 'listOfLibraries')\" onclick=\"javascript:applyFilter(this, 'listOfLibraries')\" />\n<table id='listOfLibraries' class='wikitable' style='margin-left:11px;'>".$ListHeader1.$LibraryList;
    $LibraryList .= $ListHeader2;
    $LibraryList .= "</table>\n";
    my $Title = "Upstream Tracker - List of C/C++ libraries";
    my $Keywords = "ABI, change, shared, library, upstream, C, runtime, test, C++";
    my $Description = "Test results and ABI changes analysis for the shared C/C++ libraries on x86";
    writeFile("index.html", composeHTML_Head_Descr($Title, $Keywords, $Description, $IndexStyles, $IndexJScript)."\n<body>\n<div><a name='Top'></a>\n$LibraryList</div>\n"."<br/><br/>$TOOL_SIGNATURE\n</body></html>");
    if(-f "index.html")  {
        print "index.html has been updated\n";
    }
    else  {
        print "index.html has been generated\n";
    }
}

sub interface_title($)
{
    if($_[0]==1)  {
        return "interface";
    }
    else  {
        return "interfaces";
    }
}

sub problem_title($)
{
    if($_[0]==1)  {
        return "1 problem";
    }
    else  {
        return $_[0]." problems";
    }
}

sub warning_title($)
{
    if($_[0]==1)  {
        return "1 warning";
    }
    else  {
        return $_[0]." warnings";
    }
}

sub version_title($)
{
    if($_[0]==1)  {
        return "1 version";
    }
    else  {
        return $_[0]." versions";
    }
}

sub isCheckouted($)
{
    my $Library = $_[0];
    return 0 if(not $Library);
    my $SourcePath = "$REPOSITORY_PATH/$Library/$Library-source";
    return (-d $SourcePath and split("\n", `find $SourcePath -maxdepth 1 -type f`));
}

sub getLibraryPrefix($)
{
    my $Library = $_[0];
    return 0 if(not $Library);
    foreach my $PackageTmpl (keys(%{$LibraryInfo{$Library}{"Package"}}))
    {
        if(my ($LinkPrefix, $LinkSuffix) = split(/\{VERSION\}/, $PackageTmpl))
        {
            $LinkPrefix=~s/[-_]+\Z//g;
            return $LinkPrefix;
        }
    }
    return $Library;
}

sub getCurrentVersion_Library($)
{
    my $Library = $_[0];
    return if(not $Library or not $LibraryInfo{$Library}{"Scm"});
    my $SourcePath = "$REPOSITORY_PATH/$Library/$Library-source";
    my $Date = getDate(time);
    my $DateStamp = getDateStamp($Date);
    if(isCheckouted($Library))
    {
        if(updateSource($Library) and -d $SourcePath)
        {
            $RepositoryProvide{$Library}{"current"}="current";
            unlink($VersionPath{$Library}{"current"}); # remove old snapshot
            $UpdateLog{$Library}{"NewSrc"}=1;
            $VersionPath{$Library}{"current"}=create_Package($SourcePath, getLibraryPrefix($Library)."-current-$DateStamp", "tar.gz");
        }
    }
    else
    {
        if(checkoutSource($Library) and -d $SourcePath)
        {
            $RepositoryProvide{$Library}{"current"}="current";
            unlink($VersionPath{$Library}{"current"}); # remove old snapshot
            $UpdateLog{$Library}{"NewSrc"}=1;
            $VersionPath{$Library}{"current"}=create_Package($SourcePath, getLibraryPrefix($Library)."-current-$DateStamp", "tar.gz");
        }
    }
}

sub create_Package($$$)
{
    my ($SrcPath, $To, $Format) = @_;
    return 0 if(not $SrcPath or not $Format);
    my $SrcDir = get_Directory($SrcPath);
    my $From = get_FileName($SrcPath);
    my $ResPath = "$SrcDir/$To.$Format";
    if(-e $ResPath)
    {
        unlink($ResPath);
    }
    my $ChangeDir = "";
    if($SrcDir)
    {
        $ChangeDir = "cd $SrcDir &&";
    }
    if($Format eq "tar.gz")
    {
        `$ChangeDir tar -cf $To.tar $From`;
        `$ChangeDir gzip $To.tar --best`;
    }
    elsif($Format eq "tar.bz2")
    {
        `$ChangeDir tar -cf $To.tar $From`;
        `$ChangeDir bzip2 $To.tar`;
    }
    elsif($Format eq "tgz")
    {
        `$ChangeDir tar -cf $To.tar $From`;
        `$ChangeDir gzip $To.tar --best`;
        `$ChangeDir mv $To.tar.gz $To.tgz`;
    }
    elsif($Format eq "tar.Z")
    {
        `$ChangeDir tar -cf $To.tar $From`;
        `$ChangeDir gzip $To.tar --best`;
        `$ChangeDir mv $To.tar.gz $To.tar.Z`;
    }
    elsif($Format eq "zip")
    {
        `$ChangeDir zip -r $To.zip $From`;
    }
    return $ResPath;
}

sub updateSource($)
{
    my $Library = $_[0];
    return if(not $Library);
    my $Scm = $LibraryInfo{$Library}{"Scm"};
    my $ScmAddr = $LibraryInfo{$Library}{"ScmAddr"};
    my $ScmModule = $LibraryInfo{$Library}{"ScmModule"};
    return if(not $ScmAddr or not $Scm);
    if(not check_scm($Library))
    {
        print STDERR "ERROR: source repository for $Library is not available at the moment ($ScmAddr)\n";
        return 0;
    }
    my $SourcePath = "$REPOSITORY_PATH/$Library/$Library-source";
    mkpath($SourcePath);
    print "updating $Library source from $ScmAddr\n";
    if($Scm eq "cvs")
    {
        system("cd $SourcePath && cvs update -d 2>/dev/null");
    }
    elsif($Scm eq "svn")
    {
        system("cd $SourcePath && svn update");
    }
    elsif($Scm eq "rsync")
    {
        system("cd $SourcePath && rsync -avz rsync://$ScmAddr .");
    }
    elsif($Scm eq "git")
    {
        system("cd $SourcePath && git reset --hard HEAD && git pull");
    }
    elsif($Scm eq "bazaar")
    {
        system("cd $SourcePath && bzr pull");
    }
    elsif($Scm eq "mercurial")
    {
        system("cd $SourcePath && hg pull");
    }
    elsif($Scm eq "darcs")
    {
        system("cd $SourcePath && darcs pull");
    }
    else
    {
        return 0;
    }
    if($?)
    {
        print STDERR "ERROR: can't update $Library source from $ScmAddr\n";
        return 0;
    }
    return 1;
}

sub firstDir_Path($)
{
    return $_[0] if($_[0]!~/\//);
    if($_[0]=~/\A([^\/]*)\//)
    {
        return $1;
    }
    else
    {
        return $_[0];
    }
}

sub interactive_Cmd($$)
{
    my ($Cmd, $Password) = @_;
    return if(not $Cmd);
    my $TransferCmd = "expect <<TRANSFER 
spawn bash -c \"$Cmd\"
while 1 {
    expect {
    \"(yes/no)?\" {send \"yes\\n\"}
    \"Y/n/?\" {send \"Y\\n\"}
    \"N/y/?\" {send \"N\\n\"}
    \"assword\" {send \"$Password\\n\"}
    \"Do you accept\" {send \"yes\\n\"}
    eof exit
    }
}
TRANSFER\n";#\"\" {send \"\\n\"}
    writeFile(".tmp_dir/transfer.sh", $TransferCmd);
    system("sh .tmp_dir/transfer.sh");
    unlink(".tmp_dir/transfer.sh");
}

sub checkoutSource($)
{
    my $Library = $_[0];
    return if(not $Library);
    my $Scm = $LibraryInfo{$Library}{"Scm"};
    my $ScmAddr = $LibraryInfo{$Library}{"ScmAddr"};
    my $ScmModule = $LibraryInfo{$Library}{"ScmModule"};
    my $ScmBranch = $LibraryInfo{$Library}{"ScmBranch"};
    my $ScmPassword = $LibraryInfo{$Library}{"ScmPassword"};
    return if(not $ScmAddr or not $Scm);
    if(not check_scm($Library))
    {
        print STDERR "ERROR: source repository for $Library is not available at the moment ($ScmAddr)\n";
        return 0;
    }
    print "checkouting $Library source from $ScmAddr\n";
    my $SourcePath = "$REPOSITORY_PATH/$Library/$Library-source";
    my $SourceDir = "$Library-source";
    my $LibraryDir = "$REPOSITORY_PATH/$Library";
    rmtree($SourcePath);
    if($Scm eq "cvs")
    {
        return if(not $ScmModule);
        mkpath($SourcePath);
        mkpath($LibraryDir."/".$ScmModule);
        my $Cmd = "cd $LibraryDir && cvs -z3 -d:pserver:$ScmAddr >/dev/null 2>&1";
        if($ScmPassword)
        {
            interactive_Cmd($Cmd." login", $ScmPassword);
        }
        system($Cmd." checkout ".($ScmBranch?"-r$ScmBranch ":"")."$ScmModule");
        move($LibraryDir."/".$ScmModule, $SourcePath) if(-d "$LibraryDir/$ScmModule");
    }
    elsif($Scm eq "svn")
    {
        if($ScmAddr=~/http\:/ or $ScmAddr=~/https\:/ or $ScmAddr=~/svn\:/)
        {
            system("cd $LibraryDir && svn checkout $ScmAddr $SourceDir -rHEAD >/dev/null 2>&1");
        }
        else
        {
            system("cd $LibraryDir && svn checkout svn://$ScmAddr $SourceDir >/dev/null 2>&1");
        }
    }
    elsif($Scm eq "rsync")
    {
        if($ScmAddr=~/http\:/ or $ScmAddr=~/https\:/ or $ScmAddr=~/rsync\:/)
        {
            system("cd $SourcePath && rsync -avz $ScmAddr . >/dev/null 2>&1");
        }
        else
        {
            system("cd $SourcePath && rsync -avz rsync://$ScmAddr . >/dev/null 2>&1");
        }
    }
    elsif($Scm eq "git")
    {
        if($ScmAddr=~/http\:/ or $ScmAddr=~/https\:/ or $ScmAddr=~/git\:/)
        {
            system("cd $LibraryDir && git clone $ScmAddr $SourceDir >/dev/null 2>&1");
        }
        else
        {
            system("cd $LibraryDir && git clone git://$ScmAddr $SourceDir >/dev/null 2>&1");
        }
    }
    elsif($Scm eq "bazaar")
    {
        system("cd $LibraryDir && bzr checkout $ScmAddr $SourceDir >/dev/null 2>&1");
    }
    elsif($Scm eq "mercurial")
    {
        system("cd $LibraryDir && hg clone $ScmAddr $SourceDir >/dev/null 2>&1");
    }
    elsif($Scm eq "darcs")
    {
        system("cd $LibraryDir && darcs get $ScmAddr $SourceDir >/dev/null 2>&1");
    }
    else
    {
        return 0;
    }
    if($?)
    {
        print STDERR "ERROR: can't checkout $Library from $ScmAddr\n";
        return 0;
    }
    return 1;
}

sub check_scm($)
{
    my $Library = $_[0];
    return 0 if(not $Library);
    my $Scm = $LibraryInfo{$Library}{"Scm"};
    my $ScmAddr = $LibraryInfo{$Library}{"ScmAddr"};
    if($Scm eq "cvs")
    {
        $ScmAddr=~s/\A(.*)\@//g;
        $ScmAddr=~s/:\/(.*)\Z//g;
    }
    $ScmAddr = getSiteAddr($ScmAddr);
    return check_addr($ScmAddr);
}

sub getChangeLog_SCM($)
{
    my $Library = $_[0];
    return "" if(not $Library);
    my $Scm = $LibraryInfo{$Library}{"Scm"};
    my $ScmAddr = $LibraryInfo{$Library}{"ScmAddr"};
    return "" if(not $Scm or not $ScmAddr);
    my $SourcePath = "$REPOSITORY_PATH/$Library/$Library-source";
    return "" if(not -e $SourcePath or not isCheckouted($Library));
    return "" if($Scm eq "cvs");# FIXME: cvs log in the correct format is needed
    if(not check_network())
    {
        print "WARNING: can't checkout log for current (scm) version of $Library: network is down\n";
        return "";
    }
    print "chekouting current (scm) log from $ScmAddr\n";
    if(not check_scm($Library))
    {
        print STDERR "ERROR: source repository for $Library is not available at the moment ($ScmAddr)\n";
        return "";
    }
    if($Scm eq "git")
    {
        return `cd $SourcePath && git log --summary --no-merges 2>/dev/null`;
    }
    elsif($Scm eq "svn")
    {
        return `cd $SourcePath && svn log 2>/dev/null`;
    }
    elsif($Scm eq "cvs")
    {
        return `cd $SourcePath && cvs log -N 2>/dev/null`;
    }
    elsif($Scm eq "darcs")
    {
        return `cd $SourcePath && darcs changes 2>/dev/null`;
    }
    else
    {
        return "";
    }
}

sub init_signals()
{
    return if(not defined $Config{"sig_name"}
    or not defined $Config{"sig_num"});
    my $No = 0;
    my @Numbers = split(/\s/, $Config{"sig_num"} );
    foreach my $Name (split(/\s/, $Config{"sig_name"})) 
    {
        if(not $SigName{$Numbers[$No]} or $Name=~/\A(SEGV|ABRT)\Z/)
        {
            $SigNo{$Name} = $Numbers[$No];
            $SigName{$Numbers[$No]} = $Name;
        }
        $No+=1;
    }
}

sub remove_Library($)
{
    my $Library = $_[0];
    return if(not $Library);
    rmtree("compat_reports/$Library/");
    rmtree("test_results/$Library/");
    rmtree("changelogs/$Library/");
    rmtree("tests/$Library/");
    rmtree("versions/$Library.html");
    rmtree("$INSTALL_PATH_DEFAULT/$Library/");
    rmtree("$INSTALL_PATH_DEBUG/$Library/");
    rmtree("$REPOSITORY_PATH/$Library/");
}

sub remove_LibraryVersion($$$)
{
    my ($Library, $Version, $Type) = @_;
    return if(not $Version or not $Library or not $Type);
    if($Type eq "compat_reports")
    {
        if(-d "compat_reports/$Library")
        {
            foreach my $Path (split(/\n/, `find \"compat_reports/$Library/\" -maxdepth 1 -type d -name "*_to_$Version"`),
            split(/\n/, `find \"compat_reports/$Library/\" -maxdepth 1 -type d -name "$Version\_to_*"`))
            {
                rmtree($Path);
            }
        }
    }
    elsif($Type eq "tests")
    {
        rmtree("test_results/$Library/$Version/");
        rmtree("tests/$Library/$Version/");
    }
    elsif($Type eq "tests_t2c")
    {
        rmtree("tests_t2c/$Library/$Version/");
    }
    elsif($Type eq "changelog")
    {
        rmtree("changelogs/$Library/$Version/");
    }
}

sub scenario()
{
    if(defined $Help)
    {
        HELP_MESSAGE();
        exit(0);
    }
    if(defined $HelpDB)
    {
        HELP_MESSAGE_DB();
        exit(0);
    }
    if(defined $ShowVersion)
    {
        print "Upstream Tracker $UPSTREAM_TRACKER_VERSION\nCopyright (C) The Linux Foundation\nCopyright (C) Institute for System Programming, RAS\nLicenses GPLv2 and LGPLv2 <http://www.gnu.org/licenses/>\nThis program is free software: you can redistribute it and/or modify it.\n\nWritten by Andrey Ponomarenko.\n";
        exit(0);
    }
    init_signals();
    detect_default_paths();
    check_state();
    remove_Library("simple_lib_c");
    remove_Library("simple_lib_cpp");
    if($ListKnownLibs)
    {
        foreach my $Library (sort {lc($a) cmp lc($b)} keys(%LibraryInfo))
        {
            print "$Library\n";
        }
        print "Total: ".keys(%LibraryInfo)."\n";
        exit(0);
    }
    if($TargetLibrary and not defined $LibraryInfo{$TargetLibrary}
    and defined $LibraryInfo{"lib".$TargetLibrary})
    {# add prefix to the library name
        $TargetLibrary="lib".$TargetLibrary;
    }
    if($TargetLibrary and not defined $LibraryInfo{$TargetLibrary}
    and defined $LibraryInfo{lc($TargetLibrary)})
    {
        $TargetLibrary=lc($TargetLibrary);
    }
    $TargetVersion = canonify_version($TargetVersion);
    $TargetVersion1 = canonify_version($TargetVersion1);
    $TargetVersion2 = canonify_version($TargetVersion2);
    
    if($TargetLibrary and not defined $LibraryInfo{$TargetLibrary})
    {
        print STDERR "ERROR: hasn't info about $TargetLibrary, please update lib_info.xml\n";
        exit(1);
    }
    if($RemoveInfo)
    {# complete removing
        if($TargetLibrary)
        {
            if($TargetVersion)
            {
                print "removing $TargetLibrary-$TargetVersion from the system\n";
                remove_LibraryVersion($TargetLibrary, $TargetVersion, "compat_reports");
                remove_LibraryVersion($TargetLibrary, $TargetVersion, "tests");
                remove_LibraryVersion($TargetLibrary, $TargetVersion, "changelog");
                remove_LibraryVersion($TargetLibrary, $TargetVersion, "tests_t2c");
                rmtree("$INSTALL_PATH_DEFAULT/$TargetLibrary/$TargetVersion/");
                rmtree("$INSTALL_PATH_DEBUG/$TargetLibrary/$TargetVersion/");
                rmtree($VersionPath{$TargetLibrary}{$TargetVersion}); # removing a package
            }
            else
            {
                print "removing $TargetLibrary from the system\n";
                remove_Library($TargetLibrary);
            }
        }
        else
        {
            print STDERR "ERROR: library name was not selected (-l option)\n";
            exit(1);
        }
    }
    if($GetCurrent)
    {
        if(check_network())
        {
            if($TargetLibrary)
            {
                print "searching for current (scm) version of the $TargetLibrary ...\n";
                getCurrentVersion_Library($TargetLibrary);
            }
            else
            {
                foreach my $Library (sort keys(%LibraryInfo))
                {
                    print "searching for current (scm) version of the $Library ...\n";
                    getCurrentVersion_Library($Library);
                }
            }
        }
        else
        {
            print STDERR "ERROR: network is down\n";
        }
    }
    if($GetNew)
    {
        if(check_network())
        {
            if($TargetLibrary)
            {
                getNewVersions_Library($TargetLibrary);
            }
            else
            {
                foreach my $Library (sort keys(%LibraryInfo))
                {
                    getNewVersions_Library($Library);
                }
            }
        }
        else
        {
            print STDERR "ERROR: network is down\n";
        }
    }
    if($GetChangelogs)
    {
        if($TargetLibrary)
        {
            if($TargetVersion)
            {
                changelog_LibraryVersion($TargetLibrary, $TargetVersion);
            }
            else
            {
                changelog_Library($TargetLibrary);
            }
        }
        else
        {
            foreach my $Library (sort keys(%LibraryInfo))
            {
                changelog_Library($Library);
            }
        }
    }
    if($UnInstall)
    {
        if($TargetLibrary)
        {
            if($TargetVersion)
            {
                my $IPath = $INSTALL_PATH."/".$TargetLibrary."-".$TargetVersion."/";
                print "uninstalling $TargetLibrary $TargetVersion from \'$IPath\'\n";
                rmtree($IPath);
            }
            else
            {
                my $IPath = $INSTALL_PATH."/".$TargetLibrary."/";
                print "uninstalling $TargetLibrary from \'$IPath\'\n";
                rmtree($IPath);
            }
        }
        else
        {
            print STDERR "ERROR: library name was not selected (-l option)\n";
            exit(1);
        }
    }
    if($Install or $ReInstall)
    {
        if($TargetLibrary)
        {
            if($TargetVersion)
            {
                install_LibraryVersion($TargetLibrary, $TargetVersion);
            }
            else
            {
                install_Library($TargetLibrary);
            }
        }
        else
        {
            foreach my $Library (sort keys(%LibraryInfo))
            {
                install_Library($Library);
            }
        }
    }
    if($TestAPI)
    {
        if(-f "api-sanity-autotest.pl")
        {
            if($TargetLibrary)
            {
                if($TargetVersion)
                {
                    test_LibraryVersion($TargetLibrary, $TargetVersion, "standalone");
                }
                else
                {
                    test_Library($TargetLibrary);
                }
            }
            else
            {
                foreach my $Library (sort keys(%LibraryInfo))
                {
                    test_Library($Library);
                }
            }
        }
        else
        {
            print STDERR "ERROR: api-sanity-autotest.pl was not found\n";
            exit(1);
        }
    }
    if($GenT2C)
    {
        if(-f "api-sanity-autotest.pl")
        {
            if($TargetLibrary)
            {
                if($TargetVersion)
                {
                    test_LibraryVersion($TargetLibrary, $TargetVersion, "t2c");
                }
                else
                {
                    genT2C_Library($TargetLibrary);
                }
            }
            else
            {
                foreach my $Library (sort keys(%LibraryInfo))
                {
                    genT2C_Library($Library);
                }
            }
        }
        else
        {
            print STDERR "ERROR: api-sanity-autotest.pl was not found\n";
            exit(1);
        }
    }
    if($CmpABI)
    {
        if(-f "abi-compliance-checker.pl")
        {
            if($TargetLibrary)
            {
                if($TargetVersion1 and $TargetVersion2)
                {
                    cmp_LibraryVersions($TargetLibrary, $TargetVersion1, $TargetVersion2);
                }
                else
                {
                    cmp_Library($TargetLibrary);
                }
            }
            else
            {
                foreach my $Library (sort keys(%LibraryInfo))
                {
                    cmp_Library($Library);
                }
            }
        }
        else
        {
            print STDERR "ERROR: abi-compliance-checker.pl was not found\n";
            exit(1);
        }
    }
    if($MakeCppCompat)
    {
        if($TargetLibrary)
        {
            if($TargetVersion)
            {
                mkcpp_LibraryVersion($TargetLibrary, $TargetVersion);
            }
            else
            {
                mkcpp_Library($TargetLibrary);
            }
        }
        else
        {
            print STDERR "ERROR: library name was not selected (-l option)\n";
            exit(1);
        }
    }
    if($UpdateInfo)
    {
        my $IncludeCurrent_Saved = $IncludeCurrent;
        foreach my $Library ($TargetLibrary?($TargetLibrary):sort keys(%RepositoryProvide))
        {
            $IncludeCurrent = 1 if($VersionPath{$Library}{"current"} or $IncludeCurrent);
            if(not $LibraryInfo{$Library}{"SkipOnUpdate"}{"get_new_versions"})
            {
                getNewVersions_Library($Library);
            }
            if(keys(%{$UpdateLog{$Library}{"New"}})
            or $UpdateLog{$Library}{"NewSrc"})
            {
                if(not $LibraryInfo{$Library}{"SkipOnUpdate"}{"get_changelog"})
                {
                    changelog_Library($Library);
                }
                if(not $LibraryInfo{$Library}{"SkipOnUpdate"}{"installing"})
                {
                    install_Library($Library);
                    if(not $LibraryInfo{$Library}{"SkipOnUpdate"}{"abi_tracking"})
                    {
                        cmp_Library($Library);
                    }
                    if(not $LibraryInfo{$Library}{"SkipOnUpdate"}{"runtime_tests"})
                    {
                        test_Library($Library);
                        genT2C_Library($Library);
                    }
                }
            }
            $IncludeCurrent = $IncludeCurrent_Saved;
        }
    }
    createLog();
    createIndex();
    exit(0);
}

scenario();
